Package: Rarr
Title: A Simple and Performant Native R Reader & Writer for Zarr Arrays
Version: 2.1.26
Authors@R: c(
    person("Mike", "Smith", role = c("aut", "ccp"),
           comment = c(ORCID = "0000-0002-7800-3848", "Maintainer from 2022 to 2025.")),
    person("Hugo", "Gruson", , "hugo.gruson@embl.de", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-4094-1476")),
    person("Artür", "Manukyan", role = "ctb"),
    person("Sharla", "Gelfand", role = "ctb"),
    person("German Network for Bioinformatics Infrastructure - de.NBI", role = "fnd",
           comment = c(ROR = "01vmpm840"))
  )
Description: The Zarr specification defines a format for chunked,
    compressed, N-dimensional arrays.  Its design allows efficient access
    to subsets of the stored array, and supports both local and cloud
    storage systems.  Rarr aims to implement this specification in R with
    minimal reliance on an external tools or libraries.
License: MIT + file LICENSE
URL: https://huber-group-embl.github.io/Rarr/,
        https://github.com/Huber-group-EMBL/Rarr
BugReports: https://github.com/Huber-group-EMBL/Rarr/issues
Depends: R (>= 4.2.0)
Imports: curl, grumpy, jsonlite, lifecycle, paws.storage, R.utils,
        rlang, utils
Suggests: BiocStyle, knitr, rmarkdown, testthat (>= 3.0.0), withr,
        ZarrArray
VignetteBuilder: knitr
biocViews: DataImport
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
SystemRequirements: GNU make
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
git_url: https://git.bioconductor.org/packages/Rarr
git_branch: devel
git_last_commit: 3f0925a
git_last_commit_date: 2026-07-19
Repository: Bioconductor 3.24
Date/Publication: 2026-07-20
NeedsCompilation: yes
Packaged: 2026-07-20 23:44:13 UTC; biocbuild
Author: Mike Smith [aut, ccp] (ORCID: <https://orcid.org/0000-0002-7800-3848>,
    Maintainer from 2022 to 2025.),
  Hugo Gruson [aut, cre] (ORCID: <https://orcid.org/0000-0002-4094-1476>),
  Artür Manukyan [ctb],
  Sharla Gelfand [ctb],
  German Network for Bioinformatics Infrastructure - de.NBI [fnd] (ROR:
    <https://ror.org/01vmpm840>)
Maintainer: Hugo Gruson <hugo.gruson@embl.de>

# Generated by roxygen2: do not edit by hand

export(create_empty_zarr_array)
export(read_zarr_array)
export(read_zarr_attributes)
export(read_zarr_consolidated_metadata)
export(update_zarr_array)
export(use_blosc)
export(use_bz2)
export(use_gzip)
export(use_lz4)
export(use_lzma)
export(use_zlib)
export(use_zstd)
export(write_zarr_array)
export(write_zarr_attributes)
export(write_zarr_group)
export(zarr_consolidate_metadata)
export(zarr_overview)
importFrom(R.utils,getAbsolutePath)
importFrom(grumpy,
  convert_bytes_to_array,
  parse_npy_datatype
)
importFrom(jsonlite,
  fromJSON,
  read_json,
  write_json
)
importFrom(paws.storage,s3)
importFrom(stats,setNames)
importFrom(utils,
  modifyList,
  relist,
  tail
)
useDynLib(Rarr)

Zarr arrays with Rarr
================

- [Introduction to Rarr](#introduction-to-rarr)
  - [Limitations with **Rarr**](#limitations-with-rarr)
- [Quick start guide](#quick-start-guide)
  - [Installation and setup](#installation-and-setup)
  - [Reading a from a local Zarr
    array](#reading-a-from-a-local-zarr-array)
  - [Reading from S3 storage](#read-s3)
  - [Writing to a Zarr array](#writing-to-a-zarr-array)
- [Required system libraries](#required-system-libraries)

<!-- badges: start -->

| GitHub Actions | Bioconductor Build Sysytem | Test Coverage |
|:--------------:|:-------------:|:-----:|
| [![R-CMD-check](https://github.com/Huber-group-EMBL/Rarr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/Huber-group-EMBL/Rarr/actions/workflows/R-CMD-check.yaml) | [![Bioconductor Status](https://bioconductor.org/shields/build/devel/bioc/Rarr.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/Rarr/) | [![Codecov test coverage](https://codecov.io/gh/Huber-group-EMBL/Rarr/graph/badge.svg)](https://app.codecov.io/gh/Huber-group-EMBL/Rarr) |

<!-- badges: end -->

# Introduction to Rarr

The Zarr specification defines a format for chunked, compressed,
N-dimensional arrays. It’s design allows efficient access to subsets of
the stored array, and supports both local and cloud storage systems.
Zarr is experiencing increasing adoption in a number of scientific
fields, where multi-dimensional data are prevalent.

**Rarr** is intended to be a simple interface to reading and writing
individual Zarr arrays. It is developed in R and C with no reliance on
external libraries or APIs for interfacing with the Zarr arrays.
Additional compression libraries (e.g. blosc) are bundled with **Rarr**
to provide support for datasets compressed using these tools.

## Limitations with **Rarr**

If you know about Zarr arrays already, you’ll probably be aware they can
be stored in hierarchical groups, where additional meta data can explain
the relationship between the arrays. Currently, **Rarr** is not designed
to be aware of these hierarchical Zarr array collections. However, the
component arrays can be read individually by providing the path to them
directly.

Currently, there are also limitations on the Zarr datatypes that can be
accessed using **Rarr**. For now most numeric types can be read into R,
although in some instances e.g. 64-bit integers there is potential for
loss of information.

See the dedicated [“Supported Zarr features”
vignette](https://huber-group-embl.github.io/Rarr/articles/features.html)
for more information.

# Quick start guide

## Installation and setup

If you want to quickly get started reading an existing Zarr array with
the package, this section should have the essentials covered. First, we
need to install **Rarr**[^1] with the commands below.

``` r
## we need BiocManager to perform the installation
if (!require("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}
## install Rarr
BiocManager::install("Rarr")
```

Once **Rarr** is installed, we have to load it into our R session:

``` r
library(Rarr)
```

**Rarr** can be used to read files either on local disk or on remote S3
storage systems. First lets take a look at reading from a local file.

## Reading a from a local Zarr array

To demonstrate reading a local file, we’ll pick the example file
containing 32-bit integers arranged in the “column first” ordering.

``` r
zarr_example <- system.file(
  "extdata",
  "zarr_examples",
  "column-first",
  "int32.zarr",
  package = "Rarr"
)
```

### Exploring the data

We can get an summary of the array properties, such as its shape and
datatype, using the function `zarr_overview()`[^2].

``` r
zarr_overview(zarr_example)
```

    ## Type: Array
    ## Path: /tmp/Rtmp3ZvhzD/temp_libpath541a2b36c23a/Rarr/extdata/zarr_examples/column-first/int32.zarr
    ## Shape: 30 x 20 x 10
    ## Chunk Shape: 10 x 10 x 5
    ## No. of Chunks: 12 (3 x 2 x 2)
    ## Data Type: int32
    ## Endianness: little
    ## Compressor: blosc

You can use this to check that the location is a valid Zarr array, and
that the shape and datatype of the array content are what you are
expecting. For example, we can see in the output above that the data
type (`int32`) corresponds to what we expect.

### Reading the Zarr array

The summary information retrieved above is required, as to read the
array with **Rarr** you need to know the shape and size of the array
(unless you want to read the entire array). From the previous output we
can see our example array has three dimensions of size 30 x 20 x 10. We
can select the subset we want to extract using a `list`. The list must
have the same length as the number of dimensions in our array, with each
element of the list corresponding to the indices you want to extract in
that dimension.

``` r
index <- list(1:4, 1:2, 1)
```

We then extract the subset using `read_zarr_array()`:

``` r
read_zarr_array(zarr_example, index = index)
```

    ## , , 1
    ## 
    ##      [,1] [,2]
    ## [1,]    1    2
    ## [2,]    1    0
    ## [3,]    1    0
    ## [4,]    1    0

## Reading from S3 storage

Reading files in S3 storage works in a very similar fashion to local
disk. This time the path needs to be a URL to the Zarr array. We can
again use `zarr_overview()` to quickly retrieve the array metadata.

``` r
s3_address <- "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0076A/10501752.zarr/0"
zarr_overview(s3_address)
```

    ## Type: Array
    ## Path: https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0076A/10501752.zarr/0
    ## Shape: 50 x 494 x 464
    ## Chunk Shape: 1 x 494 x 464
    ## No. of Chunks: 50 (50 x 1 x 1)
    ## Data Type: float64
    ## Endianness: little
    ## Compressor: blosc

The output above indicates that the array is stored in 50 chunks, each
containing a slice of the overall data. In the example below we use the
`index` argument to extract the first and tenth slices from the array.
Choosing to read only 2 of the 50 slices is much faster than if we opted
to download the entire array before accessing the data.

``` r
z2 <- read_zarr_array(s3_address, index = list(c(1, 10), NULL, NULL))
```

We then plot our two slices on top of one another using the `image()`
function.

``` r
## plot the first slice in blue
image(
  log2(z2[1, , ]),
  col = hsv(h = 0.6, v = 1, s = 1, alpha = 0:100 / 100),
  asp = dim(z2)[2] / dim(z2)[3],
  axes = FALSE
)
## overlay the tenth slice in green
image(
  log2(z2[2, , ]),
  col = hsv(h = 0.3, v = 1, s = 1, alpha = 0:100 / 100),
  asp = dim(z2)[2] / dim(z2)[3],
  axes = FALSE,
  add = TRUE
)
```

<img src="inst/rmd/imgs/plot-raster-1.jpeg" alt="" width="30%" />

**Note:** if you receive the error message
`"Error in stop(aws_error(request$error)) : bad error message"` it is
likely you have some AWS credentials available in to your R session,
which are being inappropriately used to access this public bucket.
Please see the section @ref(s3-client) for details on how to set
credentials for a specific request.

## Writing to a Zarr array

Up until now we’ve only covered reading existing Zarr array into R.
However, **Rarr** can also be used to write R data to disk following the
Zarr specification. To explore this, lets create an example array we
want to save as a Zarr. In this case it’s going to be a three
dimensional array and store the values 1 to 600.

``` r
x <- array(1:600, dim = c(10, 10, 6))
```

``` r
path_to_new_zarr <- file.path(tempdir(), "new.zarr")
write_zarr_array(
  x = x,
  zarr_array_path = path_to_new_zarr,
  chunk_dim = c(10, 5, 1)
)
```

We can check that the contents of the Zarr array is what we’re
expecting. Since the contents of the whole array will be too large to
display here, we use the `index` argument to extract rows 6 to 10, from
the 10th column and 1st slice. That should be the values 96, 97, 98, 99,
100, but retaining the 3-dimensional array structure of the original
array. The second line below uses `identical()` to confirm that reading
the whole Zarr returns something equivalent to our original input `x`.

``` r
read_zarr_array(zarr_array_path = path_to_new_zarr, index = list(6:10, 10, 1))
```

    ## , , 1
    ## 
    ##      [,1]
    ## [1,]   96
    ## [2,]   97
    ## [3,]   98
    ## [4,]   99
    ## [5,]  100

``` r
identical(read_zarr_array(zarr_array_path = path_to_new_zarr), x)
```

    ## [1] TRUE

# Required system libraries

To provide support for BLOSC and zstd compression tools Rarr links
against libraries providing these tools. If you have them installed on
your system Rarr will attempt to use those versions. If they are not
detected then Rarr will compile and use versions that are distributed
with the package. Either way the functionality will available, however
if you are using the system libraries and then later remove them Rarr
may fail to work correctly.

This only concerns users installing the package from source. If you are
using the pre-built binaries for Windows or Mac OSX distributed by
Bioconductor then this should not be an issue for you.

[^1]: you only need to do the installation step once

[^2]: This is essentially reading and formatting the array metadata that
    accompanies any Zarr array.

# Description

This folder contains the script used to produce the example Zarr arrays found
in the inst/extdata/zarr_examples folder.  These are created using the 
reference implementation of the Zarr specification found in 
[zarr-python](https://github.com/zarr-developers/zarr-python).

The files are used by the Rarr's unit tests to check reading consistency with
the reference implementation across a range of features.  The files aim to test
the following aspects of the specification:

* various data types e.g. different sized integers, floats, characters
* column-first and row-first layouts
* different compression tools
* special case fill values e.g. infinity

The naming of the generated files should be indicative of the feature under 
test.

# Generating the data

## Natively

If you have a python installation already available you can use either `pip`
or `conda` to install `zarr-python` as appropriate.

```
pip install zarr
```

or

```
conda install -c conda-forge zarr
```

You can then run the script using python:

```
python ./create_test_data.py
```

This will try to create files in the folder `/data`.  If thia location does not
exist you will either need to create it yourself or modify the relevant paths
in `create_test_data.py`.

## Using Docker

If you do not have a version of python and zarr readily available, you can
alternatively use the Dockerfile provided to generate the data.  The following
command will build a container called `zarr_tester`.

```
docker build -t zarr-tester .   
```

To generate the data you then use the command below.  This will produce the
example files in the folder `/tmp/zarr` on your local filesystem.  Modify that
path in the call to `docker run` if it is not appropriate for your system.

```
mkdir /tmp/zarr_examples
docker run --rm --name zarr-tester --user ${UID}:${UID} \
  -v /tmp/zarr_examples:/data \
  zarr-tester
```

Patches created to remove symbols that R CMD check has decided are warnings.  
This include abort(), exit(), fprintf() etc.  We replace them will R compatible
functions like error(), Rprintf() etc.  

If we update any of these compression libraries, apply the patch with:

cd lib/lz4-1.9.2
patch --ignore-whitespace lz4.c < ../../patches/lz4.diff

is.compact <- function(x) {
  .Call("is_compact", x, PACKAGE = "Rarr")
}

reindex <- function(x, from = 1L, to = 0L) {
  offset <- from - to
  if (is.compact(x)) {
    res <- (x[1L] - offset):(x[length(x)] - offset)
  } else {
    res <- x - offset
  }
  return(res)
}

.convert_metadata_version <- function(metadata, version_from, version_to) {
  if (version_from != 2L || version_to != 3L) {
    # nocov start
    stop(
      "Only conversion from version 2 to version 3 is supported.",
      call. = FALSE
    )
    # nocov end
  }

  dt <- metadata$datatype

  if (length(metadata$shape) == 0L) {
    # Empty tuple in shape means we are dealing with a scalar.
    metadata$shape <- metadata$chunks <- 1L
  }

  metadata_v3 <- list(
    node_type = "array",
    zarr_format = 2L,
    datatype = dt,
    data_type = .convert_dtype_version(
      dt$base_type,
      dt$nbytes,
      dt$endian,
      version_from,
      version_to
    ),
    shape = metadata$shape,
    chunk_grid = list(
      name = "regular",
      configuration = list(chunk_shape = metadata$chunks)
    ),
    chunk_key_encoding = list(
      name = "default",
      configuration = list(
        # The default was "." in v2
        separator = metadata$dimension_separator %||% "."
      )
    ),
    fill_value = metadata$fill_value,
    codecs = list()
  )

  # Transpose codec only makes sense for more than 1 dimension
  if (length(metadata_v3$shape) > 1L) {
    metadata_v3$codecs$transpose <- list(
      name = "transpose",
      configuration = list(
        order = switch(
          metadata$order,
          # default in numpy is "C"
          "C" = seq_along(metadata$shape) - 1L, # zero indexed
          "F" = rev(seq_along(metadata$shape)) - 1L
        )
      )
    )
  }

  for (filter in metadata$filters) {
    metadata_v3$codecs[[filter$id]] <- list(
      name = filter$id
    )
  }

  if (!is.null(metadata_v3$codecs[["vlen-utf8"]])) {
    # In v3, vlen-utf8 applies to 'string' type
    metadata_v3$data_type <- "string"
    metadata_v3$datatype$base_type <- "string"
  } else {
    metadata_v3$codecs$bytes <- list(
      name = "bytes",
      configuration = list("endian" = dt$endian %||% NA_character_)
    )
  }

  if (!is.null(metadata$compressor$id)) {
    metadata_v3$codecs[[metadata$compressor$id]] <- list(
      name = metadata$compressor$id,
      configuration = metadata$compressor[names(metadata$compressor) != "id"]
    )
  }

  return(metadata_v3)
}

.convert_dtype_version <- function(
  base_type,
  nbytes,
  endian,
  version_from,
  version_to
) {
  if (version_from != 2L || version_to != 3L) {
    # nocov start
    stop(
      "Only conversion from version 2 to version 3 is supported.",
      call. = FALSE
    )
    # nocov end
  }
  if (length(base_type) > 1L) {
    return(
      list(
        name = "struct",
        configuration = list(
          fields = mapply(
            .convert_dtype_version,
            base_type,
            nbytes,
            endian,
            MoreArgs = list(
              version_from = version_from,
              version_to = version_to
            ),
            SIMPLIFY = FALSE
          )
        )
      )
    )
  }
  switch(
    base_type,
    "unicode" = list(
      name = "fixed_length_utf32",
      configuration = list(
        length_bytes = 8L * nbytes
      )
    ),
    "string" = list(
      name = "null_terminated_bytes",
      configuration = list(
        length_bytes = nbytes
      )
    ),
    "bool" = "bool",
    paste0(base_type, 8L * nbytes)
  )
}

codec_blosc_decode <- function(bytes, ...) {
  .Call(
    "decompress_chunk_BLOSC",
    bytes,
    PACKAGE = "Rarr"
  )
}

codec_blosc_encode <- function(bytes, compressor_config, ...) {
  .Call(
    "compress_chunk_BLOSC",
    bytes,
    compressor_config$typesize,
    compressor_config$cname,
    compressor_config$clevel,
    which(
      compressor_config$shuffle == c("noshuffle", "shuffle", "bitshuffle")
    ) -
      1L,
    compressor_config$blocksize,
    PACKAGE = "Rarr"
  )
}

codec_zlib_decode <- codec_gzip_decode <- function(bytes, ...) {
  memDecompress(
    from = bytes,
    type = "gzip",
    asChar = FALSE
  )
}

codec_bz2_decode <- function(bytes, ...) {
  memDecompress(
    from = bytes,
    type = "bzip2",
    asChar = FALSE
  )
}

codec_lzma_decode <- function(bytes, ...) {
  memDecompress(
    from = bytes,
    type = "xz",
    asChar = FALSE
  )
}

#' @importFrom utils tail
codec_lz4_decode <- function(bytes, ...) {
  # numcodecs docs says:
  # The compressed output includes a 4-byte header storing the original size
  # of the decompressed data as a little-endian 32-bit integer.
  buffer_size <- readBin(bytes, "integer", n = 1L, size = 4L, endian = "little")
  .Call(
    "decompress_chunk_LZ4",
    tail(x = bytes, n = -4L),
    buffer_size,
    PACKAGE = "Rarr"
  )
}

codec_lz4_encode <- function(bytes, ...) {
  ## numpy stores the original size of the buffer in the first 4 bytes after
  ## compression. We should do that too for compatibility
  c(
    writeBin(length(bytes), raw(), size = 4L, endian = "little"),
    .Call("compress_chunk_LZ4", bytes, PACKAGE = "Rarr")
  )
}

codec_zstd_decode <- function(bytes, ...) {
  .Call(
    "decompress_chunk_ZSTD",
    bytes,
    PACKAGE = "Rarr"
  )
}

codec_zstd_encode <- function(bytes, compressor_config, ...) {
  .Call(
    "compress_chunk_ZSTD",
    bytes,
    as.integer(compressor_config$level),
    PACKAGE = "Rarr"
  )
}

#' @importFrom jsonlite write_json
#' @importFrom grumpy parse_npy_datatype
.write_zarr_metadata <- function(
  array_path,
  array_shape,
  dimension_names = NULL,
  chunk_shape,
  data_type,
  fill_value,
  compressor,
  dimension_separator = ".",
  order = "C",
  zarr_version = 3L
) {
  if (!is.null(compressor$id) && compressor$id == "blosc") {
    nbytes <- parse_npy_datatype(data_type)$nbytes
    compressor$typesize <- compressor$typesize %||% nbytes
    if (is.na(compressor$typesize)) {
      # e.g., vlen-utf8
      compressor$typesize <- 1L
    }
  }

  metadata_v2 <- list(
    # the spec states these need to be json arrays, so we need to avoid auto_unboxing
    shape = as.list(array_shape),
    chunks = as.list(chunk_shape),
    dtype = data_type,
    fill_value = fill_value,
    dimension_separator = dimension_separator,
    order = order,
    zarr_format = 2L,
    filters = NULL,
    compressor = compressor
  )

  if (zarr_version == 2L) {
    if (!is.null(dimension_names)) {
      # The spec says we "SHOULDN'T" do this.
      metadata_v2$dimension_names <- dimension_names
    }
    write_json(
      metadata_v2,
      file.path(array_path, ".zarray"),
      auto_unbox = TRUE,
      pretty = TRUE,
      null = "null"
    )
    return(invisible(TRUE))
  }
  if (zarr_version == 3L) {
    metadata_v2$datatype <- parse_npy_datatype(data_type)
    metadata_v3 <- .convert_metadata_version(
      metadata_v2,
      version_from = 2L,
      version_to = 3L
    )
    metadata_v3$zarr_format <- 3L
    if (!is.null(dimension_names)) {
      metadata_v3$dimension_names <- dimension_names
    }
    # FIXME: get rid of these directly in the internal
    metadata_v3$codecs <- unname(metadata_v3$codecs)
    metadata_v3$datatype <- NULL
    write_json(
      metadata_v3,
      file.path(array_path, "zarr.json"),
      auto_unbox = TRUE,
      pretty = TRUE,
      null = "null"
    )
  }
}

#' Read the .zattrs file associated with a Zarr array or group
#'
#' @param zarr_path A character vector of length 1. This provides the
#'   path to a Zarr array or group.
#' @param new.zattrs a list inserted to .zattrs at the `path`.
#' @param overwrite if `TRUE` (the default), existing .zattrs elements will be overwritten by `new.zattrs`.
#' @param zarr_version The version of the Zarr specification to use. If a
#'   metadata file already exists, the version will be inferred from the file.
#'   Otherwise, the default is `3`.
#'
#' @importFrom jsonlite write_json
#'
#' @returns Invisibly, the updated attributes as a named list.
#' This is equivalent to (but faster than) using `read_zarr_attributes()` after writing.
#' If no attributes were present before, this is identical to `new.zattrs`.
#'
#' @importFrom utils modifyList
#' @export
#' @examples
#' z1 <- withr::local_tempdir(fileext = ".zarr")
#' write_zarr_attributes(z1, list(date = "2025-01-01", author = "Jane Doe"))
#'
write_zarr_attributes <- function(
  zarr_path,
  new.zattrs = list(),
  overwrite = TRUE,
  zarr_version = if (has_metadata_v2) 2L else 3L
) {
  has_metadata_v2 <- any(file.exists(file.path(
    zarr_path,
    c(".zarray", ".zgroup", ".zattrs")
  )))
  stopifnot(
    "`zarr_version` must be 2 or 3" = zarr_version %in% c(2L, 3L),
    "list elements should be named" = !is.null(names(new.zattrs))
  )
  zarr_path <- .normalize_array_path(zarr_path)

  if ("" %in% names(new.zattrs)) {
    message("Ignoring unnamed list elements")
    new.zattrs <- new.zattrs[nzchar(names(new.zattrs))]
  }

  old.zattrs <- read_zarr_attributes(zarr_path)

  new.zattrs <- if (overwrite) {
    modifyList(old.zattrs, new.zattrs)
  } else {
    modifyList(new.zattrs, old.zattrs)
  }

  if (zarr_version == 2L) {
    write_json(
      new.zattrs,
      file.path(zarr_path, ".zattrs"),
      auto_unbox = TRUE,
      pretty = TRUE,
      null = "null"
    )
  } else if (zarr_version == 3L) {
    # FIXME: we really want a partial write to the json file
    if (file.exists(file.path(zarr_path, "zarr.json"))) {
      metadata <- read_json(file.path(zarr_path, "zarr.json"))
    } else {
      metadata <- list(
        "zarr_format" = 3L
      )
    }
    metadata$attributes <- new.zattrs
    write_json(
      metadata,
      file.path(zarr_path, "zarr.json"),
      auto_unbox = TRUE,
      pretty = TRUE,
      null = "null"
    )
  }

  invisible(new.zattrs)
}

#' Consolidate Zarr metadata files into a single file
#'
#' This function reads all the metadata files in a Zarr store and consolidates them into a single file.
#' Thanks to this, a single request can be made to retrieve all the elements and their related metadata for a Zarr store,
#' which is especially beneficial for remote stores like S3.
#'
#' @param zarr_store_path A character vector of length 1. This provides the path to a Zarr store.
#' @inheritParams zarr_overview
#' @param action A character string specifying the action to take with the consolidated metadata.
#'   If `"write"` (the default), the consolidated metadata will be written back to the Zarr store.
#'   If `"return"`, the consolidated metadata will be returned as a list without writing it back to the store.
#'   The latter is particularly useful for non-writable stores.
#' @param overwrite A logical value (default `TRUE`) indicating whether to overwrite existing
#'   consolidated metadata when `action` is `"write"`. If `FALSE` and consolidated metadata
#'   already exists, an error will be raised.
#'
#' @returns If `action` is `"return"`, a list containing the consolidated metadata.
#'   Otherwise, the function is called for its side effect and `NULL` is returned invisibly.
#'
#' @importFrom jsonlite fromJSON read_json
#' @export
#'
#' @examples
#' # v2
#' zarr_v2 <- withr::local_tempfile(fileext = ".zarr")
#' dir.create(zarr_v2)
#' jsonlite::write_json(
#'   list("zarr_format" = 2L),
#'   file.path(zarr_v2, ".zgroup")
#' )
#' write_zarr_array(
#'   array(1:4, dim = c(2, 2)),
#'   file.path(zarr_v2, "array1"),
#'   chunk_dim = c(1, 2),
#'   zarr_version = 2L
#' )
#' write_zarr_array(
#'   array(c(3.14, 42.42, 12.96, 7.89), dim = c(2, 2)),
#'   file.path(zarr_v2, "array2"),
#'   chunk_dim = c(1, 2),
#'   zarr_version = 2L
#' )
#' write_zarr_attributes(
#'  file.path(zarr_v2, "array1"),
#'  list(description = "This is array 1")
#' )
#' zarr_consolidate_metadata(zarr_v2, action = "return")
#'
#' zarr_consolidate_metadata(zarr_v2, action = "write")
#' zarr_overview(zarr_v2)
#'
zarr_consolidate_metadata <- function(
  zarr_store_path,
  s3_client = NULL,
  action = c("write", "return"),
  overwrite = TRUE
) {
  action <- match.arg(action)

  zarr_store_path <- .normalize_array_path(zarr_store_path)
  s3_client <- s3_client %||% .create_s3_client(zarr_store_path)

  if (!is.null(s3_client) && action == "write") {
    warning(
      "Consolidating metadata on S3 is not currently supported. Returning consolidated metadata instead.",
      call. = FALSE
    )
    action <- "return"
  }

  if (action == "write" && !overwrite) {
    exists <- !is.null(
      suppressWarnings(
        # We don't want to see the warning about empty consolidated metadata.
        .read_consolidated_metadata(
          zarr_store_path,
          s3_client = s3_client
        )
      )
    )
    if (exists) {
      # TODO: return current consolidated metadata
      stop(
        "Consolidated metadata already exists. ",
        "Set `overwrite = TRUE` to overwrite it.",
        call. = FALSE
      )
    }
  }

  child_meta <- .store_list(zarr_store_path, recursive = TRUE, s3_client)

  # v2 or v3?
  metadata_v2_files <- endsWith(child_meta, ".zarray") |
    endsWith(child_meta, ".zgroup") |
    endsWith(child_meta, ".zattrs")
  metadata_v3_files <- endsWith(child_meta, "zarr.json")
  if (any(metadata_v2_files) && any(metadata_v3_files)) {
    stop(
      "Found both v2 and v3 metadata files. Please resolve this conflict before consolidating metadata.",
      call. = FALSE
    )
  }
  version <- if (any(metadata_v2_files)) 2L else 3L
  metadata_files <- child_meta[metadata_v2_files | metadata_v3_files]

  consolidated <- lapply(
    paste0(zarr_store_path, metadata_files),
    .read_json_file,
    s3_client = s3_client
  )

  if (version == 2L) {
    res <- list(
      zarr_consolidated_format = 1L,
      metadata = setNames(consolidated, metadata_files)
    )
  } else {
    attrs <- read_zarr_attributes(zarr_store_path, s3_client = s3_client)
    res <- list(
      zarr_format = 3L,
      node_type = "group",
      attributes = attrs,
      consolidated_metadata = list(
        kind = "inline",
        must_understand = FALSE,
        metadata = setNames(consolidated, dirname(metadata_files))
      )
    )
  }

  if (action == "return") {
    return(res)
  }

  if (version == 2L) {
    consolidated_path <- paste0(zarr_store_path, ".zmetadata")
  } else {
    consolidated_path <- paste0(zarr_store_path, "zarr.json")
  }

  write_json(
    res,
    consolidated_path,
    auto_unbox = TRUE,
    pretty = TRUE,
    null = "null"
  )
}

test_that("zarr attributes can be written and read", {
  # create zarr array
  path <- withr::local_tempdir(fileext = ".zarr")
  x <- array(runif(n = 10), dim = c(2, 5))
  res <- write_zarr_array(
    x = x,
    zarr_array_path = path,
    chunk_dim = c(2, 5),
    zarr_version = 2
  )

  # add .zattrs to /
  zattrs <- list(foo = "foo", bar = "bar")
  write_zarr_attributes(zarr_path = path, new.zattrs = zattrs)
  expect_true(file.exists(file.path(path, ".zattrs")))

  # check .zattrs
  read.zattrs <- read_zarr_attributes(path)
  expect_identical(read.zattrs, zattrs)
})

test_that("zarr attributes can be updated with new elements", {
  # create zarr array
  path <- withr::local_tempdir(fileext = ".zarr")
  x <- array(runif(n = 10), dim = c(2, 5))
  res <- write_zarr_array(
    x = x,
    zarr_array_path = path,
    chunk_dim = c(2, 5)
  )

  zattrs <- list(foo = "foo", bar = "bar")
  write_zarr_attributes(zarr_path = path, new.zattrs = zattrs)

  # add new elements to .zattrs
  zattrs.new.elem <- list(foo2 = "foo")
  write_zarr_attributes(zarr_path = path, new.zattrs = zattrs.new.elem)
  read.zattrs <- read_zarr_attributes(path)
  expect_identical(read.zattrs, c(zattrs, zattrs.new.elem))
})

test_that("zarr attributes can be overwritten", {
  # create zarr array
  path <- withr::local_tempdir(fileext = ".zarr")
  x <- array(runif(n = 10), dim = c(2, 5))
  res <- write_zarr_array(
    x = x,
    zarr_array_path = path,
    chunk_dim = c(2, 5)
  )

  zattrs <- list(foo = "foo", bar = "bar")
  write_zarr_attributes(zarr_path = path, new.zattrs = zattrs)

  # overwrite
  zattrs.new.elem <- list(foo2 = "foo2")
  write_zarr_attributes(zarr_path = path, new.zattrs = zattrs.new.elem)
  read.zattrs <- read_zarr_attributes(path)
  zattrs[names(zattrs.new.elem)] <- zattrs.new.elem
  expect_identical(read.zattrs, c(zattrs))
})

test_that("zarr attributes overwrite parameter works correctly", {
  # create zarr array
  path <- withr::local_tempdir(fileext = ".zarr")
  x <- array(runif(n = 10), dim = c(2, 5))
  res <- write_zarr_array(
    x = x,
    zarr_array_path = path,
    chunk_dim = c(2, 5)
  )

  zattrs <- list(foo = "foo", bar = "bar", foo2 = "foo2")
  write_zarr_attributes(zarr_path = path, new.zattrs = zattrs)

  zattrs.new.elem <- list(foo2 = "foo")
  write_zarr_attributes(
    zarr_path = path,
    new.zattrs = zattrs.new.elem,
    overwrite = FALSE
  )
  read.zattrs <- read_zarr_attributes(path)
  zattrs[names(zattrs.new.elem)] <- "foo2"
  expect_true(all(names(read.zattrs) %in% names(zattrs)))
  expect_true(all(read.zattrs %in% zattrs))
})

test_that("zarr attributes handles lists with empty names", {
  # create zarr array
  path <- withr::local_tempdir(fileext = ".zarr")
  x <- array(runif(n = 10), dim = c(2, 5))
  res <- write_zarr_array(
    x = x,
    zarr_array_path = path,
    chunk_dim = c(2, 5)
  )

  zattrs <- list(foo = "foo", bar = "bar")
  write_zarr_attributes(zarr_path = path, new.zattrs = zattrs)

  # test lists with empty names
  zattrs.new.elem <- list("empty", full = "full")
  expect_message(write_zarr_attributes(
    zarr_path = path,
    new.zattrs = zattrs.new.elem
  ))
  read.zattrs <- read_zarr_attributes(path)
  zattrs[["full"]] <- "full"
  expect_true(all(names(read.zattrs) %in% names(zattrs)))
  expect_true(all(read.zattrs %in% zattrs))
})

test_that("v3 zarr attributes work correctly", {
  v3_attrs <- read_zarr_attributes(
    system.file(
      "extdata",
      "zarr_examples",
      "metadata",
      "v3_attr.zarr",
      package = "Rarr"
    )
  )
  expect_identical(
    v3_attrs,
    list(custom = "Hello, Zarr!")
  )

  expect_identical(
    read_zarr_attributes(
      system.file(
        "extdata",
        "zarr_examples",
        "metadata",
        "v3.zarr",
        package = "Rarr"
      )
    ),
    list()
  )
})

test_that("read_zarr_attributes errors clearly for invalid mixed arrays", {
  invalid_zarr <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "invalid_mixed.zarr",
    package = "Rarr"
  )

  expect_snapshot_error(
    read_zarr_attributes(invalid_zarr)
  )
})

test_that("read_zarr_attributes from s3", {
  expect_snapshot(
    read_zarr_attributes(
      "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.3/9836842.zarr/"
    )
  )
})

test_that("behaviour on missing attributes", {
  expect_warning(
    read_zarr_attributes(
      system.file(
        "extdata",
        "zarr_examples",
        "column-first",
        "boolean.zarr",
        package = "Rarr"
      ),
      missing = "warning"
    ),
    "No file that could contain attributes"
  )
  expect_error(
    read_zarr_attributes(
      system.file(
        "extdata",
        "zarr_examples",
        "column-first",
        "boolean.zarr",
        package = "Rarr"
      ),
      missing = "error"
    ),
    "No file that could contain attributes"
  )
})

test_that("zarr_consolidate_metadata() on v2", {
  zarr_v2 <- withr::local_tempfile(fileext = ".zarr")
  dir.create(zarr_v2)
  jsonlite::write_json(
    list("zarr_format" = 2L),
    file.path(zarr_v2, ".zgroup")
  )
  write_zarr_array(
    array(1:4, dim = c(2, 2)),
    file.path(zarr_v2, "array1"),
    chunk_dim = c(1, 2),
    zarr_version = 2L
  )
  write_zarr_array(
    array(c(3.14, 42.42, 12.96, 7.89), dim = c(2, 2)),
    file.path(zarr_v2, "array2"),
    chunk_dim = c(1, 2),
    zarr_version = 2L
  )
  write_zarr_attributes(
    file.path(zarr_v2, "array1"),
    list(description = "This is array 1")
  )
  expect_snapshot(
    zarr_consolidate_metadata(zarr_v2, action = "return"),
    # "Path" is absolute path so will differ between systems
    transform = function(x) gsub("^(  )?Path: .*", "Path: <path>", x)
  )
})

test_that("zarr_consoliate_metadata() on v3", {
  zarr_v3 <- withr::local_tempfile(fileext = ".zarr")
  dir.create(zarr_v3)
  jsonlite::write_json(
    list(
      zarr_format = 3L,
      node_type = "group"
    ),
    file.path(zarr_v3, "zarr.json"),
    auto_unbox = TRUE
  )
  write_zarr_array(
    array(1:4, dim = c(2, 2)),
    file.path(zarr_v3, "array1"),
    chunk_dim = c(1, 2),
    zarr_version = 3L
  )
  write_zarr_array(
    array(c(3.14, 42.42, 12.96, 7.89), dim = c(2, 2)),
    file.path(zarr_v3, "array2"),
    chunk_dim = c(1, 2),
    zarr_version = 3L
  )
  write_zarr_attributes(
    file.path(zarr_v3, "array1"),
    list(description = "This is array 1")
  )
  expect_snapshot(
    zarr_consolidate_metadata(zarr_v3, action = "return"),
    # "Path" is absolute path so will differ between systems
    transform = function(x) gsub("^(  )?Path: .*", "Path: <path>", x)
  )
})

#' Print a summary of a Zarr array or group
#'
#' When reading a Zarr array using [read_zarr_array()] it is necessary to know
#' it's shape and size. `zarr_overview()` can be used to get a quick overview of
#' the array shape and contents, based on the `.zarray` (Zarr v2) or `zarr.json`
#' (Zarr v3) metadata file each array contains.
#'
#' The function currently prints the following information to the R console:
#'  - array path
#'  - array shape and size
#'  - chunk and size
#'  - the number of chunks
#'  - the datatype of the array
#'  - codec used for data compression (if any)
#'
#' If given the path to a group of arrays the function will attempt to print the
#' details of all sub-arrays in the group.
#'
#' @param zarr_array_path A character vector of length 1.  This provides the
#'   path to a Zarr array or group of arrays. This can either be on a local file
#'   system or on S3 storage.
#' @param s3_client A list representing an S3 client.  This should be produced
#' by [paws.storage::s3()].
#' @param as_data_frame Logical determining whether the Zarr array details
#'   should be printed to screen (`FALSE`) or returned as a `data.frame`
#'   (`TRUE`) so they can be used computationally.
#'
#' @return If `as_data_frame = FALSE` the function invisible returns `TRUE` if
#'   successful.  However it is primarily called for the side effect of printing
#'   details of the Zarr array(s) to the screen.  If `as_data_frame = TRUE` then
#'   a `data.frame` containing details of the array is returned.
#'
#' @examples
#'
#' ## Using a local file provided with the package
#' z1 <- system.file("extdata", "zarr_examples", "row-first",
#'   "int32.zarr",
#'   package = "Rarr"
#' )
#'
#' ## read the entire array
#' zarr_overview(zarr_array_path = z1)
#'
#' ## using a file on S3 storage
#' \donttest{
#' z2 <- "https://noaa-nwm-retro-v2-zarr-pds.s3.amazonaws.com/feature_id/"
#' zarr_overview(z2)
#' }
#' @export
zarr_overview <- function(
  zarr_array_path,
  s3_client = NULL,
  as_data_frame = FALSE
) {
  zarr_array_path <- .normalize_array_path(zarr_array_path)
  s3_client <- s3_client %||% .create_s3_client(path = zarr_array_path)

  dot_zmeta <- .read_consolidated_metadata(
    zarr_path = zarr_array_path,
    nodes = "array",
    s3_client = s3_client
  )
  if (!is.null(dot_zmeta)) {
    is_array <- vapply(
      dot_zmeta$metadata,
      function(x) !is.null(x$node_type) && x$node_type == "array",
      FUN.VALUE = logical(1L)
    )
    arrays <- names(dot_zmeta$metadata)[is_array]
    tmp <- lapply(
      arrays,
      FUN = function(x) {
        if (endsWith(x, "/.zarray")) {
          v2_attrs <- sub(".zarray", ".zattrs", x, fixed = TRUE)
          has_attrs <- length(dot_zmeta$metadata[[v2_attrs]]) > 0L
        } else {
          has_attrs <- length(dot_zmeta$metadata[[x]]$attributes) > 0L
        }
        .rbind_array_metadata(
          array_name = x,
          metadata = dot_zmeta$metadata,
          zarr_array_path = zarr_array_path,
          has_attrs = has_attrs
        )
      }
    )
    res <- do.call(rbind.data.frame, tmp)
    if (as_data_frame) {
      return(res)
    }
    cat("Type: Group of Arrays\n")
    cat(
      "Path: ",
      normalizePath(zarr_array_path, mustWork = FALSE),
      "\n",
      sep = ""
    )
    cat("Arrays:\n---\n")
    .print_array_metadata(res, indent = "  ")
    invisible(TRUE)
  } else {
    array_metadata <- .read_array_metadata(
      zarr_path = zarr_array_path,
      s3_client = s3_client
    )
    array_attrs <- read_zarr_attributes(
      zarr_path = zarr_array_path,
      s3_client = s3_client
    )
    res <- .rbind_array_metadata(
      array_name = basename(zarr_array_path),
      metadata = array_metadata,
      zarr_array_path = dirname(zarr_array_path),
      has_attrs = length(array_attrs) > 0L
    )
    if (as_data_frame) {
      return(res)
    }
    cat("Type: Array\n")
    .print_array_metadata(res)
    invisible(TRUE)
  }
}

.rbind_array_metadata <- function(
  array_name,
  metadata,
  zarr_array_path,
  has_attrs
) {
  if (array_name %in% names(metadata)) {
    array_metadata <- metadata[[array_name]]
    array_name <- dirname(array_name)
  } else {
    array_metadata <- metadata
  }
  # FIXME: This is redundant with .read_array_metadata() but we didn't go
  # through it if we are dealing with consolidated metadata.
  names(array_metadata$codecs) <- vapply(
    array_metadata$codecs,
    function(x) x$name,
    character(1L)
  )

  chunk_shape <- unlist(array_metadata$chunk_grid$configuration$chunk_shape)
  data_shape <- unlist(array_metadata$shape)
  nchunks <- ceiling(
    data_shape / chunk_shape
  )

  codecs <- array_metadata$codecs
  compressor <- names(codecs)[match(
    TRUE,
    names(codecs) %in% CODEC_BYTES_BYTES
  )]
  endianness <- codecs[["bytes"]][["configuration"]][["endian"]]

  res <- data.frame(
    path = paste0(.normalize_array_path(zarr_array_path), array_name),
    data_type = if (is.list(array_metadata$data_type)) {
      array_metadata$data_type$name
    } else {
      array_metadata$data_type
    },
    # We already introduce default values in .convert_metadata_version() but
    # these do not exist when reading v3 metadata directly.
    endianness = endianness %||% NA_character_,
    compressor = compressor %||% NA_character_
  )
  res$dim <- list(data_shape)
  res$chunk_dim <- list(chunk_shape)
  res$nchunks <- list(nchunks)
  res$attributes <- has_attrs
  return(res)
}

.print_array_metadata <- function(array_metadata_df, indent = "") {
  fields <- c(
    "Path",
    "Shape",
    "Chunk Shape",
    "No. of Chunks",
    "Data Type",
    "Endianness",
    "Compressor",
    "Attributes"
  )
  compressor <- array_metadata_df$compressor
  compressor[is.na(compressor)] <- "None"
  fields <- paste0(indent, fields, ": %s")
  formatted <- sprintf(
    paste(fields, collapse = "\n"),
    array_metadata_df$path,
    vapply(
      array_metadata_df$dim,
      function(x) {
        paste(unlist(x), collapse = " x ")
      },
      character(1L)
    ),
    vapply(
      array_metadata_df$chunk_dim,
      function(x) {
        paste(unlist(x), collapse = " x ")
      },
      character(1L)
    ),
    vapply(
      array_metadata_df$nchunks,
      function(x) {
        chunks <- unlist(x)
        paste0(prod(chunks), " (", paste(chunks, collapse = " x "), ")")
      },
      character(1L)
    ),
    array_metadata_df$data_type,
    array_metadata_df$endianness,
    compressor,
    # Equivalent but ~5x faster than ifelse(attributes, "yes", "no")
    c("yes", "no")[2L - array_metadata_df$attributes]
  )
  cat(formatted, sep = "\n---\n")
}

#' Read the `.zarray` or `zarr.json` metadata file associated with a Zarr array
#'
#' @param zarr_path A character vector of length 1.  This provides the
#'   path to a Zarr array or group of arrays. This can either be on a local file
#'   system or on S3 storage.
#' @param s3_client A list representing an S3 client.  This should be produced
#' by [paws.storage::s3()].
#'
#' @returns A list containing the array metadata
#'
#' @importFrom grumpy parse_npy_datatype
#'
#' @keywords internal
.read_array_metadata <- function(zarr_path, s3_client = NULL) {
  zarr_path <- .normalize_array_path(zarr_path)

  metadata_file <- c(".zarray", "zarr.json") |>
    .store_check_exist(zarr_path, files = _, s3_client = s3_client)

  if (metadata_file[".zarray"] && metadata_file["zarr.json"]) {
    stop(
      "The path contains both `.zarray` (Zarr V2 specification) and ",
      "`zarr.json` (Zarr V3 specification) metadata files.\n",
      "An array or group must conform to either the Zarr V2 or V3 ",
      "specification.",
      call. = FALSE
    )
  }
  if (!any(metadata_file)) {
    stop(
      "The path does not contain any metadata files. ",
      "It must contain one of:\n",
      "  - `.zarray` (Zarr V2 specification)\n",
      "  - `zarr.json` (Zarr V3 specification)",
      call. = FALSE
    )
  }

  metadata_path <- paste0(zarr_path, names(metadata_file)[metadata_file])

  metadata <- .read_json_file(metadata_path, s3_client)

  if (metadata$zarr_format == 2L) {
    ## if we do this here, we save many repeated calls to .parse_npy_datatype
    ## the parsed version is used each time a chunk is read
    metadata$datatype <- parse_npy_datatype(metadata$dtype)
    metadata <- .convert_metadata_version(
      metadata,
      version_from = 2L,
      version_to = 3L
    )
  } else if (metadata$node_type == "array" && metadata$zarr_format == 3L) {
    metadata <- .normalize_v3_metadata(metadata)
  }
  metadata$fill_value <- .update_fill_value(
    metadata$fill_value,
    metadata$datatype,
    metadata$data_type
  )
  if (
    is.list(metadata$data_type) &&
      metadata$data_type$name %in%
        c("fixed_length_utf32", "null_terminated_bytes")
  ) {
    # It is annoying to have to handle both strings & lists in `data_type` in
    # downstream code, so we convert the fixed-length string types to their v2
    # equivalents here.
    # Their length has been calculated in .parse_datatype_v3() and stored in nbytes.
    metadata$data_type <- switch(
      metadata$data_type$name,
      "fixed_length_utf32" = "unicode",
      "null_terminated_bytes" = "string"
    )
  }

  return(metadata)
}

#' Apply Zarr v3-specific fixups to raw array metadata
#'
#' Called after reading a `zarr.json` file for an array node.  Performs four
#' normalisation steps that are needed before the metadata can be used by the
#' rest of the package:
#'
#' 1. Parse the `data_type` string into an R-friendly `datatype` list.
#' 2. Name `codecs` by their `name` field for O(1) lookup.
#' 3. Normalise scalar arrays (empty `shape`) to shape `1`.
#' 4. Inject a default `transpose` codec when absent (R uses F-order so we
#'    always need to know the intended order).
#' 5. Replicate the `endian` field across all fields of struct datatypes so
#'    that downstream chunk-reading code can treat it uniformly.
#'
#' @param metadata A list as returned by [.read_json_file()] for a Zarr v3
#'   array node (i.e. `metadata$zarr_format == 3L` and
#'   `metadata$node_type == "array"`).
#'
#' @returns The modified `metadata` list.
#'
#' @keywords internal
.normalize_v3_metadata <- function(metadata) {
  metadata$datatype <- .parse_datatype_v3(metadata$data_type)

  # We shouldn't have any case where x$name is NULL since the v3 spec states
  # 'name' MUST be a plain string.
  names(metadata$codecs) <- vapply(
    metadata$codecs,
    function(x) x$name,
    character(1L)
  )

  if (length(metadata$shape) == 0L) {
    # Empty tuple in shape means we are dealing with a scalar.
    metadata$shape <- 1L
    metadata$chunk_grid <- list(
      name = "regular",
      configuration = list(chunk_shape = 1L)
    )
  }

  # This needs to happen after we address the scalar edge case
  if (is.null(metadata$codecs[["transpose"]])) {
    # We need to make sure this is always present because we do the
    # reverse of what this codec is telling us (since R uses F-order).
    # So even when using the implicit default, we need to add it here.
    metadata$codecs[["transpose"]] <- list(
      name = "transpose",
      configuration = list(order = seq_along(metadata$shape) - 1L)
    )
  }

  # Set endian to NA for 1-byte types; replicate across struct fields
  if (!is.null(metadata$codecs[["bytes"]])) {
    endian <- metadata$codecs[["bytes"]]$configuration$endian %||%
      NA_character_
    # In v3 struct datatypes, endian is defined only once for the whole array
    # but our reading infra (based on v2) expects it per base type.
    endian <- rep(endian, length(metadata$datatype$base_type))
    metadata$codecs[["bytes"]]$configuration$endian <- endian
    metadata$datatype$endian <- endian
  }

  return(metadata)
}

#' Convert special fill values from strings to numbers
#'
#' Special case fill values (NaN, Inf, -Inf) are encoded as strings in the Zarr
#' metadata.  R will create arrays of type character if these are defined and
#' the chunk isn't present on disk. This function updates the fill value to be
#' R's representation of these special values, so numeric arrays are created. A
#' "null" fill value implies no missing values. We set this to NA as you can't
#' create an array of type NULL in R. It should have no impact if there are
#' really no missing values.
#'
#' @param fill_value The fill value as read from the Zarr metadata.
#' @param datatype A list of details for the array datatype.  Expected to be
#' produced by [grumpy::parse_npy_datatype()].
#'
#' @returns Returns a modified fill value.  The returned value will be equal
#'   to the input, but with the correct type, unless the `fill_value` entry was one
#'   of: NULL, "NaN", "Infinity" or "-Infinity".
#'
#' @keywords internal
#' @noRd
.update_fill_value <- function(fill_value, datatype, data_type) {
  if (length(datatype$base_type) > 1L) {
    if (data_type$name == "struct") {
      return(mapply(
        function(fill, bt, endian, nbytes) {
          .update_fill_value(
            fill,
            list(base_type = bt, endian = endian, nbytes = nbytes)
          )
        },
        fill_value,
        datatype$base_type,
        datatype$endian,
        datatype$nbytes
      ))
    }
    if (data_type$name == "structured") {
      decoded_fill_value <- jsonlite::base64_dec(fill_value)
      fill_value <- vector("list", length(datatype$base_type))
      for (i in seq_along(datatype$base_type)) {
        el_fill_value <- decoded_fill_value[seq_len(datatype$nbytes[i])]
        fill_value[[i]] <- convert_bytes_to_array(
          el_fill_value,
          datatype$base_type[i],
          shape = NULL,
          datatype$nbytes[i],
          "little"
        )
        decoded_fill_value <- decoded_fill_value[-seq_len(datatype$nbytes[i])]
      }
      return(fill_value)
    }
  }
  ## a null fill value implies no missing values.
  ## We set to NA as you can't create an array of NULL in R
  if (is.null(fill_value)) {
    fill_value <- NA
  } else if (fill_value %in% c("NaN", "Infinity", "-Infinity")) {
    if (datatype$base_type != "string") {
      fill_value <- switch(
        fill_value,
        "NaN" = NaN,
        "Infinity" = Inf,
        "-Infinity" = -Inf
      )
    }
  } else if (is.numeric(fill_value)) {
    fill_value <- switch(
      datatype$base_type,
      "float" = as.double(fill_value),
      "int" = as.integer(fill_value),
      "uint" = as.integer(fill_value),
      "complex" = as.complex(fill_value),
      "string" = as.character(fill_value),
      "unicode" = as.character(fill_value),
      fill_value
    )
  } else if (datatype$base_type == "float" && startsWith(fill_value, "0x")) {
    # FIXME: the spec only defines this for floats but surely it makes
    # sense to also apply it to int and uint?
    hex_clean <- sub("^0x", "", fill_value)
    byte_pairs <- paste0(
      "0x",
      regmatches(hex_clean, gregexpr(".{2}", hex_clean))[[1L]]
    )
    # Swap endianness
    endian <- switch(
      datatype$endian,
      "little" = "big",
      "big" = "little"
    )
    fill_value <- readBin(
      as.raw(byte_pairs),
      what = "double",
      size = datatype$nbytes,
      endian = endian
    )
  }
  return(fill_value)
}

#' Read consolidated metadata file
#'
#' @details
#' This is stored in the `.zmetadata` file at the root of a Zarr store.
#' Note that it is not documented in the official Zarr specification, because
#' it is not (yet?) part of the standard.
#'
#' It is implemented in zarr-python and discussed under the
#' "consolidated metadata" phrase.
#'
#' In particular, it lists the location of all the metadata files for arrays in
#' the current group, so it is not necessary to crawl to discover them.
#'
#' @references
#' <https://zarr.readthedocs.io/en/latest/user-guide/consolidated_metadata.html>
#'
#'
#' @inheritParams .read_array_metadata
#'
#' @importFrom grumpy parse_npy_datatype
#'
#' @keywords internal
.read_consolidated_metadata <- function(
  zarr_path,
  nodes = c("group", "array"),
  s3_client = NULL
) {
  s3_client <- s3_client %||% .create_s3_client(zarr_path)

  metadata_file <- .store_check_exist(
    zarr_path,
    files = c(".zmetadata", "zarr.json"),
    s3_client
  )
  if (!metadata_file[".zmetadata"] && !metadata_file["zarr.json"]) {
    return(NULL)
  }
  if (metadata_file[".zmetadata"] && metadata_file["zarr.json"]) {
    stop(
      "The path contains both `.zmetadata` (Zarr V2 specification) and ",
      "`zarr.json` (Zarr V3 specification) consolidated metadata files.\n",
      "An array or group must conform to either the Zarr V2 or V3 ",
      "specification.",
      call. = FALSE
    )
  }

  zmeta_path <- paste0(zarr_path, names(metadata_file)[metadata_file])

  # At this stage, we are sure the file exists
  zmeta <- .read_json_file(zmeta_path, s3_client)

  if (metadata_file[".zmetadata"]) {
    arrays <- names(zmeta$metadata)[endsWith(
      names(zmeta$metadata),
      "/.zarray"
    )]
    zmeta$metadata[arrays] <- lapply(
      zmeta$metadata[arrays],
      function(metadata) {
        metadata$datatype <- parse_npy_datatype(metadata$dtype)
        .convert_metadata_version(
          metadata,
          version_from = 2L,
          version_to = 3L
        )
      }
    )
  } else {
    if (zmeta$node_type == "array") {
      return(NULL)
    }
    zmeta <- zmeta$consolidated_metadata
  }
  if (length(zmeta$metadata) == 0L) {
    warning(
      "The consolidated metadata file was found but was empty. ",
      "Consider filling it with `zarr_consolidate_metadata()`.",
      call. = FALSE
    )
  }
  return(zmeta)
}

#' Read the attributes associated with a Zarr array or group
#'
#' @inheritParams .read_array_metadata
#' @param missing A character vector of length 1. This determines the behaviour
#'   when no file containing attributes is found. This can be one of:
#'  - "ignore" (the default): an empty list is returned silently
#'  - "warning": a warning is issued and an empty list is returned.
#'  - "error": an error is raised.
#'
#' @returns A list containing the attributes. If the file containing attributes
#' (`.zattrs` for Zarr v2 or `zarr.json` for Zarr v3) exists but no attributes
#' are provided, an empty list is returned.
#'
#' @examples
#' read_zarr_attributes(
#'   "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0048A/9846152.zarr"
#' )
#'
#' @export
read_zarr_attributes <- function(
  zarr_path,
  s3_client = NULL,
  missing = c("ignore", "warning", "error")
) {
  missing <- match.arg(missing)
  zarr_path <- .normalize_array_path(zarr_path)
  ## determine if this is a local or S3 array
  s3_client <- s3_client %||% .create_s3_client(path = zarr_path)

  exists_attribute_files <- .store_check_exist(
    zarr_path,
    c(".zattrs", "zarr.json"),
    s3_client
  )

  if (!any(exists_attribute_files)) {
    msg <- paste(
      "No file that could contain attributes",
      "(either `.zattrs` for v2 or `zarr.json` for v3)",
      "was found in the path."
    )
    if (missing == "error") {
      stop(msg, call. = FALSE)
    }
    if (missing == "warning") {
      warning(msg, call. = FALSE)
    }
    return(list())
  }
  if (all(exists_attribute_files)) {
    stop(
      "The path contains both `.zattrs` (v2 Zarr specification) ",
      "and `zarr.json` (v3 Zarr specification) files, which is not allowed.",
      call. = FALSE
    )
  }
  attribute_file <- names(exists_attribute_files)[exists_attribute_files]
  attribute_path <- paste0(
    zarr_path,
    attribute_file
  )

  zattrs <- .read_json_file(attribute_path, s3_client)

  if (attribute_file == "zarr.json") {
    zattrs <- zattrs[["attributes"]]
  }

  # Normalize cases of empty named list vs empty list vs NULL
  if (length(zattrs) == 0L) {
    return(list())
  }

  return(zattrs)
}

# Counting function calls ensures we don't repeatedly perform the same task when
# trying to read data (e.g., reading metadata over and over), which would be
# terrible for performance, but could happen more easily that one imagine in a
# complex codebase.

test_that("metadata is read and converted only once", {
  # Eventually, it would be great to use
  # https://github.com/r-lib/withr/issues/200
  local({
    read_metadata_calls <- 0L
    parsedatatype_calls <- 0L
    trace(
      grumpy::parse_npy_datatype,
      function(x) {
        parsedatatype_calls <<- parsedatatype_calls + 1L
      },
      print = FALSE
    )
    trace(
      Rarr:::.read_array_metadata,
      function(x) {
        read_metadata_calls <<- read_metadata_calls + 1L
      },
      print = FALSE
    )
    read_zarr_array(
      system.file(
        "extdata",
        "zarr_examples",
        "column-first",
        "int32.zarr",
        package = "Rarr"
      )
    )
    expect_identical(read_metadata_calls, 1L)
    expect_identical(parsedatatype_calls, 1L)
  })

  local({
    convert_metadata_calls <- 0L
    trace(
      Rarr:::.convert_metadata_version,
      function(x) {
        convert_metadata_calls <<- convert_metadata_calls + 1L
      },
      print = FALSE
    )
    zarr_overview(
      system.file(
        "extdata",
        "zarr_examples",
        "column-first",
        "int32.zarr",
        package = "Rarr"
      )
    )
    expect_identical(convert_metadata_calls, 1L)
  })
})

test_that("zarr_overview returns data.frame for single array", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int64.zarr",
    package = "Rarr"
  )

  # Return results as a data.frame
  df <- zarr_overview(zarr_c, as_data_frame = TRUE)
  expect_s3_class(df, "data.frame")
  expect_identical(dim(df), c(1L, 8L))
})

test_that("zarr_overview console output matches snapshot for single array", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int64.zarr",
    package = "Rarr"
  )

  expect_snapshot(
    zarr_overview(zarr_c, as_data_frame = FALSE),
    # "Path" is absolute path so will differ between systems
    transform = function(x) gsub("^Path: .*", "Path: <path>", x)
  )
})

test_that("zarr_overview works with consolidated metadata store", {
  zarr_store_consolidated <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "consolidated.zarr",
    package = "Rarr"
  )

  df <- zarr_overview(zarr_store_consolidated, as_data_frame = TRUE)
  expect_s3_class(df, "data.frame")
  expect_identical(dim(df), c(3L, 8L))
  expect_identical(
    colnames(df),
    c(
      "path",
      "data_type",
      "endianness",
      "compressor",
      "dim",
      "chunk_dim",
      "nchunks",
      "attributes"
    )
  )

  # From https://objects.eodc.eu:443/e05ab01a9d56408d82ac32d69a5aae2a:202505-s02msil2a/30/products/cpm_v256/S2B_MSIL2A_20250530T101559_N0511_R065_T32TPT_20250530T130924.zarr
  complex_consolidated <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "complex_consolidated.zarr",
    package = "Rarr"
  )
  metadata_df <- zarr_overview(complex_consolidated, as_data_frame = TRUE)
  # "path" is absolute path so will differ between systems
  metadata_df$path <- gsub(
    complex_consolidated,
    "",
    metadata_df$path,
    fixed = TRUE
  )
  expect_snapshot(
    metadata_df
  )
})

test_that("zarr_overview console output matches snapshot for consolidated store", {
  zarr_store_consolidated <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "consolidated.zarr",
    package = "Rarr"
  )

  expect_snapshot(
    zarr_overview(zarr_store_consolidated, as_data_frame = FALSE),
    # "Path" is absolute path so will differ between systems
    transform = function(x) gsub("^(  )?Path: .*", "Path: <path>", x)
  )
})

test_that("zarr_overview works with consolidated v3 metadata", {
  zarr_store_consolidated_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "consolidated_v3.zarr",
    package = "Rarr"
  )
  expect_snapshot(
    zarr_overview(zarr_store_consolidated_v3, as_data_frame = FALSE),
    # "Path" is absolute path so will differ between systems
    transform = function(x) gsub("^(  )?Path: .*", "Path: <path>", x)
  )
})

test_that("zarr_overview works with v3 metadata", {
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "v3.zarr",
    package = "Rarr"
  )

  df <- zarr_overview(zarr_v3, as_data_frame = TRUE)
  expect_s3_class(df, "data.frame")
  expect_identical(dim(df), c(1L, 8L))
})

test_that("zarr_overview doesn't choke on empty consolidated metadata", {
  zarr_empty_consolidated <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "empty_consolidated.zarr",
    package = "Rarr"
  )
  zarr_overview(zarr_empty_consolidated) |>
    expect_warning("The consolidated metadata file was found but was empty.")
})

test_that("zarr_overview console output matches snapshot for v3 metadata", {
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "v3.zarr",
    package = "Rarr"
  )

  expect_snapshot(
    zarr_overview(zarr_v3, as_data_frame = FALSE),
    # "Path" is absolute path so will differ between systems
    transform = function(x) gsub("^Path: .*", "Path: <path>", x)
  )
})

test_that("zarr_overview throws error for missing metadata files", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "missing.zarr",
    package = "Rarr"
  )

  expect_error(
    zarr_overview(zarr_c),
    "The path does not contain any metadata files."
  )
})

test_that("zarr_overview throws error for mixed v2/v3 zarr", {
  invalid_zarr <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "invalid_mixed.zarr",
    package = "Rarr"
  )

  expect_snapshot_error(
    zarr_overview(invalid_zarr)
  )

  invalid_consolidated <- system.file(
    "extdata",
    "zarr_examples",
    "metadata",
    "invalid_mixed_consolidated.zarr",
    package = "Rarr"
  )
  expect_snapshot_error(
    zarr_overview(invalid_consolidated)
  )
})

test_that("missing metadata on S3 results in a clear error", {
  with_mocked_bindings(
    expect_error(
      zarr_overview(
        "http://s3.example.com/bucket/example.zarr/",
        list()
      ),
      "The path does not contain any metadata files."
    ),
    .s3_object_exists = function(...) FALSE
  )
})

test_that("Anonymous S3 access is detected properly", {
  path <- "https://www.test.com/bucket/file1"
  s3_client <- .create_s3_client(path)

  ## This ensures .get_credentials always returns an error, even if the
  ## host machine has credentials available
  with_mocked_bindings(
    .get_credentials = function(...) stop(),
    expect_true(s3_client$.internal$config$credentials$anonymous)
  )
})

test_that("Authenticated S3 access works", {
  skip_if_offline("s3.embl.de")
  skip_on_bioc()
  skip_on_ci()

  withr::with_envvar(
    c(
      "AWS_ACCESS_KEY_ID" = "bYUBYVg1AsEreuDgtg5K",
      "AWS_SECRET_ACCESS_KEY" = "r8FrLXc9dseD6V1P3htsu7ZBzP7Gszsd3sM1G4KX"
    ),
    {
      expect_snapshot(zarr_overview(
        "https://s3.embl.de/rarr-testing/bzip2.zarr"
      ))
      expect_snapshot(read_zarr_array(
        "https://s3.embl.de/rarr-testing/bzip2.zarr"
      ))
    }
  )
})

test_that("Denied access errors return clear error messages", {
  skip_if_offline("s3.embl.de")
  skip_on_bioc()
  skip_on_ci()

  # This is a real zarr store, but we don't have access
  expect_error(
    zarr_overview("https://s3.embl.de/rarr-testing/lzma.zarr"),
    "Denied"
  )
})

test_that("JSON metadata is not unboxed", {
  skip_if_offline("livingobjects.ebi.ac.uk")
  skip_on_bioc()
  skip_on_ci()

  # We don't want to save the metadata and test this locally. It has to be tested on a real S3 store because it relates
  # to a bug in the code branch.
  # See https://github.com/Huber-group-EMBL/Rarr/pull/170
  zarr_overview(
    "https://livingobjects.ebi.ac.uk/idr/share/ome2024-ngff-challenge/idr0066/ExpA_VIP_ASLM_on_XZ_slice1088.zarr/0"
  ) |>
    expect_snapshot()
})

#' @export
#' @rdname ZarrArray-deprecated
#'
#' @title Deprecated DelayedArray backend functions
#'
#' @description
#' `r lifecycle::badge('superseded')`
#'
#' The DelayedArray backend has moved to a dedicated package: the ZarrArray package
#' (<https://github.com/Bioconductor/ZarrArray>).
#'
#' @param ... Passed to the new function in the ZarrArray package.
#'
#' @importFrom lifecycle deprecate_warn
ZarrArray <- function(...) {
  deprecate_warn(
    when = "1.12.0",
    what = "ZarrArray()",
    with = "ZarrArray::ZarrArray()",
    details = paste(
      "The functions related to the DelayedArray backend have moved",
      "to the dedicated ZarrArray package (https://github.com/Bioconductor/ZarrArray)."
    )
  )
  if (!requireNamespace("ZarrArray", quietly = TRUE)) {
    stop("ZarrArray() requires the ZarrArray package")
  }
  ZarrArray::ZarrArray(...)
}

#' @export
#' @rdname ZarrArray-deprecated
writeZarrArray <- function(...) {
  deprecate_warn(
    when = "1.12.0",
    what = "writeZarrArray()",
    with = "ZarrArray::writeZarrArray()",
    details = paste(
      "The functions related to the DelayedArray backend have moved",
      "to the dedicated ZarrArray package (https://github.com/Bioconductor/ZarrArray)."
    )
  )
  if (!requireNamespace("ZarrArray", quietly = TRUE)) {
    stop("writeZarrArray() requires the ZarrArray package")
  }
  ZarrArray::writeZarrArray(...)
}

#!/usr/bin/env Rscript

args <- commandArgs(trailingOnly = TRUE)

# simple --key=val parser
options <- args |>
  grepv(pattern = "^--", x = _) |>
  strsplit("=", fixed = TRUE) |>
  do.call(rbind, args = _)

opt_list <- setNames(options[, 2], gsub("^--", "", options[, 1]))

res <- Rarr::read_zarr_array(
  zarr_array_path = opt_list[["array_path"]]
)

.create_chunk_names <- function(chunk_indices, metadata) {
  if (nrow(chunk_indices) == 0L) {
    # In the DelayedArray framework, we can have integer(0) indices
    # https://github.com/Huber-group-EMBL/Rarr/issues/112.
    return(character(0L))
  }
  if (metadata$zarr_format == 3L && identical(metadata$shape, 1L)) {
    # Special case for scalar in v3
    return("c")
  }

  dim_separator <- metadata$chunk_key_encoding$configuration$separator %||% "/"

  # Faster and ready for https://github.com/zarr-developers/zarr-extensions/pull/69
  chunk_name_template <- paste(
    rep("%s", ncol(chunk_indices)),
    collapse = dim_separator
  )
  if (metadata$zarr_format == 3L) {
    chunk_name_template <- paste("c", chunk_name_template, sep = dim_separator)
  }

  chunk_names <- do.call(sprintf, c(chunk_name_template, chunk_indices))

  return(chunk_names)
}

#' Precompute index positions grouped by chunk
#'
#' For each chunk touched by `index`, returns the positions (1-based) within
#' each dimension of `index` that fall inside that chunk, together with the
#' within-chunk indices needed to extract values from the chunk array.
#'
#' @param index A list of integer vectors, one per dimension, giving the
#'   requested array indices (1-based).
#' @param metadata List of array metadata as returned by `.read_array_metadata()`.
#'   Used to derive chunk name keys via `.create_chunk_names()`.
#' @param chunk_dim Integer vector of length equal to the number of dimensions
#'   of the array, specifying the size of each chunk in each dimension.
#'
#' @returns A named list keyed by chunk names (same format as
#'   `.create_chunk_names()`, e.g. `"c/0/1/0"` for Zarr V3).  Each element is
#'   a list with two components:
#'   * `positions`: a per-dimension list of integer vectors of positions into
#'     the corresponding `index` vector that map to that chunk.
#'   * `index_in_chunk`: a per-dimension list of 1-based integer vectors
#'     giving the within-chunk coordinates corresponding to `positions`.
#'
#' @importFrom utils relist
#'
#' @keywords internal
#' @noRd
.chunk_positions_by_chunk <- function(
  index,
  metadata,
  chunk_dim = unlist(metadata$chunk_grid$configuration$chunk_shape)
) {
  index0 <- lapply(index, reindex, from = 1L, to = 0L)
  if (!is.integer(chunk_dim)) {
    chunk_dim <- as.integer(chunk_dim)
  }
  # FIXME:
  # - make this work for compat sequence that don't start at one
  # - fold the second step (index_in_chunk) here
  if (
    all(vapply(index, is.compact, logical(1L))) &&
      all(vapply(index, min, integer(1L)) == 1L)
  ) {
    per_dim <- mapply(
      \(i, cs) {
        res <- .Call(
          "chop_vec",
          i,
          cs,
          PACKAGE = "Rarr"
        )
        setNames(res, 0L:(length(res) - 1L))
      },
      index,
      chunk_dim,
      SIMPLIFY = FALSE
    )
  } else {
    index0 <- unlist(index0)
    per_dim <- (index0 %/% rep(chunk_dim, times = lengths(index))) |>
      relist(index) |>
      lapply(function(x) split(seq_along(x), x))
    index0 <- relist(index0, index)
  }
  chunk_keys <- do.call(expand.grid, lapply(per_dim, names))
  key_strings <- .create_chunk_names(chunk_keys, metadata)
  setNames(
    lapply(seq_along(key_strings), function(i) {
      positions <- mapply(
        \(d, k) d[[k]],
        per_dim,
        chunk_keys[i, ],
        SIMPLIFY = FALSE
      )
      index_in_chunk <- mapply(
        \(idx, pos, cs) reindex(idx[pos] %% cs, from = 0L, to = 1L),
        index0,
        positions,
        chunk_dim,
        SIMPLIFY = FALSE
      )
      list(positions = positions, index_in_chunk = index_in_chunk)
    }),
    key_strings
  )
}

#' @useDynLib Rarr
#' @keywords internal
"_PACKAGE"

## usethis namespace: start
## usethis namespace: end
NULL

# Backport from R 4.4.0
"%||%" <- function(x, y) {
  if (is.null(x)) y else x # nolint: coalesce_linter.
}

# Backport from R 4.6.0
`%notin%` <- function(x, table) {
  match(x, table, nomatch = 0L) == 0L
}

# Backport from R 4.7.0
Partial <- function(f, ...) {
  f <- match.fun(f)
  args <- list(...)
  function(...) {
    do.call(f, c(args, list(...)))
  }
}

Partial_with_splicing <- function(f, ...) {
  f <- match.fun(f)
  args <- rlang::list2(...)
  function(...) {
    do.call(f, c(args, list(...)))
  }
}

#' @keywords internal
check_index <- function(index, metadata) {
  index_len <- length(index)
  shape <- metadata$shape
  ## check we have the correct number of dimensions
  if (index_len != length(shape)) {
    stop(
      "The number of dimensions provided to 'index' does not match the shape of the array"
    )
  }

  ## If any dimensions are NULL transform into the entirety of that dimension
  ## Otherwise check provided indices are valid
  failed <- rep_len(FALSE, index_len)
  for (i in seq_len(index_len)) {
    if (is.null(index[[i]])) {
      index[[i]] <- seq_len(shape[[i]])
    } else if (
      length(index[[i]]) > 0L &&
        (min(index[[i]]) < 1L || max(index[[i]]) > shape[[i]])
    ) {
      failed[i] <- TRUE
    } else {
      index[[i]] <- as.integer(index[[i]])
    }
  }

  if (any(failed)) {
    stop(sprintf(
      "Selected indices for dimension(s) %s are out of range.",
      paste(which(failed), collapse = " & ")
    ))
  }

  return(index)
}

#' Subset extraction for an array with a variable number of dimensions.
#'
#' @param x Array to extract from.
#' @param idx List of index vectors, one per dimension.
#'
#' @returns The extracted sub-array (with `drop = FALSE`).
#'
#' @keywords internal
.extract_chunk <- function(x, idx) {
  # do.call() has the same performance if we ever need to drop rlang dependency
  # but this is more aesthetically pleasing and rlang is likely to always be
  # somewhere in the dependency tree.
  rlang::inject(x[!!!idx, drop = FALSE])
}

.parse_datatype_v3 <- function(typestr) {
  if (is.list(typestr)) {
    if (typestr$name == "fixed_length_utf32") {
      return(list(
        base_type = "unicode",
        nbytes = typestr$configuration$length_bytes
      ))
    }
    if (typestr$name == "null_terminated_bytes") {
      return(list(
        base_type = "string",
        nbytes = typestr$configuration$length_bytes
      ))
    }
    if (typestr$name == "struct") {
      internal_types <- lapply(typestr$configuration$fields, function(field) {
        .parse_datatype_v3(field$data_type)
      })
      return(
        list(
          base_type = vapply(
            internal_types,
            `[[`,
            "base_type",
            FUN.VALUE = character(1L)
          ),
          nbytes = vapply(
            internal_types,
            `[[`,
            "nbytes",
            FUN.VALUE = integer(1L)
          )
        )
      )
    }
    if (typestr$name == "structured") {
      internal_types <- lapply(typestr$configuration$fields, function(field) {
        .parse_datatype_v3(field[[2L]])
      })
      return(
        list(
          base_type = vapply(
            internal_types,
            `[[`,
            "base_type",
            FUN.VALUE = character(1L)
          ),
          nbytes = vapply(
            internal_types,
            `[[`,
            "nbytes",
            FUN.VALUE = integer(1L)
          )
        )
      )
    }
    # nocov start
    stop("Unsupported data type: ", typestr$name, call. = FALSE)
    # nocov end
  }

  entry <- SUPPORTED_V3_TYPES[[typestr]]
  if (is.null(entry)) {
    # nocov start
    stop("Unsupported data type: ", typestr, call. = FALSE)
    # nocov end
  }
  return(list(
    base_type = entry$base_type,
    nbytes = entry$nbytes
  ))
}

test_that("check_index errors for mismatched dimensions", {
  expect_error(
    check_index(index = list(1), metadata = list(shape = c(10, 10)))
  )
})


test_that("check_index errors for negative indices", {
  expect_error(
    check_index(index = list(-1, 1), metadata = list(shape = c(10, 10)))
  )
})


test_that("check_index errors for indices outside extent", {
  expect_error(
    check_index(index = list(100, 1), metadata = list(shape = c(10, 10)))
  )
})

test_that("integer(0) indexing works correctly", {
  zarr_v2 <- system.file(
    package = "Rarr",
    "extdata",
    "zarr_examples",
    "column-first",
    "int32.zarr"
  )
  expect_no_condition(
    res <- read_zarr_array(zarr_v2, list(integer(0), integer(0), integer(0)))
  )
  expect_shape(res, dim = c(0, 0, 0))

  zarr_v3 <- system.file(
    package = "Rarr",
    "extdata",
    "zarr_examples",
    "column-first",
    "int32_v3.zarr"
  )
  expect_no_condition(
    res <- read_zarr_array(zarr_v3, list(integer(0), integer(0), integer(0)))
  )
  expect_shape(res, dim = c(0, 0, 0))

  # Edge case from https://github.com/Huber-group-EMBL/Rarr/issues/112#issuecomment-3842837618.
  # This happens because the scalar case is special-cased.
  zarr_scalar_v3 <- system.file(
    package = "Rarr",
    "extdata",
    "zarr_examples",
    "scalar",
    "scalar_v3.zarr"
  )
  expect_no_condition(
    res <- read_zarr_array(zarr_scalar_v3, list(integer(0)))
  )
  expect_identical(res, integer(0), ignore_attr = "dim")
})

test_that("out of order indices are handled correctly", {
  # Used in ZarrArray for example.
  # See https://github.com/Huber-group-EMBL/Rarr/issues/211
  zarr_v2 <- system.file(
    package = "Rarr",
    "extdata",
    "zarr_examples",
    "column-first",
    "int32.zarr"
  )
  # Non-compact indices
  expect_no_condition(
    res1 <- read_zarr_array(zarr_v2, list(c(3, 1), c(2, 1), c(1, 3)))
  )
  expect_shape(res1, dim = c(2, 2, 2))
  expect_no_condition(
    res2 <- read_zarr_array(zarr_v2, list(c(1, 3), c(1, 2), c(3, 1)))
  )
  expect_identical(
    res1,
    res2[c(2, 1), c(2, 1), c(2, 1)]
  )

  # Compact index ranges
  expect_no_condition(
    res3 <- read_zarr_array(zarr_v2, list(3:1, 2:1, 1:3))
  )
  expect_shape(res3, dim = c(3, 2, 3))
  expect_no_condition(
    res4 <- read_zarr_array(zarr_v2, list(1:3, 1:2, 3:1))
  )
  expect_identical(
    res3,
    res4[c(3, 2, 1), c(2, 1), c(3, 2, 1)]
  )
})

parse_s3_path <- function(path) {
  if (!any(startsWith(path, c("http://", "https://", "s3://")))) {
    return(NULL)
  }
  if (grepl("amazonaws.com", path, fixed = TRUE)) {
    return(.url_parse_aws(path))
  }
  return(.url_parse_other(path))
}

#' @keywords internal
.url_parse_aws <- function(url) {
  if (grepl(pattern = "^https?://s3\\.", x = url, ignore.case = TRUE)) {
    ## path style address
    url_parts <- regmatches(
      url,
      regexec(
        "^https?://s3\\.([a-z0-9-]*)\\.amazonaws\\.com/?([a-z0-9\\.-]*)/(.*)",
        url,
        ignore.case = TRUE
      )
    )[[1L]]
    region <- url_parts[2L]
    bucket <- url_parts[3L]
    object <- url_parts[4L]
  } else if (
    grepl(
      pattern = "^https?://[a-z0-9\\.-]*\\.s3\\.",
      x = url,
      ignore.case = TRUE
    )
  ) {
    ## virtual-host style address
    tmp <- curl::curl_parse_url(url, params = FALSE)
    bucket <- sub(
      x = tmp$host,
      pattern = "^([a-z0-9\\.-]*)\\.s3.*",
      replacement = "\\1",
      ignore.case = TRUE
    )
    object <- sub("^/?(.*)", "\\1", tmp$path)
    region <- sub(
      x = tmp$host,
      pattern = "^.*\\.s3\\.([a-z0-9-]*)\\.amazonaws\\.com$",
      replacement = "\\1",
      ignore.case = TRUE
    )
  } else {
    # nocov start
    stop(
      "Unknown AWS path style.  Please report this to the package maintainer."
    )
    # nocov end
  }

  res <- list(
    bucket = bucket,
    object = object,
    region = region,
    hostname = "https://s3.amazonaws.com"
  )
  return(res)
}

#' @keywords internal
.url_parse_other <- function(url) {
  parsed_url <- curl::curl_parse_url(url, params = FALSE)
  path_parts <- regmatches(
    parsed_url$path,
    regexec("^/?([a-z0-9:\\.-]*)/(.*)", parsed_url$path, ignore.case = TRUE)
  )[[1L]]
  bucket <- path_parts[2L]
  object <- path_parts[3L]
  hostname <- paste0(parsed_url$scheme, "://", parsed_url$host)

  if (!is.null(parsed_url$port)) {
    hostname <- paste0(hostname, ":", parsed_url$port)
  }

  res <- list(
    bucket = bucket,
    object = object,
    region = "auto",
    hostname = hostname
  )
  return(res)
}

#' This is a modified version of paws.storage:::get_credentials().  It is
#' included to prevent using the `:::` operator.  Look at that function if
#' things stop working.
#'
#' @param credentials Content stored at `.internal$config$credentials` in
#' an object created by `paws.storage::s3()`.
#'
#' @returns A credentials list to be reinserted into a `paws.storage` s3 object.
#' If no valid credentials are found this function will error, which is expected
#' and is caught by `.check_credentials`.
#'
#' @keywords internal
.get_credentials <- function(credentials) {
  for (provider in credentials$provider) {
    args <- names(formals(provider))
    if (is.null(args)) {
      creds <- provider()
    } else {
      creds <- do.call(provider, as.list(credentials)[args])
    }
    if (!is.null(creds)) {
      credentials$creds <- creds
      break
    }
  }
  return(credentials)
}

.check_credentials <- function(s3_client, parsed_url) {
  test <- try(
    .get_credentials(s3_client$.internal$config$credentials),
    silent = TRUE
  )

  if (inherits(test, "try-error")) {
    s3_client <- s3(
      config = list(
        credentials = list(
          anonymous = TRUE
        ),
        region = parsed_url$region,
        endpoint = parsed_url$hostname
      )
    )
  }

  return(s3_client)
}


#' @importFrom paws.storage s3
.create_s3_client <- function(path) {
  parsed_url <- parse_s3_path(path)

  if (is.null(parsed_url)) {
    return(NULL)
  }
  s3_client <- s3(
    config = list(
      region = parsed_url$region,
      endpoint = parsed_url$hostname
    )
  )

  .check_credentials(s3_client, parsed_url)
}

.s3_object_exists <- function(s3_client, Bucket, Key) {
  exists <- s3_client$list_objects_v2(
    Bucket = Bucket,
    Prefix = Key
  )$KeyCount >
    0L

  return(exists)
}

test_that("AWS host URL parses correctly (aws_host)", {
  url <- "https://DOC-EXAMPLE-BUCKET1.s3.us-west-2.amazonaws.com/puppy.png"
  expect_silent(parsed <- Rarr:::.url_parse_aws(url))
  expect_identical(parsed$bucket, "DOC-EXAMPLE-BUCKET1")
  expect_identical(parsed$hostname, "https://s3.amazonaws.com")
  expect_identical(parsed$object, "puppy.png")
  expect_identical(parsed$region, "us-west-2")
})


test_that("AWS path URL parses correctly (aws_path)", {
  url <- "https://s3.us-west-2.amazonaws.com/DOC-EXAMPLE-BUCKET1/puppy.png"
  expect_silent(parsed <- .url_parse_aws(url))
  expect_identical(parsed$bucket, "DOC-EXAMPLE-BUCKET1")
  expect_identical(parsed$hostname, "https://s3.amazonaws.com")
  expect_identical(parsed$object, "puppy.png")
  expect_identical(parsed$region, "us-west-2")
})

test_that("Other S3-like URLs parse correctly (EMBL and EO paths)", {
  paths <- list(
    embl_path = "https://s3.embl.de/rarr-testing/bz2.zarr/.zarray",
    eopf_path_1 = "https://objectstore.eodc.eu:2222/e05ab01a9d56408d82ac32d69a5aae2a:202505-s02msil2a/17/products/cpm_v256/S2A_MSIL2A_20250517T085541_N0511_R064_T35QKA_20250517T112203.zarr/",
    eopf_path_2 = "https://objects.eodc.eu:443/e05ab01a9d56408d82ac32d69a5aae2a:202506-s01siwgrh/03/products/cpm_v256/S1C_IW_GRDH_1SDV_20250603T053151_20250603T053216_002614_0056AB_EE86.zarr"
  )

  # EMBL
  expect_silent(parsed <- Rarr:::.url_parse_other(paths$embl_path))
  expect_identical(parsed$bucket, "rarr-testing")
  expect_identical(parsed$hostname, "https://s3.embl.de")
  expect_identical(parsed$object, "bz2.zarr/.zarray")
  expect_identical(parsed$region, "auto")

  # EOPF style path 1
  expect_silent(parsed <- .url_parse_other(paths$eopf_path_1))
  expect_identical(
    parsed$bucket,
    "e05ab01a9d56408d82ac32d69a5aae2a:202505-s02msil2a"
  )
  expect_identical(parsed$hostname, "https://objectstore.eodc.eu:2222")
  expect_identical(
    parsed$object,
    "17/products/cpm_v256/S2A_MSIL2A_20250517T085541_N0511_R064_T35QKA_20250517T112203.zarr/"
  )
  expect_identical(parsed$region, "auto")

  # EOPF style path 2
  expect_silent(parsed <- .url_parse_other(paths$eopf_path_2))
  expect_identical(
    parsed$bucket,
    "e05ab01a9d56408d82ac32d69a5aae2a:202506-s01siwgrh"
  )
  expect_identical(parsed$hostname, "https://objects.eodc.eu:443")
  expect_identical(
    parsed$object,
    "03/products/cpm_v256/S1C_IW_GRDH_1SDV_20250603T053151_20250603T053216_002614_0056AB_EE86.zarr"
  )
  expect_identical(parsed$region, "auto")
})

#' Convert the metadata `codecs` elements into functions
#'
#' @param codecs A list containing the Zarr v3 codecs
#' @param operation One of "encode" or "decode"
#'
#' @returns An list of 3 lists containing functions:
#' - `bytes_bytes_codecs`: functions to encode/decode raw bytes
#' - `array_array_codecs`: functions to encode/decode R arrays
#' - `array_bytes_codecs`: functions to encode/decode between R arrays and raw bytes
#'
#' @keywords internal
#' @noRd
.configure_codecs <- function(codecs, operation = c("encode", "decode")) {
  codecs_names <- names(codecs)
  operation <- match.arg(operation)

  bytes_bytes_codecs <- intersect(codecs_names, CODEC_BYTES_BYTES)
  array_array_codecs <- intersect(codecs_names, CODEC_ARRAY_ARRAY)
  array_bytes_codecs <- intersect(codecs_names, CODEC_ARRAY_BYTES)

  if (length(array_bytes_codecs) != 1L) {
    stop(
      "A single 'bytes-array' codec must be provided",
      call. = FALSE
    )
  }

  unsupported_codecs <- setdiff(
    codecs_names,
    c(
      bytes_bytes_codecs,
      array_array_codecs,
      array_bytes_codecs
    )
  )
  if (length(unsupported_codecs) > 0L) {
    stop(
      "The following codecs are not supported: ",
      toString(unsupported_codecs)
    )
  }

  bytes_bytes_env <- list()
  array_array_env <- list()
  array_bytes_env <- list()

  if ("transpose" %in% array_array_codecs) {
    # R is already F ordered, so we reverse the order in config
    cfg <- unlist(codecs$transpose$configuration$order) + 1L
    if ("sharding_indexed" %notin% array_bytes_codecs) {
      # For sharding_indexed, the bytes->array function already transposes the
      # array, so we don't need to reverse the order here.
      # FIXME: I'm probably overcomplicating this. A better process within the
      # sharding codec would avoid the special case here.
      cfg <- rev(cfg)
    }
    if (is.unsorted(cfg)) {
      array_array_env[["transpose"]] <- switch(
        operation,
        "encode" = Partial(codec_transpose_encode, indices = cfg),
        "decode" = Partial(codec_transpose_decode, indices = cfg)
      )
    }
  }

  # No loop here since there is always a single array_bytes codec
  cfg <- codecs[[array_bytes_codecs]]$configuration %||% NA_character_
  func_name <- paste("codec", array_bytes_codecs, operation, sep = "_")
  array_bytes_env[[array_bytes_codecs]] <- Partial_with_splicing(
    func_name,
    !!!cfg
  )

  # Compressors
  for (candidate_codec in bytes_bytes_codecs) {
    if (candidate_codec %in% codecs_names) {
      cfg <- codecs[[candidate_codec]]$configuration
      # Aliases
      candidate_codec <- switch(
        candidate_codec,
        "numcodecs.lz4" = "lz4",
        "zstd" = if (
          !is.na(extSoftVersion()["zstd"]) && nzchar(extSoftVersion()["zstd"])
        ) {
          "zstd_base"
        } else {
          "zstd_custom"
        },
        candidate_codec
      )
      if (
        candidate_codec == "zstd_custom" &&
          !identical(Sys.getenv("OLDREL_TESTS"), "true")
      ) {
        warning(
          "Rarr now relies on the base R memCompress() and memDecompress() ",
          "functions for zstd with the expectation it is available everywhere. ",
          "If you are seeing this warning, it means your R installation does ",
          "not have zstd support. Please report at ",
          "https://github.com/Huber-group-EMBL/Rarr/issues. ",
          "If no reports are received, this support will be removed in a ",
          "future release.",
          call. = FALSE
        )
      }
      func_name <- paste(
        "codec",
        candidate_codec,
        operation,
        sep = "_"
      )
      bytes_bytes_env[[candidate_codec]] <- Partial_with_splicing(
        func_name,
        !!!cfg
      )
    }
  }

  return(
    list(
      array_array = array_array_env,
      bytes_bytes = bytes_bytes_env,
      array_bytes = array_bytes_env
    )
  )
}

#' Normalize a Zarr array path
#'
#' Taken from https://zarr.readthedocs.io/en/stable/spec/v2.html#logical-storage-paths
#'
#' @param path Character vector of length 1 giving the path to be normalised.
#'
#' @returns A character vector of length 1 containing the normalised path.
#'
#' @importFrom R.utils getAbsolutePath
#'
#' @keywords internal
.normalize_array_path <- function(path) {
  ## we strip the protocol because it gets messed up by the slash removal later
  supported_protocols <- c("http://", "https://", "s3://")
  if (any(startsWith(path, supported_protocols))) {
    root <- supported_protocols[startsWith(path, supported_protocols)]
  } else {
    ## Replace all backward slash ("\\") with forward slash ("/")
    path <- gsub(x = path, pattern = "\\", replacement = "/", fixed = TRUE)
    path <- getAbsolutePath(path, expandTilde = TRUE)
    root <- sub(x = path, "(^[[:alnum:]:.]*/)?(.*)", replacement = "\\1")
  }
  path <- substring(path, nchar(root) + 1L)

  ## Strip any leading "/" characters
  if (startsWith(path, "/")) {
    path <- substring(path, 2L)
  }
  ## Collapse any sequence of more than one "/" character into a single "/"
  path <- gsub(x = path, pattern = "//+", replacement = "/", fixed = FALSE)
  ## The key prefix is then obtained by appending a single "/" character to
  ## the normalized logical path.
  if (endsWith(path, "/")) {
    path <- paste0(root, path)
  } else {
    path <- paste0(root, path, "/")
  }

  return(path)
}

#' @importFrom stats setNames
.store_check_exist <- function(
  zarr_array_path,
  files,
  s3_client
) {
  if (is.null(s3_client)) {
    is_present <- setNames(
      file.exists(paste0(zarr_array_path, files, recycle0 = TRUE)),
      files
    )
  } else {
    parse_url <- parse_s3_path(zarr_array_path)
    is_present <- vapply(
      files,
      FUN = function(f) {
        key <- paste0(parse_url$object, f)
        .s3_object_exists(s3_client, parse_url$bucket, key)
      },
      FUN.VALUE = logical(1L)
    )
  }

  return(is_present)
}

.store_list <- function(path, recursive, s3_client) {
  if (is.null(s3_client)) {
    list.files(
      path,
      recursive = recursive,
      include.dirs = TRUE,
      all.files = TRUE
    )
  } else {
    s3_client$list_objects(path, recursive = recursive)
  }
}

#' Read a JSON file from local disk or S3
#'
#' @param path Full path (local or S3) to a JSON file.
#' @param s3_client An S3 client produced by [paws.storage::s3()], or `NULL`
#'   for local files.
#'
#' @returns A list as returned by [jsonlite::read_json()] /
#'   [jsonlite::fromJSON()].
#'
#' @importFrom jsonlite read_json fromJSON
#'
#' @keywords internal
.read_json_file <- function(path, s3_client = NULL) {
  if (!is.null(s3_client)) {
    parsed_url <- parse_s3_path(path)
    s3_object <- s3_client$get_object(
      Bucket = parsed_url$bucket,
      Key = parsed_url$object
    )
    # simplifyVector = FALSE is used for consistency with read_json(),
    # used on local files.
    fromJSON(rawToChar(s3_object$Body), simplifyVector = FALSE)
  } else {
    read_json(path)
  }
}

test_that("normalize_array_path works with various path formats", {
  paths <- c(
    # Windows paths
    "c:/foo/bar/baz.zarr",
    "d:\\foo\\bar\\baz.zarr",
    "e:\\foo\\bar\\baz.zarr/",
    "z://foo//bar//baz.zarr",
    # Unix-like paths
    "/foo/bar/baz.zarr",
    "/foo///bar//baz.zarr",
    # URLs
    "https://s3.foo.com/bar/baz.zarr"
  )

  expected_normalized_paths <- c(
    "c:/foo/bar/baz.zarr/",
    "d:/foo/bar/baz.zarr/",
    "e:/foo/bar/baz.zarr/",
    "z:/foo/bar/baz.zarr/",
    "/foo/bar/baz.zarr/",
    "/foo/bar/baz.zarr/",
    "https://s3.foo.com/bar/baz.zarr/"
  )

  actual_normalized_paths <- vapply(
    paths,
    .normalize_array_path,
    character(1),
    USE.NAMES = FALSE
  )

  expect_identical(actual_normalized_paths, expected_normalized_paths)
})

test_that("normalize_array_path works with existing filesystem paths", {
  existing_paths <- c(
    "~/foo/bar/baz.zarr",
    "../foo/bar/baz.zarr",
    "./foo/bar/baz.zarr",
    "foo/bar/baz.zarr",
    "foo/bar/baz.zarr/",
    file.path(tempdir(), "foo", "/bar//", "baz.zarr")
  )

  actual_normalized_existing_paths <- vapply(
    existing_paths,
    .normalize_array_path,
    character(1),
    USE.NAMES = FALSE
  )

  # All normalized paths should match the expected pattern
  expect_match(
    actual_normalized_existing_paths,
    "/([^(.|..)]+/)*foo/bar/baz.zarr/$"
  )
})

test_that("normalize_array_path treats relative paths consistently", {
  expect_identical(
    .normalize_array_path("baz.zarr"),
    .normalize_array_path("./baz.zarr")
  )
})

test_that("normalize_array_path on S3", {
  expect_identical(
    .normalize_array_path(
      "https://noaa-nwm-retro-v2-zarr-pds.s3.amazonaws.com/feature_id"
    ),
    "https://noaa-nwm-retro-v2-zarr-pds.s3.amazonaws.com/feature_id/"
  )
})

#' Define compression tool and settings
#'
#' These functions select a compression tool and its setting when writing a Zarr
#' file
#'
#' @param level Specify the compression level to use.  The range of possible
#' values is dependant on the compression tool being used.  For example, for
#' `use_zlib()` this argument can be between 1 & 9, while for `use_zstd()`the
#' valid range is 0 (default; level chosen automatically) to 22.
#'
#' @returns A list containing the details of the selected compression tool. This
#'   will be written to the .zarray metadata when the Zarr array is created.
#'
#' @examples
#'
#' ## define 2 compression filters for blosc (using snappy) and bzip2 (level 5)
#' blosc_with_snappy_compression <- use_blosc(cname = "snappy")
#' bzip2_compression <- use_bz2(level = 5)
#'
#' ## create an example array to write to a file
#' x <- array(runif(n = 1000, min = -10, max = 10), dim = c(10, 20, 5))
#'
#' ## write the array to two files using each compression filter
#' blosc_path <- tempfile()
#' bzip2_path <- tempfile()
#' write_zarr_array(
#'   x = x, zarr_array_path = blosc_path, chunk_dim = c(2, 5, 1),
#'   compressor = blosc_with_snappy_compression
#' )
#' write_zarr_array(
#'   x = x, zarr_array_path = bzip2_path, chunk_dim = c(2, 5, 1),
#'   compressor = bzip2_compression
#' )
#'
#' ## the contents of the two arrays should be the same
#' identical(read_zarr_array(blosc_path), read_zarr_array(bzip2_path))
#'
#' ## the size of the files on disk are not the same
#' sum(file.size(list.files(blosc_path, full.names = TRUE)))
#' sum(file.size(list.files(bzip2_path, full.names = TRUE)))
#'
#' @name compressors
NULL


#' @rdname compressors
#'
#' @param cname Blosc is a 'meta-compressor' providing access to several
#'   compression algorithms.  This argument defines which compression tool
#'   should be used.  Valid options are: `"lz4"`, `"lz4hc"`, `"blosclz"`,
#'   `"zstd"`, `"zlib"`, `"snappy"`.
#' @param clevel An integer from 0 to 9 which controls the speed and level of
#'   compression. A level of 1 is the fastest compression method and produces
#'   the least compressions, while 9 is slowest and produces the most compression.
#'   Compression is turned off completely when level is 0. Defaults to 5.
#' @param shuffle Specifies the type of shuffling to perform, if any, prior to
#'   compression. Must be one of `"noshuffle"`, to indicate no shuffling;
#'   `"shuffle"` (default), to indicate byte-wise shuffling; `"bitshuffle"`, to
#'   indicate bit-wise shuffling.
#' @param typesize The data type size in bytes used by Blosc shuffling. If
#'   `NULL` (default), this will be inferred from the array datatype. Ignored if
#'   `shuffle = "noshuffle"`.
#' @param blocksize The requested size of the compressed blocks in bytes. Use 0
#'   (default) to let Blosc choose automatically.
#'
#' @export
use_blosc <- function(
  cname = c("lz4", "lz4hc", "blosclz", "zstd", "zlib", "snappy"),
  clevel = 5L,
  shuffle = c("shuffle", "noshuffle", "bitshuffle"),
  typesize = NULL,
  blocksize = 0L
) {
  cname <- tolower(cname)
  cname <- match.arg(cname)

  shuffle <- match.arg(shuffle)

  if (clevel < 0L || clevel > 9L) {
    stop(
      "Blosc compresssion `clevel` value must be an integer between 0 and 9",
      call. = FALSE
    )
  }
  clevel <- as.integer(clevel)

  res <- list(
    id = "blosc",
    cname = cname,
    clevel = as.integer(clevel),
    shuffle = shuffle,
    typesize = typesize,
    blocksize = as.integer(blocksize)
  )
  return(res)
}

#' @rdname compressors
#' @export
use_zlib <- function(level = 6L) {
  res <- list(id = "zlib", level = as.integer(level))
  return(res)
}

#' @rdname compressors
#' @export
use_gzip <- function(level = 6L) {
  res <- list(id = "gzip", level = as.integer(level))
  return(res)
}

#' @rdname compressors
#' @export
use_bz2 <- function(level = 6L) {
  res <- list(id = "bz2", level = as.integer(level))
  return(res)
}

#' @rdname compressors
#' @export
use_lzma <- function(level = 9L) {
  res <- list(id = "lzma", format = 1L, level = as.integer(level))
  return(res)
}

#' @rdname compressors
#' @export
use_lz4 <- function() {
  res <- list(id = "lz4", acceleration = 1L)
  return(res)
}

#' @rdname compressors
#' @export
use_zstd <- function(level = 0L) {
  # Question: we could potentially save level 0 in the metadata as the actual
  # value picked by the zstd library. For now, we follow zarr python approach.
  if (level < 0L || level > 22L) {
    stop(
      "Zstd compression `level` value must be an integer between 0 and 22.",
      call. = FALSE
    )
  }
  res <- list(id = "zstd", level = as.integer(level))
  return(res)
}

test_that("blosc works with variable length types", {
  blosc_vlen_zarr <- withr::local_tempfile(fileext = ".zarr")

  x <- matrix(
    c("Paris", "London", "Berlin", "Madrid"),
    nrow = 2
  )
  # TODO: update this to the canonical way to write vlen-utf8 once we have it.
  # It's this way only because the vlen-utf8 interface has not been created yet.
  create_empty_zarr_array(
    blosc_vlen_zarr,
    dim = dim(x),
    chunk_dim = c(2, 2),
    data_type = "|O",
    compressor = use_blosc()
  )
  zarr_json_path <- file.path(blosc_vlen_zarr, "zarr.json")
  zarr_json <- jsonlite::read_json(zarr_json_path)
  zarr_json$data_type <- "string"
  # There should be only one bytes-array codec
  zarr_json$codecs <- lapply(
    zarr_json$codecs,
    function(codec) {
      if (codec$name == "bytes") {
        list(name = "vlen-utf8", configuration = list())
      } else {
        codec
      }
    }
  )
  jsonlite::write_json(
    zarr_json,
    zarr_json_path,
    auto_unbox = TRUE,
    pretty = TRUE,
    null = "null"
  )

  expect_no_condition(
    update_zarr_array(blosc_vlen_zarr, x, index = list(1:2, 1:2))
  )

  expect_identical(
    read_zarr_array(blosc_vlen_zarr),
    x
  )
})

test_that("error when setting up compressors with invalid level", {
  expect_error(
    use_blosc(clevel = 30),
    "`clevel` value must be an integer between 0 and 9"
  )
  expect_error(
    use_zstd(level = 30),
    "`level` value must be an integer between 0 and 22"
  )
})

test_that("empty zarr arrays can be created and read correctly", {
  path <- withr::local_tempdir(fileext = ".zarr")

  expect_silent(
    res <- create_empty_zarr_array(
      zarr_array_path = path,
      dim = 100,
      chunk_dim = 10,
      data_type = "integer",
      fill_value = 100L
    )
  )

  expect_identical(read_zarr_array(path), array(100L, dim = 100))
})

test_that("zarr arrays can be written and read with various compression methods", {
  x <- array(runif(n = 1000, min = -10, max = 10), dim = c(10, 20, 5))

  # Standard writing and reading with zlib compression
  path <- withr::local_tempfile(pattern = "zlib_", fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = x,
      zarr_array_path = path,
      chunk_dim = c(2, 5, 1),
      order = "F",
      compressor = use_zlib()
    )
  )
  expect_identical(read_zarr_array(path), x)

  # Standard writing and reading with gzip compression
  path <- withr::local_tempfile(pattern = "gzip_", fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = x,
      zarr_array_path = path,
      chunk_dim = c(2, 5, 1),
      order = "F",
      compressor = use_gzip()
    )
  )
  expect_identical(read_zarr_array(path), x)

  # Testing blosc compression
  path <- withr::local_tempfile(pattern = "blosc_", fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = x,
      zarr_array_path = path,
      chunk_dim = c(2, 5, 1),
      compressor = use_blosc(
        cname = "blosclz",
        clevel = 9L,
        shuffle = "bitshuffle",
        typesize = 8L,
        blocksize = 0L
      )
    )
  )
  expect_identical(read_zarr_array(path), x)

  # Testing LZMA compression
  path <- withr::local_tempfile(pattern = "lzma_", fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = x,
      zarr_array_path = path,
      chunk_dim = c(2, 5, 1),
      compressor = use_lzma()
    )
  )
  expect_identical(read_zarr_array(path), x)

  # Testing LZ4 compression
  path <- withr::local_tempfile(pattern = "lz4_", fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = x,
      zarr_array_path = path,
      chunk_dim = c(2, 5, 1),
      compressor = use_lz4()
    )
  )
  expect_identical(read_zarr_array(path), x)

  # Testing BZIP2 compression
  path <- withr::local_tempfile(pattern = "bz2_", fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = x,
      zarr_array_path = path,
      chunk_dim = c(2, 5, 1),
      compressor = use_bz2()
    )
  )
  expect_identical(read_zarr_array(path), x)

  # Testing ZSTD compression
  path <- withr::local_tempfile(pattern = "zstd_", fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = x,
      zarr_array_path = path,
      chunk_dim = c(2, 5, 1),
      compressor = use_zstd()
    )
  )
  expect_identical(read_zarr_array(path), x)

  # Testing no compression
  path <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = x,
      zarr_array_path = path,
      chunk_dim = c(2, 5, 1),
      compressor = NULL
    )
  )
  expect_identical(read_zarr_array(path), x)
})

test_that("zarr arrays work with non-aligned chunk dimensions", {
  x <- array(runif(n = 1000, min = -10, max = 10), dim = c(10, 20, 5))

  # Testing chunk dimensions that don't align perfectly with the array extent
  path <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = x,
      zarr_array_path = path,
      chunk_dim = c(6, 11, 3)
    )
  )
  expect_identical(read_zarr_array(path), x)
})

test_that("zarr arrays work with row-major ordering", {
  x2d <- array(1:8, dim = c(2, 4))
  path2d <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    write_zarr_array(
      x = x2d,
      zarr_array_path = path2d,
      chunk_dim = c(1, 2),
      order = "C"
    )
  )
  expect_identical(read_zarr_array(path2d), x2d)

  x3d <- array(seq_len(300), dim = c(6, 10, 5))

  path3d <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    write_zarr_array(
      x = x3d,
      zarr_array_path = path3d,
      chunk_dim = c(2, 5, 1),
      order = "C"
    )
  )
  expect_identical(read_zarr_array(path3d), x3d)
})

test_that("dimension names roundtrip", {
  x <- array(1:32, dim = c(2, 4, 4))
  dimnames(x) <- list("X" = NULL, "Y" = NULL, "Z" = NULL)

  path <- withr::local_tempfile(fileext = ".zarr")

  expect_silent(
    write_zarr_array(
      x = x,
      zarr_array_path = path,
      chunk_dim = c(2, 4, 1)
    )
  )
  res <- read_zarr_array(path)
  expect_identical(res, x)
})

test_that("NAs roundtrip", {
  int_nas <- array(c(1L, NA_integer_, 3L))
  path_int <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    write_zarr_array(
      x = int_nas,
      zarr_array_path = path_int,
      chunk_dim = 3
    )
  )

  expect_identical(
    read_zarr_array(path_int),
    int_nas
  )

  dbl_nas <- array(c(1.5, NA_real_, 3.5))
  path_dbl <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    write_zarr_array(
      x = dbl_nas,
      zarr_array_path = path_dbl,
      chunk_dim = 3
    )
  )

  expect_identical(
    read_zarr_array(path_dbl),
    dbl_nas
  )

  char_nas <- array(c("z", NA_character_, "WORD"))
  path_unicode <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    write_zarr_array(
      x = char_nas,
      zarr_array_path = path_unicode,
      chunk_dim = 3,
      data_type = "<U"
    )
  )

  expect_identical(
    read_zarr_array(path_unicode),
    char_nas
  )

  path_char <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    write_zarr_array(
      x = char_nas,
      zarr_array_path = path_char,
      chunk_dim = 3,
      data_type = "|S"
    )
  )

  expect_identical(
    read_zarr_array(path_char),
    char_nas
  )
})

test_that("use_blosc() accepts all configurable options", {
  compressor <- use_blosc(
    cname = "blosclz",
    clevel = 9L,
    shuffle = "bitshuffle",
    typesize = 8L,
    blocksize = 0L
  )

  expect_identical(compressor$id, "blosc")
  expect_identical(compressor$cname, "blosclz")
  expect_identical(compressor$clevel, 9L)
  expect_identical(compressor$shuffle, "bitshuffle")
  expect_identical(compressor$typesize, 8L)
  expect_identical(compressor$blocksize, 0L)
})

#' Truncate overflow values and warn about bool NA values before writing
#'
#' Applies the two pre-write data checks that are needed identically in both
#' [write_zarr_array()] and [update_zarr_array()]:
#'
#' 1. If the target type is narrower than R's native representation (e.g.
#'    `int8`, `int16`, `float32`), clamp out-of-range values via
#'    `.truncate_overflow()`.
#' 2. If the target type is `bool`, warn when `x` contains `NA` values (which
#'    the Zarr bool type cannot represent).
#'
#' @param x The R array to be written.
#' @param metadata Array metadata list as returned by [.read_array_metadata()].
#'
#' @returns `x`, possibly with overflow values clamped.
#'
#' @keywords internal
.prepare_write_data <- function(x, metadata) {
  same_type_lower_bytesize <- metadata$data_type %in%
    c("int8", "int16", "float32")
  lower_bytesize_type <- storage.mode(x) == "double" &&
    metadata$data_type == "float32"

  if (same_type_lower_bytesize || lower_bytesize_type) {
    x <- .truncate_overflow(x, metadata$datatype$nbytes)
  }

  if (metadata$data_type == "bool" && anyNA(x)) {
    warning(
      "Zarr native 'bool' data type does not support NA values. ",
      "NA values will be converted to FALSE. ",
      "To preserve NA values, use 'uint8' datatype in `write_zarr_array()` and `as.logical()` after reading.",
      call. = FALSE
    )
  }

  return(x)
}

.truncate_overflow <- function(x, nbytes) {
  max_value <- 2L^(nbytes * 8L - 1L) - 1L
  min_value <- -max_value - 1L

  positive_overflow <- x > max_value
  negative_overflow <- x < min_value
  if (any(positive_overflow | negative_overflow)) {
    warning(
      "Some values in 'x' are too large to be represented by the ",
      "specified data type. They will be truncated when written.",
      call. = FALSE
    )
    storage.mode(max_value) <- storage.mode(x)
    storage.mode(min_value) <- storage.mode(x)
    x[positive_overflow] <- max_value
    x[negative_overflow] <- min_value
  }
  return(x)
}

.guess_int_size <- function(x, data_type) {
  if (anyNA(x)) {
    # Lower bitsize types cannot represent NA values.
    return("<i4")
  }
  # There is no situation where we may return i8 or u8 because these large integers cannot exist in R.
  # FIXME: potential optimization, avoid checking for overflow later if we used this function.
  max_value <- max(x, na.rm = TRUE)
  min_value <- min(x, na.rm = TRUE)

  if (min_value < 0L) {
    if (min_value >= -128L && max_value <= 127L) {
      return("|i1")
    } else if (min_value >= -32768L && max_value <= 32767L) {
      return("<i2")
    } else {
      return("<i4")
    }
  }
  if (max_value <= 255L) {
    return("|u1")
  } else if (max_value <= 65535L) {
    return("<u2")
  } else {
    return("<u4")
  }
}

# nolint next: cyclocomp_linter.
.check_datatype <- function(data_type, fill_value = NULL, nchar = NULL) {
  # TODO: Error if fill_value is incompatible with data_type in Zarr v3.
  # V3 spec says:
  # "The permitted values depend on the data type.
  # Fill values for core data types are listed in Permitted fill values."
  if (missing(data_type) && is.null(fill_value)) {
    stop(
      "Data type cannot be determined if both 'data_type' and 'fill_value' arguments are missing."
    )
  }
  if (missing(data_type) && !is.null(fill_value)) {
    ## if we only have a fill value, infer the data type from that
    data_type <- storage.mode(fill_value)
  }

  ## if data type was supplied directly, always use that
  supported_types <- c(
    "|i1",
    "<i2",
    "<i4",
    "<i8",
    "|u1",
    "<u2",
    "<u4",
    "<u8",
    "<f4",
    "<f8",
    "|S",
    "<U",
    "|b1",
    "|O"
  )
  r_base_types <- c("integer", "double", "character", "logical")

  if (data_type %notin% supported_types) {
    if (data_type %notin% r_base_types) {
      stop(
        "Currently only able to write integer, double, character and logical arrays"
      )
    }
    data_type <- switch(
      data_type,
      "integer" = "<i4",
      "double" = "<f8",
      "character" = "|S",
      "logical" = "|b1"
    )
  }

  ## set a default fill value if needed
  fill_value <- fill_value %||%
    switch(
      data_type,
      "|i1" = 0L,
      "<i2" = 0L,
      "<i4" = 0L,
      "<i8" = 0L,
      "|u1" = 0L,
      "<u2" = 0L,
      "<u4" = 0L,
      "<u8" = 0L,
      "<f4" = 0.0,
      "<f8" = 0.0,
      "|S" = "",
      "<U" = "",
      "|O" = "",
      "|b1" = FALSE,
      NULL
    )

  if (data_type %in% c("|S", "<U", ">U")) {
    if (is.null(nchar) || nchar < 1L) {
      stop(
        "The 'nchar' argument must be provided when working with ",
        "character data types and be a positive integer"
      )
    }
    data_type <- paste0(data_type, as.integer(nchar))
  }

  double_types <- c("<f4", "<f8")
  # FIXME: the spec only defines this for floats but surely it makes
  # sense to also apply it to int and uint?
  if (is.na(fill_value) && data_type %in% double_types) {
    float_size <- switch(
      data_type,
      "<f4" = 4L,
      "<f8" = 8L
    )
    # "0xYYYYYYYY", specifying the byte representation of the floating point number as an unsigned integer.
    fill_value <- paste(
      c("0x", writeBin(fill_value, raw(), size = float_size, endian = "swap")),
      collapse = ""
    )
  }

  return(list(data_type = data_type, fill_value = fill_value))
}

#' Create an (empty) Zarr array
#'
#' @param zarr_array_path Character vector of length 1 giving the path to the
#'   new Zarr array.
#' @param dim Dimensions of the new array.  Should be a numeric vector with the
#'   same length as the number of dimensions.
#' @param dimension_names Optional character vector with the same length as
#'   `dim`.
#' @param chunk_dim Dimensions of the array chunks. Should be a numeric vector
#'   with the same length as the `dim` argument.
#' @param data_type Character vector giving the data type of the new array.
#'   Valid options are: "integer", "double", "character", "logical", which are
#'   based on standard R data types. You can also use the analogous Numpy
#'   formats: "|i1", "<i2", "<i4", "<f4", "<f8", "|S", "|b1".
#'   If this argument isn't provided the `fill_value` will be used to determine
#'   the datatype.
#' @param order Define the layout of the bytes within each chunk.  Valid options
#'   are 'column', 'row', 'F' & 'C'.  'column' or 'F' will specify
#'   "column-major" ordering, which is how R arrays are arranged in memory.
#'   'row' or 'C' will specify "row-major" order.
#' @param compressor What (if any) compression tool should be applied to the
#'   array chunks.  The default is to use `zstd` compression. Supplying `NULL`
#'   will disable chunk compression. See [compressors] for more details.
#' @param fill_value The default value for uninitialized portions of the array.
#'   Does not have to be provided, in which case the default for the specified
#'   data type will be used.
#' @param nchar For `datatype = "character"` this parameter gives the maximum
#'   length of the stored strings. It is an error not to specify this for a
#'   character array, but it is ignored for other data types.
#' @param dimension_separator The character used to to separate the dimensions
#'   in the names of the chunk files.  Valid options are limited to "." and "/".
#' @param zarr_version The version of the Zarr specification to use. Currently,
#'   either `2` or `3`. The default is `3`.
#'
#' @returns This function is primarily called for the side effect of
#'   initialising a Zarr array location and creating the `.zarray` or
#'   `zarr.json` metadata file.
#'   Returns (invisibly) the normalized path it wrote the metadata to.
#'
#' @seealso [write_zarr_array()], [update_zarr_array()]
#'
#' @examples
#'
#' new_zarr_array <- file.path(tempdir(), "temp.zarr")
#' create_empty_zarr_array(new_zarr_array,
#'   dim = c(10, 20), chunk_dim = c(2, 5),
#'   data_type = "integer"
#' )
#'
#' @export
create_empty_zarr_array <- function(
  zarr_array_path,
  dim,
  chunk_dim,
  data_type,
  order = c("F", "C"),
  compressor = use_zstd(),
  fill_value = NULL,
  nchar = NULL,
  dimension_separator = if (zarr_version == 2L) "." else "/",
  dimension_names = NULL,
  zarr_version = 3L
) {
  if (!is.null(dimension_names) && length(dimension_names) != length(dim)) {
    stop(
      "`dimension_names` must have the same length as `dim`.",
      call. = FALSE
    )
  }
  order <- match.arg(order)

  dt <- .check_datatype(
    data_type = data_type,
    fill_value = fill_value,
    nchar = nchar
  )
  data_type <- dt$data_type
  fill_value <- dt$fill_value

  .check_chunk_shape(x_dim = dim, chunk_dim = chunk_dim)

  path <- .normalize_array_path(zarr_array_path)
  if (!dir.exists(path)) {
    dir.create(path)
  }

  ## create the metadata file
  .write_zarr_metadata(
    array_path = path,
    array_shape = dim,
    dimension_names = dimension_names,
    chunk_shape = chunk_dim,
    data_type = data_type,
    order = order,
    fill_value = fill_value,
    compressor = compressor,
    dimension_separator = dimension_separator,
    zarr_version = zarr_version
  )

  return(invisible(path))
}

#' Write an R array to Zarr
#'
#' @param x The R array that will be written to the Zarr array.
#' @param zarr_array_path Character vector of length 1 giving the path to the
#'   new Zarr array.
#' @param nchar For character arrays this parameter gives the maximum length of
#'   the stored strings. If this argument is not specified the array provided to
#'   `x` will be checked and the length of the longest string found will be used
#'   so no data are truncated. However this may be slow and providing a value to
#'   `nchar` can provide a modest performance improvement.
#' @inheritParams create_empty_zarr_array
#'
#' @note If `x` has `dimnames`, `names(dimnames(x))` will be stored as the
#' `dimension_names` field in the Zarr metadata.
#'
#' @returns The function is primarily called for the side effect of writing to
#'   disk. Returns (invisibly) `TRUE` if the array is successfully written.
#'
#' @examples
#'
#' new_zarr_array <- file.path(tempdir(), "integer.zarr")
#' x <- array(1:50, dim = c(10, 5))
#' write_zarr_array(
#'   x = x, zarr_array_path = new_zarr_array,
#'   chunk_dim = c(2, 5)
#' )
#'
#' @export
write_zarr_array <- function(
  x,
  zarr_array_path,
  chunk_dim,
  data_type = storage.mode(x),
  order = c("F", "C"),
  compressor = use_zstd(),
  fill_value = NULL,
  nchar,
  dimension_separator = if (zarr_version == 2L) "." else "/",
  zarr_version = 3L
) {
  if (!is.array(x) && !is.atomic(x)) {
    stop(
      "`x` must be an atomic array. ",
      "You can maybe coerce your object with `as.array()`.",
      call. = FALSE
    )
  }

  if (storage.mode(x) == "character" && missing(nchar)) {
    # +1 to add NUL terminator
    # c(0, ) to deal with array full of NAs
    # base::nchar() to avoid collision with var name
    nchar <- max(c(0L, base::nchar(x)), na.rm = TRUE) + 1L
  }
  if (data_type == "integer") {
    data_type <- .guess_int_size(x, data_type)
  }

  path <- create_empty_zarr_array(
    zarr_array_path = zarr_array_path,
    dim = dim(x),
    dimension_names = names(dimnames(x)),
    chunk_dim = chunk_dim,
    data_type = data_type,
    order = order,
    fill_value = fill_value,
    compressor = compressor,
    nchar = nchar,
    dimension_separator = dimension_separator,
    zarr_version = zarr_version
  )
  # FIXME: it's not optimal to write and read again because we do multiple
  # steps to get it ready for writing, and then ready to use internally.
  # Related to https://github.com/Huber-group-EMBL/Rarr/issues/60.
  metadata <- .read_array_metadata(path)

  metadata$configured_encoders <- .configure_codecs(
    codecs = metadata$codecs,
    operation = "encode"
  )

  x <- .prepare_write_data(x, metadata)

  ## build index covering the entire array
  index <- lapply(dim(x), seq_len)

  ## precompute, for each chunk, the positions in `index` that belong to it
  chunk_positions <- .chunk_positions_by_chunk(index, metadata, chunk_dim)
  chunk_names <- names(chunk_positions)
  chunk_paths <- paste0(path, chunk_names)

  ## iterate over each chunk
  ## TODO: maybe this can be done in parallel with bpmapply() ?
  res <- mapply(
    FUN = .write_chunk,
    chunk_paths,
    chunk_names,
    MoreArgs = list(
      x = x,
      chunk_positions = chunk_positions,
      metadata = metadata
    )
  )

  return(invisible(all(res)))
}

.write_chunk <- function(chunk_path, chunk_name, x, chunk_positions, metadata) {
  chunk_dim <- unlist(metadata$chunk_grid$configuration$chunk_shape)

  chunk_info <- chunk_positions[[chunk_name]]
  idx_in_array <- chunk_info$positions
  idx_in_chunk <- chunk_info$index_in_chunk

  chunk_in_mem <- .extract_chunk(x, idx_in_array)

  # FIXME: can this check be faster?
  # isTRUE() because metadata$fill_value can be NA.
  if (isTRUE(all(chunk_in_mem == metadata$fill_value))) {
    ## if the chunk only contains the fill value, we can skip writing it
    return(invisible(TRUE))
  }

  # Spec says:
  # "Chunks at the border of an array always have the full chunk size,
  # even when the array only covers parts of it."
  if (any(dim(chunk_in_mem) != chunk_dim)) {
    ## create a new "complete" chunk filled with the fill value, then insert
    ## our partial chunk into it
    temp_chunk <- array(metadata$fill_value, dim = chunk_dim)
    rlang::inject(temp_chunk[!!!idx_in_chunk] <- chunk_in_mem) # nolint: implicit_assignment_linter.
    chunk_in_mem <- temp_chunk
  }

  .compress_and_write_chunk(
    input_chunk = chunk_in_mem,
    chunk_path = chunk_path,
    metadata
  )

  return(invisible(TRUE))
}

#' Update (a subset of) an existing Zarr array
#'
#' @param zarr_array_path Character vector of length 1 giving the path to the
#'   Zarr array that is to be modified.
#' @param x The R array (or object that can be coerced to an array) that will be
#'   written to the Zarr array.
#' @param index A list with the same length as the number of dimensions of the
#'   target array. This argument indicates which elements in the target array
#'   should be updated.
#'
#' @returns The function is primarily called for the side effect of writing to
#'   disk. Returns (invisibly) `TRUE` if the array is successfully updated.
#'
#' @examples
#'
#' ## first create a new, empty, Zarr array
#' new_zarry_array <- file.path(tempdir(), "new_array.zarr")
#' create_empty_zarr_array(
#'   zarr_array_path = new_zarry_array, dim = c(20, 10),
#'   chunk_dim = c(10, 5), data_type = "double"
#' )
#'
#' ## create a matrix smaller than our Zarr array
#' small_matrix <- matrix(runif(6), nrow = 3)
#'
#' ## insert the matrix into the first 3 rows, 2 columns of the Zarr array
#' update_zarr_array(new_zarry_array, x = small_matrix, index = list(1:3, 1:2))
#'
#' ## reading back a slightly larger subset,
#' ## we can see only the top left corner has been changed
#' read_zarr_array(new_zarry_array, index = list(1:5, 1:5))
#'
#' @export
update_zarr_array <- function(zarr_array_path, x, index) {
  stopifnot(is.list(index))

  zarr_array_path <- .normalize_array_path(zarr_array_path)

  metadata <- .read_array_metadata(zarr_array_path)

  index <- check_index(index, metadata = metadata)

  existing_storage <- switch(
    metadata$datatype$base_type,
    "uint" = "integer",
    "int" = "integer",
    "float" = "double",
    "bool" = "logical",
    "string" = "character",
    "unicode" = "character",
    NULL
  )
  if (!identical(storage.mode(x), existing_storage)) {
    stop("New data is not of the same type as the existing array.")
  }

  x <- .prepare_write_data(x, metadata)

  metadata$configured_encoders <- .configure_codecs(
    codecs = metadata$codecs,
    operation = "encode"
  )
  metadata$configured_decoders <- .configure_codecs(
    codecs = metadata$codecs,
    operation = "decode"
  )

  ## coerce x to the same shape as the zarr to be updated
  x <- array(x, dim = lengths(index))

  ## precompute, for each chunk, the positions in `index` that belong to it
  chunk_dim <- unlist(metadata$chunk_grid$configuration$chunk_shape)
  chunk_positions <- .chunk_positions_by_chunk(
    index,
    metadata,
    chunk_dim
  )
  chunk_names <- names(chunk_positions)

  ## only update the chunks that need to be
  ## TODO: maybe this can be done in parallel is bpmapply() ?
  res <- mapply(
    FUN = .update_chunk,
    chunk_names,
    MoreArgs = list(
      x = x,
      chunk_positions = chunk_positions,
      zarr_array_path = zarr_array_path,
      chunk_dim = chunk_dim,
      metadata = metadata
    )
  )

  return(invisible(all(res)))
}

.update_chunk <- function(
  chunk_name,
  x,
  zarr_array_path,
  chunk_positions,
  chunk_dim,
  metadata
) {
  ## determine which elements of x are being used and where in this specific
  ## chunk they should be inserted
  chunk_path <- file.path(zarr_array_path, chunk_name)
  chunk_info <- chunk_positions[[chunk_name]]
  idx_in_x <- chunk_info$positions
  idx_in_chunk <- chunk_info$index_in_chunk

  if (.store_check_exist(zarr_array_path, chunk_name, s3_client = NULL)) {
    size <- file.info(chunk_path)$size
    raw_chunk <- readBin(con = chunk_path, what = "raw", n = size)
    chunk_in_mem <- read_chunk(
      chunk_bytes = raw_chunk,
      chunk_dim = chunk_dim,
      decoders = metadata$configured_decoders,
      datatype = metadata$datatype,
      fill_value = metadata$fill_value
    )
  } else {
    chunk_in_mem <- array(
      metadata$fill_value,
      dim = chunk_dim
    )
  }

  ## extract the new values from x and insert them into the chunk
  rlang::inject(chunk_in_mem[!!!idx_in_chunk] <- x[!!!idx_in_x]) # nolint: implicit_assignment_linter.
  ## re-compress updated chunk and write back to disk
  .compress_and_write_chunk(
    input_chunk = chunk_in_mem,
    chunk_path = chunk_path,
    metadata
  )
}

#' Compress and write a single chunk
#'
#' @param input_chunk Array containing the chunk data to be compressed.  Will be
#'   converted to a raw vector before compression.
#' @param chunk_path Character string giving the path to the chunk that should
#'   be written.
#'
#' @returns Returns `TRUE` if writing is successful.  Mostly called for the
#'   side-effect of writing the compressed chunk to disk.
#'
#' @keywords internal
.compress_and_write_chunk <- function(
  input_chunk,
  chunk_path,
  metadata
) {
  ## check the chunk path exists, and create if not
  if (!dir.exists(dirname(chunk_path))) {
    dir.create(dirname(chunk_path), recursive = TRUE, showWarnings = FALSE)
  }

  # Array to array codecs
  for (codec in metadata$configured_encoders[["array_array"]]) {
    input_chunk <- do.call(codec, list(input_chunk))
  }
  # Array to bytes codecs
  for (codec in metadata$configured_encoders[["array_bytes"]]) {
    raw_chunk <- codec(as.vector(input_chunk), metadata$datatype)
  }

  # Bytes to bytes codecs
  codecs <- metadata$codecs
  compressor <- NULL
  compressor$id <- names(codecs)[match(
    TRUE,
    names(codecs) %in% CODEC_BYTES_BYTES
  )]
  compressor_config <- codecs[[compressor$id]]$configuration
  if (is.na(compressor$id)) {
    compressed_chunk <- raw_chunk
  } else if (compressor$id == "blosc") {
    compressed_chunk <- codec_blosc_encode(raw_chunk, compressor_config)
  } else if (compressor$id == "zlib") {
    compressed_chunk <- memCompress(from = raw_chunk, type = "gzip")
  } else if (compressor$id == "gzip") {
    con <- gzfile(
      chunk_path,
      open = "wb",
      compression = compressor_config$level
    )
    on.exit(close(con))
  } else if (compressor$id == "bz2") {
    con <- bzfile(
      chunk_path,
      open = "wb",
      compression = compressor_config$level
    )
    on.exit(close(con))
  } else if (compressor$id == "lzma") {
    con <- xzfile(
      chunk_path,
      open = "wb",
      compression = compressor_config$level
    )
    on.exit(close(con))
  } else if (compressor$id %in% c("lz4", "numcodecs.lz4")) {
    compressed_chunk <- codec_lz4_encode(raw_chunk)
  } else if (compressor$id == "zstd") {
    compressed_chunk <- codec_zstd_encode(raw_chunk, compressor_config)
  }

  if (compressor$id %in% c("gzip", "bz2", "lzma")) {
    writeBin(raw_chunk, con = con, useBytes = TRUE)
  } else {
    writeBin(compressed_chunk, con = chunk_path)
  }

  return(invisible(TRUE))
}

.check_chunk_shape <- function(x_dim, chunk_dim) {
  if (length(x_dim) != length(chunk_dim)) {
    stop("The dimensions of the chunk must equal the dimensions of the array.")
  }

  oversized_chunk <- any(chunk_dim > x_dim)
  if (oversized_chunk) {
    # One valid use case is nullable arrays in anndata.
    warning(
      "Chunk dimensions are larger than array dimensions. ",
      "This is allowed by the Zarr specification but may lead to inefficient storage and retrieval.\n",
      "In most cases, this is likely to be a mistake. Please check your `chunk_dim` argument.",
      call. = FALSE
    )
  }
  return(invisible(!oversized_chunk))
}

test_that("uint32 zarr arrays can be read correctly", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "uint32.zarr",
    package = "Rarr"
  )
  zarr_f <- system.file(
    "extdata",
    "zarr_examples",
    "row-first",
    "uint32.zarr",
    package = "Rarr"
  )
  index <- list(1:30, 1:20, 1)

  ## Shouldn't be any warnings for this input
  expect_silent(column_major <- read_zarr_array(zarr_c, index = index))
  expect_silent(row_major <- read_zarr_array(zarr_f, index = index))

  ## row and column major should be read the same in R
  expect_identical(column_major, row_major)

  ## we return an array
  expect_true(is.array(column_major))

  ## dimensions equal to the index
  expect_identical(dim(column_major), c(30L, 20L, 1L))

  ## first row should be sequence 1 to 20
  expect_identical(column_major[1, , ], 1:20)
  ## first column should be all 1s
  expect_identical(column_major[, 1, ], rep(1L, 30))

  ## this data point should be outside the range of an int32 and throw a warning
  expect_warning(
    column_major <- read_zarr_array(zarr_c, index = list(30, 20, 10))
  )
  expect_warning(row_major <- read_zarr_array(zarr_f, index = list(30, 20, 10)))
  expect_true(all(is.na(column_major)))
  expect_true(all(is.na(row_major)))
})

test_that("v2 and v3 return identical results", {
  zarr_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "uint32.zarr",
    package = "Rarr"
  )
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "uint32_v3.zarr",
    package = "Rarr"
  )

  u32_v2 <- read_zarr_array(zarr_v2)
  u32_v3 <- read_zarr_array(zarr_v3)

  expect_identical(u32_v2, u32_v3)
})

test_that("uint32 zarr array can be written", {
  uint32_zarr <- withr::local_tempfile(fileext = ".zarr")

  content_uint32_array <- array(1:24, dim = c(4, 3, 2))

  write_zarr_array(
    content_uint32_array,
    uint32_zarr,
    data_type = "<u4",
    chunk_dim = c(2, 2, 1),
    compressor = NULL
  )

  roundtrip_uint32_array <- read_zarr_array(uint32_zarr)

  expect_identical(
    content_uint32_array,
    roundtrip_uint32_array
  )
})

test_that("Inf fill-values are understood correctly", {
  zarr <- system.file(
    "extdata",
    "zarr_examples",
    "fill-values",
    "double-inf.zarr",
    package = "Rarr"
  )

  fill_val <- Inf
  expect_silent(data <- read_zarr_array(zarr, index = list(1:2, 1:10)))

  # The first row should all be 1
  expect_identical(data[1, ], rep(1, 10))

  # Second row is Inf
  expect_equal(data[2, ], rep(fill_val, 10))

  # Read a chunk that isn't on disk, only generated from fill value
  expect_silent(data2 <- read_zarr_array(zarr, index = list(19:20, 1:10)))
  expect_equal(as.vector(data2), rep(fill_val, length(data2)))
})

test_that("-Inf fill-values are understood correctly", {
  zarr <- system.file(
    "extdata",
    "zarr_examples",
    "fill-values",
    "double-neginf.zarr",
    package = "Rarr"
  )

  fill_val <- -Inf
  expect_silent(data <- read_zarr_array(zarr, index = list(1:2, 1:10)))

  # The first row should all be 1
  expect_identical(data[1, ], rep(1, 10))

  # Second row is -Inf
  expect_equal(data[2, ], rep(fill_val, 10))

  # Read a chunk that isn't on disk, only generated from fill value
  expect_silent(data2 <- read_zarr_array(zarr, index = list(19:20, 1:10)))
  expect_equal(as.vector(data2), rep(fill_val, length(data2)))
})

test_that("NA fill-values are understood correctly", {
  zarr_na_fill <- withr::local_tempfile(fileext = ".zarr/")

  create_empty_zarr_array(
    zarr_na_fill,
    dim = c(4, 10),
    chunk_dim = c(2, 5),
    data_type = "double",
    fill_value = NA_real_
  )

  expect_true(
    is.na(.read_array_metadata(zarr_na_fill)$fill_value)
  )

  index <- list(2:3, 2:6)
  res <- read_zarr_array(zarr_na_fill, index = index)

  expect_true(
    all(is.na(res))
  )
  expect_type(res, "double")

  update_zarr_array(
    zarr_na_fill,
    matrix(10.5, nrow = 2, ncol = 5),
    index
  )
  res <- read_zarr_array(zarr_na_fill)
  expect_type(res, "double")
  expect_identical(res[1, 1], NA_real_)
  expect_identical(res[2, 2], 10.5)
})

test_that("non-NA values represented as bytes work", {
  zarr <- system.file(
    "extdata",
    "zarr_examples",
    "fill-values",
    "bytes_representation.zarr",
    package = "Rarr"
  )

  expect_silent(data <- read_zarr_array(zarr, index = list(1:2, 1:10)))
  expect_true(all(data == 222))
})

test_that("fill-values on empty slices", {
  # https://github.com/Huber-group-EMBL/Rarr/issues/137
  m <- matrix(runif(30), c(5, 6))
  my_zarr3 <- withr::local_tempfile(fileext = ".zarr")
  write_zarr_array(m, zarr_array_path = my_zarr3, chunk_dim = c(2, 3))

  expect_type(
    read_zarr_array(my_zarr3, list(1:2, integer(0))),
    "double"
  )
})

test_that("int32 zarr arrays can be read correctly", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int32.zarr",
    package = "Rarr"
  )
  zarr_f <- system.file(
    "extdata",
    "zarr_examples",
    "row-first",
    "int32.zarr",
    package = "Rarr"
  )
  index <- list(1:30, 1:20, 1)

  # Shouldn't be any warnings for this input
  expect_silent(column_major <- read_zarr_array(zarr_c, index = index))
  expect_silent(row_major <- read_zarr_array(zarr_f, index = index))

  # Row and column major should be read the same in R
  expect_identical(column_major, row_major)

  # We return an array
  expect_true(is.array(column_major))

  # Dimensions equal to the index
  expect_identical(dim(column_major), c(30L, 20L, 1L))

  # First row should be sequence 1 to 20
  expect_identical(column_major[1, , ], 1:20)

  # First column should be all 1s
  expect_identical(column_major[, 1, ], rep(1L, 30))
})

test_that("i32 zarr array can be written", {
  i32_zarr <- withr::local_tempfile(fileext = ".zarr")

  content_i32_array <- array(1:24, dim = c(4, 3, 2))

  write_zarr_array(
    content_i32_array,
    i32_zarr,
    data_type = "<i4",
    chunk_dim = c(2, 2, 1),
    compressor = NULL
  )

  roundtrip_i32_array <- read_zarr_array(i32_zarr)

  expect_identical(
    content_i32_array,
    roundtrip_i32_array
  )
})

test_that("v2 and v3 return identical results", {
  zarr_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int32.zarr",
    package = "Rarr"
  )
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int32_v3.zarr",
    package = "Rarr"
  )

  expect_no_condition(i32_v2 <- read_zarr_array(zarr_v2))
  expect_no_condition(i32_v3 <- read_zarr_array(zarr_v3))

  expect_identical(i32_v2, i32_v3)
})

test_that("boolean zarr array can be read correctly", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "boolean.zarr",
    package = "Rarr"
  )
  index <- list(1:5, NULL)

  # Shouldn't be any warnings for this input
  expect_silent(column_major <- read_zarr_array(zarr_c, index = index))

  # we return an array
  expect_true(is.array(column_major))

  # dimensions equal to the index
  expect_identical(dim(column_major), c(5L, 10L))

  # first row should be TRUE
  # second row should be FALSE
  expect_true(all(column_major[1, ]))
  expect_false(any(column_major[2, ]))
})

test_that("boolean zarr array can be written", {
  bool_zarr <- withr::local_tempfile(fileext = ".zarr")

  content_bool_array <- matrix(c(TRUE, FALSE, TRUE, FALSE), nrow = 2)
  write_zarr_array(
    content_bool_array,
    bool_zarr,
    chunk_dim = c(2, 2),
    compressor = NULL
  )

  roundtrip_bool_array <- read_zarr_array(bool_zarr)

  expect_identical(
    content_bool_array,
    roundtrip_bool_array
  )

  content_bool_array_na <- matrix(c(TRUE, NA, TRUE, FALSE), nrow = 2)
  write_zarr_array(
    content_bool_array_na,
    bool_zarr,
    chunk_dim = c(2, 2),
    compressor = NULL
  ) |>
    expect_warning("not support NA values")

  read_zarr_array(bool_zarr) |>
    expect_identical(
      content_bool_array
    )
})

test_that("v2 and v3 return identical results", {
  zarr_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "boolean.zarr",
    package = "Rarr"
  )
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "boolean_v3.zarr",
    package = "Rarr"
  )

  expect_no_condition(bool_v2 <- read_zarr_array(zarr_v2))
  expect_no_condition(bool_v3 <- read_zarr_array(zarr_v3))

  expect_identical(bool_v2, bool_v3)
})

test_that("int64 zarr arrays can be read correctly", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int64.zarr",
    package = "Rarr"
  )
  zarr_f <- system.file(
    "extdata",
    "zarr_examples",
    "row-first",
    "int64.zarr",
    package = "Rarr"
  )
  index <- list(1:30, 1:20, 1)

  # Shouldn't be any warnings for this input
  expect_silent(column_major <- read_zarr_array(zarr_c, index = index))
  expect_silent(row_major <- read_zarr_array(zarr_f, index = index))

  # Row and column major should be read the same in R
  expect_identical(column_major, row_major)

  # We return an array
  expect_true(is.array(column_major))

  # Dimensions equal to the index
  expect_identical(dim(column_major), c(30L, 20L, 1L))

  # First row should be sequence 1 to 20
  expect_identical(column_major[1, , ], 1:20)

  # First column should be all 1s
  expect_identical(column_major[, 1, ], rep(1L, 30))
})

test_that("int64 zarr arrays handle out-of-range values", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int64.zarr",
    package = "Rarr"
  )
  zarr_f <- system.file(
    "extdata",
    "zarr_examples",
    "row-first",
    "int64.zarr",
    package = "Rarr"
  )

  # This data point should be outside the range of an int32 and throw a warning
  expect_warning(
    column_major <- read_zarr_array(zarr_c, index = list(30, 20, 10))
  )
  expect_warning(row_major <- read_zarr_array(zarr_f, index = list(30, 20, 10)))
  expect_true(all(is.na(column_major)))
  expect_true(all(is.na(row_major)))

  # This data point should be smaller than the range of an int32 and throw a warning
  expect_warning(
    column_major <- read_zarr_array(zarr_c, index = list(30, 20, 9))
  )
  expect_warning(row_major <- read_zarr_array(zarr_f, index = list(30, 20, 9)))
  expect_true(all(is.na(column_major)))
  expect_true(all(is.na(row_major)))
})

test_that("v2 and v3 return identical results", {
  zarr_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int64.zarr",
    package = "Rarr"
  )
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int64_v3.zarr",
    package = "Rarr"
  )

  i64_v2 <- read_zarr_array(zarr_v2)
  i64_v3 <- read_zarr_array(zarr_v3)

  expect_identical(i64_v2, i64_v3)
})

test_that("int64 zarr array can be written", {
  int64_zarr <- withr::local_tempfile(fileext = ".zarr")

  content_int64_array <- array(1:24, dim = c(4, 3, 2))

  write_zarr_array(
    content_int64_array,
    int64_zarr,
    data_type = "<i8",
    chunk_dim = c(2, 2, 1),
    compressor = NULL
  )

  roundtrip_int64_array <- read_zarr_array(int64_zarr)

  expect_identical(
    content_int64_array,
    roundtrip_int64_array
  )
})

test_that("An empty zarr array with dimension 0 can be created", {
  # This is used in anndataR for example
  empty_zarr <- withr::local_tempfile(fileext = ".zarr")
  expect_no_condition(
    Rarr::create_empty_zarr_array(
      empty_zarr,
      dim = 0,
      chunk_dim = 0,
      data_type = "integer",
      zarr_version = 2
    )
  )
  empty_metadata <- Rarr::zarr_overview(empty_zarr, as_data_frame = TRUE)
  expect_identical(empty_metadata$dim, list(0L))
  expect_identical(empty_metadata$chunk_dim, list(0L))

  empty_zarr_v3 <- withr::local_tempfile(fileext = ".zarr")
  expect_no_condition(
    Rarr::create_empty_zarr_array(
      empty_zarr_v3,
      dim = 0,
      chunk_dim = 0,
      data_type = "integer",
      zarr_version = 3
    )
  )
  empty_metadata_v3 <- Rarr::zarr_overview(empty_zarr_v3, as_data_frame = TRUE)
  expect_identical(empty_metadata_v3$dim, list(0L))
  expect_identical(empty_metadata_v3$chunk_dim, list(0L))
})

test_that("Empty zarr arrays have expected metadata", {
  empty_zarr <- withr::local_tempfile(fileext = ".zarr")
  Rarr::create_empty_zarr_array(
    empty_zarr,
    dim = c(10, 30),
    chunk_dim = c(5, 3),
    data_type = "integer",
    compressor = use_blosc(
      cname = "blosclz",
      clevel = 9L,
      shuffle = "bitshuffle",
      blocksize = 0L
    ),
    zarr_version = 2L
  )
  expect_snapshot(
    jsonlite::read_json(file.path(empty_zarr, ".zarray"))
  )
})

test_that("Creating an empty zarr array via `write_zarr_array()`", {
  empty_zarr <- withr::local_tempfile(fileext = ".zarr")
  expect_no_condition(
    write_zarr_array(
      array(5L, dim = c(20, 20, 20)),
      empty_zarr,
      chunk_dim = c(2, 5, 5),
      fill_value = 5L
    )
  )
  expect_identical(
    list.files(empty_zarr, recursive = TRUE),
    "zarr.json"
  )
})

test_that("float32 zarr arrays can be read correctly", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "float32.zarr",
    package = "Rarr"
  )
  zarr_f <- system.file(
    "extdata",
    "zarr_examples",
    "row-first",
    "float32.zarr",
    package = "Rarr"
  )
  index <- list(1:30, 1:20, 1)

  # Shouldn't be any warnings for this input
  expect_silent(column_major <- read_zarr_array(zarr_c, index = index))
  expect_silent(row_major <- read_zarr_array(zarr_f, index = index))

  # Row and column major should be read the same in R
  expect_identical(column_major, row_major)

  # We return an array
  expect_true(is.array(column_major))

  # Dimensions equal to the index
  expect_identical(dim(column_major), lengths(index))

  # First row should be 2 to 20 except the first element which is 10.52
  expect_identical(column_major[1, , ], c(10.52, 2:20), tolerance = 1e-7)

  # First column should be all 10.52
  expect_identical(column_major[, 1, ], rep(10.52, 30), tolerance = 1e-7)
})

test_that("float32 zarr arrays can be written", {
  f32_zarr <- withr::local_tempfile(fileext = ".zarr")

  content_f32_array <- array(runif(24), dim = c(4, 3, 2))

  write_zarr_array(
    content_f32_array,
    f32_zarr,
    data_type = "<f4",
    chunk_dim = c(2, 2, 1),
    compressor = NULL
  )

  roundtrip_f32_array <- read_zarr_array(f32_zarr)

  # Conversion to float32 leads to loss of precision
  expect_equal(
    content_f32_array,
    roundtrip_f32_array,
    tolerance = 1e-7
  )
})

test_that("v2 and v3 return identical results", {
  zarr_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "float32.zarr",
    package = "Rarr"
  )
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "float32_v3.zarr",
    package = "Rarr"
  )

  expect_no_condition(f32_v2 <- read_zarr_array(zarr_v2))
  expect_no_condition(f32_v3 <- read_zarr_array(zarr_v3))

  expect_identical(f32_v2, f32_v3)
})

test_that("update_zarr_array replaces chunks for 1D arrays", {
  path <- withr::local_tempfile(fileext = ".zarr")
  res <- create_empty_zarr_array(
    zarr_array_path = path,
    dim = 100,
    chunk_dim = 10,
    data_type = "integer",
    fill_value = 100L,
    zarr_version = 2
  )

  x <- array(1:5, dim = 5)
  expect_true(update_zarr_array(path, x = x, index = list(1:5)))
  expect_identical(
    read_zarr_array(path),
    array(c(1:5, rep(100L, 95)), dim = 100)
  )
  # only a single chunk file should have been created
  expect_identical(list.files(path), "0")

  x <- rep(20L, 5)
  expect_true(update_zarr_array(
    path,
    x = x,
    index = list(c(91, 93, 95, 97, 99))
  ))
  expect_identical(
    read_zarr_array(path)[91:100],
    array(rep(c(20L, 100L), 5), dim = 10)
  )
  # expect two chunk files were created
  expect_length(list.files(path), 2)

  path_v3 <- withr::local_tempfile(fileext = ".zarr")
  res_v3 <- create_empty_zarr_array(
    zarr_array_path = path_v3,
    dim = 100,
    chunk_dim = 10,
    data_type = "integer",
    fill_value = 100L,
    zarr_version = 3
  )

  x <- array(1:5, dim = 5)
  expect_true(update_zarr_array(path_v3, x = x, index = list(1:5)))
  expect_identical(
    read_zarr_array(path_v3),
    array(c(1:5, rep(100L, 95)), dim = 100)
  )
  # only a single chunk file should have been created
  expect_identical(list.files(path_v3, recursive = TRUE), c("c/0", "zarr.json"))

  x <- rep(20L, 5)
  expect_true(update_zarr_array(
    path_v3,
    x = x,
    index = list(c(91, 93, 95, 97, 99))
  ))
  expect_identical(
    read_zarr_array(path_v3)[91:100],
    array(rep(c(20L, 100L), 5), dim = 10)
  )
  # expect two chunk files were created
  expect_identical(
    list.files(path_v3, recursive = TRUE),
    c("c/0", "c/9", "zarr.json")
  )
})

test_that("update_zarr_array handles NULL in index for some dimensions (2D case)", {
  path <- withr::local_tempfile(fileext = ".zarr")
  res <- create_empty_zarr_array(
    zarr_array_path = path,
    dim = c(20, 20),
    chunk_dim = c(10, 10),
    data_type = "integer",
    fill_value = 0L
  )
  x <- matrix(1:40L, ncol = 2)
  index <- list(NULL, 1:2)

  expect_true(update_zarr_array(path, x = x, index = index))
  expect_identical(read_zarr_array(path, index = index), x)
})

test_that("update_zarr_array on C-order arrays", {
  path <- withr::local_tempfile(fileext = ".zarr")
  res <- create_empty_zarr_array(
    zarr_array_path = path,
    dim = c(20, 20),
    chunk_dim = c(10, 10),
    data_type = "integer",
    fill_value = 0L,
    order = "C",
    compressor = NULL
  )

  x <- matrix(1:400, ncol = 20)

  expect_true(update_zarr_array(path, x, index = list(NULL, NULL)))
  expect_identical(read_zarr_array(path), x)
})

test_that("update higher precision from lower precision internal type", {
  path <- withr::local_tempfile(fileext = ".zarr")
  res <- create_empty_zarr_array(
    zarr_array_path = path,
    dim = c(10, 10),
    chunk_dim = c(5, 5),
    data_type = "<i8",
    fill_value = 0L
  )

  x <- matrix(1:100, ncol = 10)

  expect_true(update_zarr_array(path, x, index = list(NULL, NULL)))
  expect_identical(read_zarr_array(path), x)
})


test_that("update unsigned array", {
  path <- withr::local_tempfile(fileext = ".zarr")
  res <- create_empty_zarr_array(
    zarr_array_path = path,
    dim = c(10, 10),
    chunk_dim = c(5, 5),
    data_type = "<u4",
    fill_value = 0L
  )

  x <- matrix(1:100, ncol = 10)

  expect_true(update_zarr_array(path, x, index = list(NULL, NULL)))
  expect_identical(read_zarr_array(path), x)
})

test_that("update VLen-UTF8 array", {
  path <- withr::local_tempfile(fileext = ".zarr")
  dir.create(path)
  list(
    shape = list(12L, 12L),
    chunks = list(6L, 6L),
    dtype = "|O",
    fill_value = "",
    order = "C",
    filters = list(list(id = "vlen-utf8")),
    dimension_separator = ".",
    compressor = list(id = "zstd", level = 0L),
    zarr_format = 2L
  ) |>
    jsonlite::write_json(
      file.path(path, ".zarray"),
      auto_unbox = TRUE,
      pretty = TRUE
    )

  x <- rep_len("ça marche", 12L)

  expect_true(update_zarr_array(path, x, index = list(1, NULL)))
  res <- read_zarr_array(path)
  expect_identical(res[1, ], x)
  expect_identical(res[2:12, ], array("", dim = c(11, 12)))
})

test_that("update v3 array", {
  path <- withr::local_tempfile(fileext = ".zarr")
  dir.create(path)
  list(
    shape = list(30L, 20L, 10L),
    data_type = "int32",
    chunk_grid = list(
      name = "regular",
      configuration = list(chunk_shape = list(10L, 10L, 5L))
    ),
    chunk_key_encoding = list(
      name = "default",
      configuration = list(separator = "/")
    ),
    fill_value = 0L,
    codecs = list(
      list(name = "bytes", configuration = list(endian = "little")),
      list(name = "zstd", configuration = list(level = 0L, checksum = FALSE))
    ),
    zarr_format = 3L,
    node_type = "array"
  ) |>
    jsonlite::write_json(
      file.path(path, "zarr.json"),
      auto_unbox = TRUE,
      pretty = TRUE
    )

  x <- 1:30

  expect_true(update_zarr_array(path, x, index = list(NULL, 1, 1)))
  res <- read_zarr_array(path)
  expect_identical(res[, 1, 1], x)
  expect_identical(res[, 1:20, 2:10], array(0L, dim = c(30, 20, 9)))
})

test_that("update existing chunk", {
  # https://github.com/Huber-group-EMBL/Rarr/issues/213
  zarr_path <- tempfile(fileext = ".zarr")
  create_empty_zarr_array(
    zarr_path,
    dim = c(8, 10),
    chunk_dim = c(4, 5),
    data_type = "integer"
  )
  data <- matrix(101:106, nrow = 2)
  update_zarr_array(zarr_path, data, index = list(1:2, 1:3))
  expect_no_condition(
    update_zarr_array(zarr_path, data, index = list(3:4, 1:3))
  )
  expect_snapshot(
    read_zarr_array(zarr_path)
  )
})

test_that("int8 zarr arrays can be read correctly", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int8.zarr",
    package = "Rarr"
  )
  zarr_f <- system.file(
    "extdata",
    "zarr_examples",
    "row-first",
    "int8.zarr",
    package = "Rarr"
  )
  index <- list(1:30, 1:20, 1)

  # Shouldn't be any warnings for this input
  expect_silent(column_major <- read_zarr_array(zarr_c, index = index))
  expect_silent(row_major <- read_zarr_array(zarr_f, index = index))

  # Row and column major should be read the same in R
  expect_identical(column_major, row_major)

  # We return an array
  expect_true(is.array(column_major))

  # Dimensions equal to the index
  expect_identical(dim(column_major), c(30L, 20L, 1L))

  # First row should be sequence 1 to 20
  expect_identical(column_major[1, , ], c(-1L, 2:20))

  # First column should be all 1s
  expect_identical(column_major[, 1, ], rep(-1L, 30))
})

test_that("i8 zarr array can be written", {
  i8_zarr <- withr::local_tempfile(fileext = ".zarr")

  content_i8_array <- array(1:24, dim = c(4, 3, 2))

  write_zarr_array(
    content_i8_array,
    i8_zarr,
    data_type = "|i1",
    chunk_dim = c(2, 2, 1),
    compressor = NULL
  )

  roundtrip_i8_array <- read_zarr_array(i8_zarr)

  expect_identical(
    content_i8_array,
    roundtrip_i8_array
  )
})

test_that("overflow when writing cause warning", {
  overflowing_zarr <- withr::local_tempfile(fileext = ".zarr")

  content_overflowing_array <- array(128L, c(3, 3, 3))

  expect_warning(
    write_zarr_array(
      content_overflowing_array,
      overflowing_zarr,
      data_type = "|i1",
      chunk_dim = c(3, 3, 3),
      compressor = NULL
    ),
    "truncate"
  )

  expect_identical(
    read_zarr_array(overflowing_zarr),
    array(127L, c(3, 3, 3))
  )

  overflowing_zarr2 <- withr::local_tempfile(fileext = ".zarr")
  content_overflowing_array2 <- array(1000L, c(3, 3, 3))

  expect_warning(
    write_zarr_array(
      content_overflowing_array2,
      overflowing_zarr2,
      data_type = "|i1",
      chunk_dim = c(3, 3, 3),
      compressor = NULL
    ),
    "truncate"
  )

  expect_identical(
    read_zarr_array(overflowing_zarr2),
    array(127L, c(3, 3, 3))
  )

  negative_overflowing_zarr <- withr::local_tempfile(fileext = ".zarr")
  content_negative_overflowing_array <- array(-1000L, c(3, 3, 3))

  expect_warning(
    write_zarr_array(
      content_negative_overflowing_array,
      negative_overflowing_zarr,
      data_type = "|i1",
      chunk_dim = c(3, 3, 3),
      compressor = NULL
    ),
    "truncate"
  )

  expect_identical(
    read_zarr_array(negative_overflowing_zarr),
    array(-128L, c(3, 3, 3))
  )
})

test_that("v2 and v3 return identical results", {
  zarr_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int8.zarr",
    package = "Rarr"
  )
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "int8_v3.zarr",
    package = "Rarr"
  )

  expect_no_condition(i8_v2 <- read_zarr_array(zarr_v2))
  expect_no_condition(i8_v3 <- read_zarr_array(zarr_v3))

  expect_identical(i8_v2, i8_v3)
})

test_that("string zarr arrays can be read correctly", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "string.zarr",
    package = "Rarr"
  )
  zarr_f <- system.file(
    "extdata",
    "zarr_examples",
    "row-first",
    "string.zarr",
    package = "Rarr"
  )
  index <- list(1:5, 1:10, 1)

  ## Shouldn't be any warnings for this input
  expect_silent(column_major <- read_zarr_array(zarr_c, index = index))
  expect_silent(row_major <- read_zarr_array(zarr_f, index = index))

  ## row and column major should be read the same in R
  expect_identical(column_major, row_major)

  ## we return an array
  expect_true(is.array(column_major))

  ## dimensions equal to the index
  expect_identical(dim(column_major), lengths(index))

  ## first row should be all "test" except the first element which is "ready"
  expect_identical(
    column_major[1, , ],
    c("ready", rep("test", length(index[[2]]) - 1))
  )
  ## first column should be all "ready"
  expect_true(all(column_major[, 1, ] == "ready"))

  ## check read/write produce the same things
  path <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = column_major,
      zarr_array_path = path,
      chunk_dim = c(5, 2, 1)
    )
  )
  expect_identical(read_zarr_array(path), column_major)

  ## check we truncate strings if the exceed the specified length
  path <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    res <- write_zarr_array(
      x = column_major,
      zarr_array_path = path,
      chunk_dim = c(5, 2, 1),
      nchar = 1
    )
  )
  expect_identical(max(nchar(read_zarr_array(path))), 1L)

  ## check if nchar argument is provided when creating empty character zarr arrays
  path <- withr::local_tempfile(fileext = ".zarr")
  expect_error(
    create_empty_zarr_array(
      zarr_array_path = path,
      dim = dim(column_major),
      chunk_dim = c(5, 2, 1),
      data_type = storage.mode(column_major)
    )
  )

  ########################

  greetings <- c(
    '¡Hola mundo!',
    'Hej Världen!',
    'Servus Woid!',
    'Hei maailma!',
    'Xin chào thế giới',
    'Njatjeta Botë!',
    'Γεια σου κόσμε!',
    'こんにちは世界',
    '世界，你好！',
    'Helló, világ!',
    'Zdravo svete!',
    'เฮลโลเวิลด์'
  )

  unicode_fixed_length <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "Unicode.zarr",
    package = "Rarr"
  )

  expect_silent(
    res <- read_zarr_array(unicode_fixed_length)
  )
  expect_identical(res[, 1], greetings)
  expect_identical(res[1, ], greetings)

  unicode_var_length <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "vlenUTF8.zarr",
    package = "Rarr"
  )

  expect_silent(
    res <- read_zarr_array(unicode_var_length)
  )
  expect_identical(res[, 1], greetings)
  expect_identical(res[1, ], greetings)

  ## writing & reading unicode
  path <- withr::local_tempfile(fileext = ".zarr")
  create_empty_zarr_array(
    path,
    dim = c(12, 12),
    chunk_dim = c(6, 6),
    data_type = "<U",
    nchar = 20
  )
  x <- array("", dim = c(12, 12))
  x[1, ] <- greetings
  x[, 1] <- greetings
  update_zarr_array(zarr_array_path = path, x, index = list(1:12, 1:12))
  expect_identical(read_zarr_array(path), x)
})

test_that("v2 and v3 return identical results", {
  zarr_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "string.zarr",
    package = "Rarr"
  )
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "string_v3.zarr",
    package = "Rarr"
  )

  expect_no_condition(s_v2 <- read_zarr_array(zarr_v2))
  expect_no_condition(s_v3 <- read_zarr_array(zarr_v3))

  expect_identical(s_v2, s_v3)

  zarr_vlen_utf8_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "vlenUTF8.zarr",
    package = "Rarr"
  )
  zarr_vlen_utf8_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "vlenUTF8_v3.zarr",
    package = "Rarr"
  )
  expect_no_condition(vlen_utf8_v2 <- read_zarr_array(zarr_vlen_utf8_v2))
  expect_no_condition(vlen_utf8_v3 <- read_zarr_array(zarr_vlen_utf8_v3))

  expect_identical(vlen_utf8_v2, vlen_utf8_v3)
})

test_that("roundtrip unicode", {
  path <- withr::local_tempfile(fileext = ".zarr")
  greetings <- array(
    c(
      '¡Hola mundo!',
      'Hej Världen!',
      'Servus Woid!',
      'Hei maailma!',
      'Xin chào thế giới',
      'Njatjeta Botë!',
      'Γεια σου κόσμε!',
      'こんにちは世界',
      '世界，你好！',
      'Helló, világ!',
      'Zdravo svete!',
      'เฮลโลเวิลด์'
    ),
    dim = c(12, 1)
  )
  write_zarr_array(
    greetings,
    path,
    chunk_dim = c(6, 1),
    data_type = "<U"
  )
  expect_identical(
    read_zarr_array(path),
    greetings
  )
})

test_that("fill_value is converted to string", {
  # https://github.com/Huber-group-EMBL/Rarr/issues/94
  path <- withr::local_tempfile(fileext = ".zarr")
  expect_silent(
    create_empty_zarr_array(
      zarr_array_path = path,
      dim = c(4, 4),
      chunk_dim = c(2, 2),
      data_type = "character",
      fill_value = 0,
      nchar = 10,
      zarr_version = 2
    )
  )
  expect_identical(read_zarr_array(path), matrix("0", nrow = 4, ncol = 4))
})

test_that("uint16 zarr arrays can be read correctly", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "uint16.zarr",
    package = "Rarr"
  )
  index <- list(1:30, 1:20, 1)

  ## Shouldn't be any warnings for this input
  expect_silent(column_major <- read_zarr_array(zarr_c, index = index))

  ## we return an array
  expect_true(is.array(column_major))

  ## dimensions equal to the index
  expect_identical(dim(column_major), c(30L, 20L, 1L))

  ## first row should be sequence 1 to 20
  expect_identical(column_major[1, , ], c(1:20))
  ## first column should be all 1s
  expect_identical(column_major[, 1, ], rep(1L, 30))
})

test_that("v2 and v3 return identical results", {
  zarr_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "uint16.zarr",
    package = "Rarr"
  )
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "uint16_v3.zarr",
    package = "Rarr"
  )

  expect_no_condition(u16_v2 <- read_zarr_array(zarr_v2))
  expect_no_condition(u16_v3 <- read_zarr_array(zarr_v3))

  expect_identical(u16_v2, u16_v3)
})

test_that("uint16 zarr array can be written", {
  uint16_zarr <- withr::local_tempfile(fileext = ".zarr")

  content_uint16_array <- array(1:24, dim = c(4, 3, 2))

  write_zarr_array(
    content_uint16_array,
    uint16_zarr,
    data_type = "<u2",
    chunk_dim = c(2, 2, 1),
    compressor = NULL
  )

  roundtrip_uint16_array <- read_zarr_array(uint16_zarr)

  expect_identical(
    content_uint16_array,
    roundtrip_uint16_array
  )
})

test_that("check_datatype works with data_type argument", {
  expect_identical(
    .check_datatype(data_type = "integer")[["data_type"]],
    "<i4"
  )
  expect_identical(
    .check_datatype(data_type = "double")[["data_type"]],
    "<f8"
  )
  expect_identical(
    .check_datatype(data_type = "character", nchar = "10")[[
      "data_type"
    ]],
    "|S10"
  )
})

test_that("check_datatype works with fill_value argument", {
  expect_identical(.check_datatype(fill_value = 4L)[["data_type"]], "<i4")
  expect_identical(.check_datatype(fill_value = 4)[["data_type"]], "<f8")
  expect_identical(
    .check_datatype(fill_value = "Test", nchar = 8)[["data_type"]],
    "|S8"
  )
})

test_that("data_type has precedence over fill_value when incompatible", {
  expect_identical(
    .check_datatype(data_type = "integer", fill_value = 4)[[
      "data_type"
    ]],
    "<i4"
  )
  expect_identical(
    .check_datatype(data_type = "double", fill_value = "Test")[[
      "data_type"
    ]],
    "<f8"
  )
  expect_identical(
    .check_datatype(
      data_type = "character",
      fill_value = 4L,
      nchar = 4
    )[[
      "data_type"
    ]],
    "|S4"
  )
})

test_that("check_datatype throws errors for invalid inputs", {
  expect_error(.check_datatype(data_type = "not-a-data-type"))
  expect_error(.check_datatype(data_type = "<f2"))
  expect_error(.check_datatype(data_type = list(1:10)))
  expect_error(.check_datatype(data_type = raw(10)))
})

test_that("check_datatype requires nchar for character types", {
  expect_error(
    .check_datatype(data_type = "|S"),
    "The 'nchar' argument must be provided"
  )
  expect_error(
    .check_datatype(fill_value = "Testing"),
    "The 'nchar' argument must be provided"
  )
  expect_error(
    .check_datatype(fill_value = "Testing", nchar = -4),
    "The 'nchar' argument must be provided"
  )
})

# This file regroups the specific conscious spec deviations we have implemented

test_that("chunk_size can be 0", {
  # https://github.com/zarr-developers/zarr-specs/issues/378
  new_zarr_array <- tempfile(fileext = ".zarr")
  write_zarr_array(
    array(numeric(0)),
    zarr_array_path = new_zarr_array,
    chunk_dim = 0
  )
  expect_identical(
    read_zarr_array(new_zarr_array),
    array(numeric(0))
  )
})

test_that("dimension_names is written for v2", {
  # TODO: implement reading as well?
  # https://github.com/zarr-developers/zarr-specs/blame/fc7dd9c9beb5a50b87f9b08b00bf50fc0048482f/docs/v2/v2.0.rst#L91-L92
  x <- array(
    1:60,
    c(4, 5, 3),
    dimnames = list(
      x = as.numeric(1:4),
      y = as.numeric(1:5),
      c = c("r", "g", "b")
    )
  )
  z <- tempfile(fileext = ".zarr")
  write_zarr_array(
    x,
    zarr_array_path = z,
    chunk_dim = c(2, 5, 1),
    zarr_version = 2L
  )
  expect_identical(
    jsonlite::read_json(file.path(z, ".zarray"))$dimension_names |> unlist(),
    c("x", "y", "c")
  )
})

#' @importFrom grumpy convert_bytes_to_array
codec_bytes_decode <- function(
  input,
  chunk_dim,
  datatype,
  endian,
  ...
) {
  convert_bytes_to_array(
    input,
    datatype$base_type,
    chunk_dim,
    datatype$nbytes,
    endian
  )
}

codec_bytes_encode <- function(input, datatype, endian) {
  if (is.character(input)) {
    ## we need to create fixed length strings either via padding or trimming
    if (datatype$base_type == "unicode") {
      to <- if (endian == "little") "UCS-4LE" else "UCS-4BE"
      raw_list <- iconv(input, to = to, toRaw = TRUE)
    } else {
      raw_list <- iconv(input, toRaw = TRUE)
    }
    unlist(
      lapply(
        raw_list,
        FUN = function(x, nbytes) {
          # NULL generated by iconv() for NA_character_
          if (is.null(x)) {
            x <- writeBin(NA_integer_, raw())
          }
          length(x) <- nbytes
          return(x)
        },
        nbytes = datatype$nbytes
      )
    )
  } else {
    if (is.na(endian)) {
      endian <- "little"
    }
    writeBin(input, raw(), size = datatype$nbytes, endian = endian)
  }
}


# -- Variable-length UTF-8 ------------------------
`codec_vlen-utf8_encode` <- function(input, ...) {
  raw_nvalues <- writeBin(length(input), raw(), size = 4L, endian = "little")
  # charToRaw() converts NA_character_ to "NA"
  raw_strings <- lapply(input, function(x) charToRaw(enc2utf8(x)))
  raw_string_lens <- .Call(
    "chop_vec",
    writeBin(lengths(raw_strings), raw(), size = 4L),
    4L,
    PACKAGE = "Rarr"
  )

  raw_vlen_utf8 <- c(
    raw_nvalues,
    unlist(mapply(
      function(len, str) c(len, str),
      raw_string_lens,
      raw_strings
    ))
  )
  return(raw_vlen_utf8)
}

`codec_vlen-utf8_decode` <- function(input, chunk_dim, ...) {
  .Call("codec_vlen_utf8_decode_c", input, chunk_dim, PACKAGE = "Rarr")
}

codec_sharding_indexed_decode <- function(
  input,
  shard_dim,
  datatype,
  fill_value,
  ...
) {
  config <- list(...)
  chunk_dim <- unlist(config$chunk_shape)
  nb_dims <- length(shard_dim)
  index_shape <- c(2L, shard_dim / chunk_dim)
  index_length <- prod(index_shape)
  index_nbytes <- index_length * 8L
  names(config$index_codecs) <- vapply(
    config$index_codecs,
    function(x) x$name,
    character(1L)
  )
  if ("crc32c" %in% names(config$index_codecs)) {
    index_nbytes <- index_nbytes + 4L
  }

  index_location <- config$index_location %||% "end"
  if (index_location == "end") {
    index_raw <- input[seq.int(to = length(input), length.out = index_nbytes)]
  } else {
    index_raw <- input[seq_len(index_nbytes)]
  }

  index <- readBin(
    index_raw,
    what = "integer",
    n = index_length,
    size = 8L,
    endian = config$index_codecs$bytes$configuration[["endian"]] %||%
      .Platform$endian
  )
  dim(index) <- index_shape
  index <- codec_transpose_decode(index, c(1L, rev(seq_len(nb_dims) + 1L)))

  configured_decoders <- config$codecs |>
    setNames(vapply(
      config$codecs,
      function(x) x$name,
      FUN.VALUE = character(1L)
    )) |>
    .configure_codecs(operation = "decode")

  chunks <- apply(
    index,
    seq_len(nb_dims) + 1L,
    function(x) {
      chunk_offset <- x[[1L]]
      chunk_nbytes <- x[[2L]]
      if (chunk_nbytes <= 0L) {
        return()
      }
      # Offset caused by index being at the start is already contained in the
      # offset returned by index.
      chunk_raw <- input[seq.int(
        chunk_offset + 1L,
        length.out = chunk_nbytes
      )]
      # FIXME: Once stores are implemented, treat this as a memory store and use `read_data()` directly
      read_chunk(
        chunk_raw,
        chunk_dim = chunk_dim,
        decoders = configured_decoders,
        datatype = datatype,
        fill_value = fill_value
      )
    }
  )
  shard <- array(fill_value, dim = shard_dim)
  non_empty_coords <- which(
    asplit(index, 1L)[[2L]] > 0L,
    arr.ind = TRUE
  )
  # Merge chunks into shard
  for (i in seq_len(nrow(non_empty_coords))) {
    coords_in_shard <- non_empty_coords[i, ]
    index_in_result <- mapply(
      FUN = function(coord, chunk_dim) {
        seq.int((coord - 1L) * chunk_dim + 1L, length.out = chunk_dim)
      },
      coords_in_shard,
      chunk_dim = chunk_dim,
      SIMPLIFY = FALSE
    )
    chunk_to_insert <- drop(.extract_chunk(chunks, coords_in_shard))[[1L]]
    rlang::inject(shard[!!!index_in_result] <- chunk_to_insert) # nolint implicit_assignment_linter.
  }

  return(shard)
}

# -- Transpose -------------------
codec_transpose_encode <- function(array, indices = seq_along(dim(array))) {
  # FIXME: we do two copies here (aperm and dim assignment) which can be expensive for large arrays
  inv_indices <- order(indices)
  array <- aperm(array, inv_indices)
  dim(array) <- dim(array)[inv_indices]

  return(array)
}

codec_transpose_decode <- function(array, indices = seq_along(dim(array))) {
  # FIXME: we manually dispatch aperm() to t() but eventually, there should be
  # an aperm.matrix() method.
  # https://bugs.r-project.org/show_bug.cgi?id=19133
  # When this happens, we only have to handle the drop() case.
  to_drop <- dim(array) == 1L
  if (any(to_drop)) {
    final_dim <- dim(array)
    array <- drop(array)
    indices <- order(indices[!to_drop])
  }

  if (is.null(dim(array)) || length(dim(array)) < 2L) {
    array <- array
  } else if (is.matrix(array) && all(indices == c(2L, 1L))) {
    dim(array) <- rev(dim(array))
    array <- t(array)
  } else {
    dim(array) <- dim(array)[indices]
    array <- aperm(array, indices)
  }

  # Restore dropped dimensions
  if (any(to_drop)) {
    dim(array) <- final_dim
  }
  return(array)
}

test_that("codec: transpose", {
  x <- array(1:60, dim = c(3, 4, 5))

  expect_identical(
    x |> codec_transpose_encode(c(1, 2, 3)),
    x
  )

  ind <- c(2, 3, 1)
  expect_identical(
    x |> codec_transpose_encode(ind) |> codec_transpose_decode(ind),
    x
  )

  ind <- c(3, 1, 2)
  expect_identical(
    x |> codec_transpose_encode(ind) |> codec_transpose_decode(ind),
    x
  )

  ind <- c(1, 3, 2)
  expect_identical(
    x |> codec_transpose_encode(ind) |> codec_transpose_decode(ind),
    x
  )
})

test_that("unsupported codec", {
  invalid_codec <- withr::local_tempfile(fileext = ".zarr")
  dir.create(invalid_codec)
  file.copy(
    system.file(
      "extdata",
      "zarr_examples",
      "column-first",
      "boolean_v3.zarr",
      package = "Rarr"
    ),
    invalid_codec,
    recursive = TRUE
  )
  meta <- jsonlite::read_json(
    file.path(invalid_codec, "boolean_v3.zarr", "zarr.json")
  )
  meta$codecs <- c(
    meta$codecs,
    list(
      list(
        "name" = "unsupported_codec",
        "configuration" = list()
      )
    )
  )
  jsonlite::write_json(
    meta,
    file.path(invalid_codec, "boolean_v3.zarr", "zarr.json"),
    auto_unbox = TRUE,
    pretty = TRUE
  )
  expect_error(
    Rarr::read_zarr_array(file.path(invalid_codec, "boolean_v3.zarr")),
    "not supported: unsupported_codec"
  )
})

#' Read a Zarr array
#'
#' @param zarr_array_path Path to a Zarr array. A character vector of length 1.
#'   This can either be a location on a local file system or the URI to an array
#'   in S3 storage.
#' @param index A list of the same length as the number of dimensions in the
#'   Zarr array.  Each entry in the list provides the indices in that dimension
#'   that should be read from the array.  Setting a list entry to `NULL` will
#'   read everything in the associated dimension.  If this argument is missing
#'   the entirety of the the Zarr array will be read.
#' @param s3_client Object created by [paws.storage::s3()]. Only required for a
#'   file on S3. Leave as `NULL` for a file on local storage.
#'
#' @returns An array with the same number of dimensions as the input array. The
#'   extent of each dimension will correspond to the length of the values
#'   provided to the `index` argument.
#'
#' @examples
#'
#' ## Using a local file provided with the package
#' ## This array has 3 dimensions
#' z1 <- system.file("extdata", "zarr_examples", "row-first", "int32.zarr", package = "Rarr")
#'
#' ## read the entire array
#' read_zarr_array(zarr_array_path = z1)
#'
#' ## extract values for first 10 rows, all columns, first slice
#' read_zarr_array(zarr_array_path = z1, index = list(1:10, NULL, 1))
#'
#' \donttest{
#' ## using a Zarr file hosted on Amazon S3
#' ## This array has a single dimension with length 2729077
#' z2 <- "https://noaa-nwm-retro-v2-zarr-pds.s3.amazonaws.com/feature_id/"
#'
#' ## read the entire array
#' read_zarr_array(zarr_array_path = z2)
#'
#' ## read alternating elements
#' read_zarr_array(zarr_array_path = z2, index = list(seq(1, 576, 2)))
#' }
#'
#' @export
read_zarr_array <- function(zarr_array_path, index, s3_client = NULL) {
  zarr_array_path <- .normalize_array_path(zarr_array_path)
  ## determine if this is a local or S3 array
  s3_client <- s3_client %||% .create_s3_client(path = zarr_array_path)

  metadata <- .read_array_metadata(
    zarr_array_path,
    s3_client = s3_client
  )

  if (metadata$node_type == "group") {
    stop(
      "The provided path points to a Zarr group, but `read_zarr_array()` can ",
      "only read arrays. Please provide the path to an array within the group.",
      call. = FALSE
    )
  }

  metadata$configured_decoders <- .configure_codecs(
    metadata$codecs,
    operation = "decode"
  )

  ## if no index provided we will return everything
  if (missing(index)) {
    index <- vector(mode = "list", length = length(metadata$shape))
  }
  index <- check_index(index = index, metadata = metadata)

  res <- read_data(zarr_array_path, s3_client, index, metadata)

  if (!is.null(metadata$dimension_names)) {
    dimnames(res) <- setNames(
      vector("list", length = length(dim(res))),
      metadata$dimension_names
    )
  }

  return(res)
}


read_data <- function(
  zarr_array_path,
  s3_client,
  index,
  metadata
) {
  if (!is.null(s3_client)) {
    parsed_s3_url <- parse_s3_path(zarr_array_path)
    bucket <- parsed_s3_url$bucket
    root <- parsed_s3_url$object
  } else {
    bucket <- NULL
    root <- zarr_array_path
  }

  ## precompute, for each chunk, the positions in `index` that belong to it
  chunk_dim <- unlist(metadata$chunk_grid$configuration$chunk_shape)
  chunk_positions <- .chunk_positions_by_chunk(
    index,
    metadata,
    chunk_dim
  )

  chunk_names <- names(chunk_positions)
  # In the DelayedArray framework, we can have integer(0) indices
  # https://github.com/Huber-group-EMBL/Rarr/issues/112
  chunk_paths <- paste0(root, chunk_names, recycle0 = TRUE)

  ## Vectorized check for chunk existence
  chunk_exists <- .store_check_exist(zarr_array_path, chunk_names, s3_client)
  existing_idx <- which(chunk_exists)

  warnings <- list()
  ## hopefully we can eventually do this in parallel
  chunk_selections <- withCallingHandlers(
    lapply(
      # We skip missing chunks here since they will just be filled with the fill value
      # when initializing the consolidated array.
      existing_idx,
      function(i) {
        .extract_elements(
          chunk_name = chunk_names[i],
          current_chunk_path = chunk_paths[i],
          metadata = metadata,
          chunk_dim = chunk_dim,
          s3_client = s3_client,
          s3_bucket = bucket,
          chunk_positions = chunk_positions
        )
      }
    ),
    warning = function(w) {
      warnings <<- c(warnings, list(w)) # nolint: undesirable_operator_linter.
      invokeRestart("muffleWarning")
    }
  )
  for (w in unique(warnings)) {
    warning(w)
  }

  ## predefine our array to be populated from the read chunks
  output <- array(metadata$fill_value, dim = lengths(index))

  ## proceed in serial and update the output with each chunk selection in turn
  for (i in seq_along(chunk_selections)) {
    index_in_result <- chunk_selections[[i]][[2L]]
    rlang::inject(output[!!!index_in_result] <- chunk_selections[[i]][[1L]]) # nolint: implicit_assignment_linter.
    if (
      is.list(metadata$data_type) &&
        metadata$data_type$name %in% c("struct", "structured")
    ) {
      # Assigning a list drops the dim attribute so we have to continuously add it again
      dim(output) <- lengths(index)
    }
  }
  return(output)
}

.extract_elements <- function(
  chunk_name,
  current_chunk_path,
  metadata,
  chunk_dim,
  s3_client,
  s3_bucket,
  chunk_positions
) {
  ## find elements to select from the chunk and what in the output we replace
  chunk_info <- chunk_positions[[chunk_name]]
  index_in_result <- chunk_info$positions
  index_in_chunk <- chunk_info$index_in_chunk

  # When we get here, we know the chunk exists, so we can read it without worrying about
  # handling missing.
  if (nzchar(Sys.getenv("RARR_DEBUG"))) {
    message(current_chunk_path)
  }

  if (is.null(s3_client)) {
    size <- file.size(current_chunk_path)
    raw_chunk <- readBin(con = current_chunk_path, what = "raw", n = size)
  } else {
    raw_chunk <- s3_client$get_object(
      Bucket = s3_bucket,
      Key = current_chunk_path
    )$Body
  }

  ## read this chunk
  chunk <- read_chunk(
    chunk_bytes = raw_chunk,
    chunk_dim = chunk_dim,
    decoders = metadata$configured_decoders,
    datatype = metadata$datatype,
    fill_value = metadata$fill_value
  )

  ## extract the required elements from the chunk
  # FIXME: optimization: skip this step if we are taking everything in the chunk
  chunk <- .extract_chunk(chunk, index_in_chunk)
  return(list(chunk, index_in_result))
}

#' Read a single Zarr chunk
#'
#' @param chunk_bytes A raw vector containing the bytes of the chunk to be read.
#' @param chunk_dim The dimensions of the chunk to be read. A numeric vector.
#' @param decoders A list of configured decoders for the array.
#' @param datatype A list describing the datatype of the array.
#'
#' @returns An array containing the decompressed chunk values.
#'
#' @keywords internal
#' @noRd
read_chunk <- function(
  chunk_bytes,
  chunk_dim,
  decoders,
  datatype,
  fill_value
) {
  # Bytes -> Bytes codecs
  for (codec in decoders[["bytes_bytes"]]) {
    chunk_bytes <- codec(
      bytes = chunk_bytes
    )
  }

  # Bytes -> Array codecs
  for (codec in decoders[["array_bytes"]]) {
    converted_chunk <- codec(
      chunk_bytes,
      chunk_dim,
      datatype,
      fill_value
    )
  }
  # Array -> Array codecs
  for (codec in decoders[["array_array"]]) {
    converted_chunk <- do.call(codec, list(converted_chunk))
  }

  return(converted_chunk)
}

test_that("compressed zarr arrays (bzip2) can be read correctly", {
  zarr <- system.file(
    "extdata",
    "zarr_examples",
    "compression",
    "bzip2.zarr",
    package = "Rarr"
  )

  expect_silent(data <- read_zarr_array(zarr))

  expect_identical(data[1, ], 1:10)
  expect_true(all(data[, 1] == seq(1, 60, 3)))
})

test_that("compressed zarr arrays (lz4) can be read correctly", {
  zarr_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "compression",
    "lz4.zarr",
    package = "Rarr"
  )
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "compression",
    "lz4_v3.zarr",
    package = "Rarr"
  )

  expect_silent(data_v2 <- read_zarr_array(zarr_v2))
  expect_silent(data_v3 <- read_zarr_array(zarr_v3))

  expect_identical(data_v2, data_v3)

  expect_identical(data_v2[1, ], 1:10)
  expect_true(all(data_v2[, 1] == seq(1, 60, 3)))
})

test_that("compressed zarr arrays (lzma) can be read correctly", {
  zarr <- system.file(
    "extdata",
    "zarr_examples",
    "compression",
    "lzma.zarr",
    package = "Rarr"
  )

  expect_silent(data <- read_zarr_array(zarr))

  expect_identical(data[1, ], 1:10)
  expect_true(all(data[, 1] == seq(1, 60, 3)))
})

test_that("compressed zarr arrays (zlib) can be read correctly", {
  zarr <- system.file(
    "extdata",
    "zarr_examples",
    "compression",
    "zlib.zarr",
    package = "Rarr"
  )

  expect_silent(data <- read_zarr_array(zarr))

  expect_identical(data[1, ], 1:10)
  expect_true(all(data[, 1] == seq(1, 60, 3)))
})

test_that("compressed zarr arrays (zstd) can be read correctly", {
  zarr <- system.file(
    "extdata",
    "zarr_examples",
    "compression",
    "zstd.zarr",
    package = "Rarr"
  )

  expect_silent(data <- read_zarr_array(zarr))

  expect_identical(data[1, ], 1:10)
  expect_true(all(data[, 1] == seq(1, 60, 3)))

  # Decompression of unknown buffer size
  # https://github.com/Huber-group-EMBL/Rarr/issues/20
  zarr_zstd_vlen <- system.file(
    "extdata",
    "zarr_examples",
    "compression",
    "zstd_vlen.zarr",
    package = "Rarr"
  )

  expect_silent(data_vlen <- read_zarr_array(zarr_zstd_vlen))

  expect_identical(data_vlen, array(as.character(0:5), dim = 6))
})

test_that("scalar structured data types", {
  common_type_file <- system.file(
    "extdata",
    "zarr_examples",
    "structured",
    "common_type.zarr",
    package = "Rarr"
  )

  expect_silent(
    res <- read_zarr_array(common_type_file)
  )

  expect_shape(res, dim = c(10, 10))
})

test_that("unicode structured data types", {
  unicode_type_file <- system.file(
    "extdata",
    "zarr_examples",
    "structured",
    "unicode.zarr",
    package = "Rarr"
  )

  expect_silent(
    res <- read_zarr_array(unicode_type_file)
  )

  expect_shape(res, dim = c(10, 10))
})

test_that("v3 structured data types", {
  v3_struct_file <- system.file(
    "extdata",
    "zarr_examples",
    "structured",
    "struct_v3.zarr",
    package = "Rarr"
  )

  expect_silent(
    res <- read_zarr_array(v3_struct_file)
  )

  expect_shape(res, dim = c(10, 2))
  dim(res) <- NULL
  for (el in res) {
    expect_length(el, 2L)
    # FIXME: support names in structured datatypes
    # expect_named(el, c("x", "y"))
    expect_type(el[[1]], "integer")
    expect_type(el[[2]], "double")
  }

  v3_deprecated_structured_file <- system.file(
    "extdata",
    "zarr_examples",
    "structured",
    "structured_v3.zarr",
    package = "Rarr"
  )

  expect_silent(
    res <- read_zarr_array(v3_deprecated_structured_file)
  )

  expect_shape(res, dim = 100)
  dim(res) <- NULL
  for (el in res) {
    expect_length(el, 6L)
    expect_all_true(vapply(el, is.double, FUN.VALUE = logical(1)))
  }
})

test_that("scalar can be read", {
  # Not explicitly supported (but still used) in v2.
  # v3 spec says: "A shape can be the empty tuple in the case of zero-dimension arrays (scalars)."
  scalar_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "scalar",
    "scalar_v2.zarr",
    package = "Rarr"
  )
  expect_identical(read_zarr_array(scalar_v2), structure("A string", dim = 1))

  scalar_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "scalar",
    "scalar_v3.zarr",
    package = "Rarr"
  )
  expect_identical(read_zarr_array(scalar_v3), structure(42L, dim = 1))
})

test_that("other zarr arrays can be read correctly", {
  test_zarr_file <- system.file(
    "extdata",
    "zarr_examples",
    "row-first",
    "other.zarr",
    package = "Rarr"
  )

  # Shouldn't be any warnings for this input
  expect_silent(
    test_array <- read_zarr_array(test_zarr_file)
  )

  # Should return an array of character values from 0 to 5
  expect_identical(
    test_array,
    array(as.character(0:5))
  )
})

test_that("bfloat16 can be read", {
  bfloat16_array <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "bfloat16.zarr",
    package = "Rarr"
  ) |>
    read_zarr_array() |>
    expect_no_condition()

  expect_shape(bfloat16_array, dim = c(4L, 6L))
  expect_equal(
    bfloat16_array,
    matrix(
      1:24,
      nrow = 4L,
      ncol = 6L,
      byrow = TRUE
    )
  )
})

test_that("read sharded files", {
  sharded <- system.file(
    "extdata",
    "zarr_examples",
    "sharding",
    "int32_sharded.zarr",
    package = "Rarr"
  )

  arr <- read_zarr_array(sharded) |>
    expect_no_condition() |>
    expect_shape(dim = c(30L, 20L, 10L)) |>
    expect_type("integer")

  expect_identical(
    arr[1L, , 1L],
    seq_len(20L)
  )
  expect_identical(
    arr[, 1L, 1L],
    rep_len(1L, 30L)
  )

  expect_identical(
    arr[2L:30L, 2L:20L, 2L:10L],
    array(0L, dim = c(29L, 19L, 9L))
  )

  # Index location doesn't matter
  sharded_index_start <- system.file(
    "extdata",
    "zarr_examples",
    "sharding",
    "int32_sharded_index_at_start.zarr",
    package = "Rarr"
  )

  expect_identical(
    read_zarr_array(sharded_index_start),
    arr
  )
})

test_that("read_zarr_array() gives an informative error on v3 groups", {
  zarr_group <- withr::local_tempdir(fileext = ".zarr")

  jsonlite::write_json(
    list(
      zarr_format = 3L,
      node_type = "group"
    ),
    file.path(zarr_group, "zarr.json"),
    auto_unbox = TRUE
  )
  expect_error(
    read_zarr_array(zarr_group),
    "The provided path points to a Zarr group, but `read_zarr_array()` can only read arrays.",
    fixed = TRUE
  )
})

test_that("float16 zarr arrays can be read correctly", {
  zarr_c <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "float16.zarr",
    package = "Rarr"
  )
  zarr_f <- system.file(
    "extdata",
    "zarr_examples",
    "row-first",
    "float16.zarr",
    package = "Rarr"
  )
  index <- list(1:30, 1:20, 1)

  # Shouldn't be any warnings for this input
  expect_silent(column_major <- read_zarr_array(zarr_c, index = index))
  expect_silent(row_major <- read_zarr_array(zarr_f, index = index))

  # Row and column major should be read the same in R
  expect_identical(column_major, row_major)

  # We return an array
  expect_true(is.array(column_major))

  # Dimensions equal to the index
  expect_identical(dim(column_major), lengths(index))

  # First row should be 2 to 20 except the first element which is -1
  expect_identical(column_major[1, , ], c(-1, 2:20))

  # First column should be all 10.52
  expect_identical(column_major[2:20, 1, ], rep(10.52, 19), tolerance = 0.0005)

  expect_identical(
    column_major[2, 2:4, ],
    c(0.00005693, -5.97e-8, 0.0000039),
    tolerance = 0.0001
  )

  expect_identical(column_major[3, 2:4, ], c(NaN, Inf, -Inf))
})

test_that("v2 and v3 return identical results", {
  zarr_v2 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "float16.zarr",
    package = "Rarr"
  )
  zarr_v3 <- system.file(
    "extdata",
    "zarr_examples",
    "column-first",
    "float16_v3.zarr",
    package = "Rarr"
  )

  expect_no_condition(f16_v2 <- read_zarr_array(zarr_v2))
  expect_no_condition(f16_v3 <- read_zarr_array(zarr_v3))

  expect_identical(f16_v2, f16_v3)
})

CODEC_BYTES_BYTES <- c(
  "blosc",
  "zlib",
  "gzip",
  "bz2",
  "lzma",
  "numcodecs.lz4",
  "lz4",
  "zstd"
)
CODEC_ARRAY_ARRAY <- c("transpose")
CODEC_ARRAY_BYTES <- c("bytes", "vlen-utf8", "sharding_indexed")

SUPPORTED_V3_TYPES <- list(
  "bool" = list(base_type = "bool", nbytes = 1L),
  "int8" = list(base_type = "int", nbytes = 1L),
  "int16" = list(base_type = "int", nbytes = 2L),
  "int32" = list(base_type = "int", nbytes = 4L),
  "int64" = list(base_type = "int", nbytes = 8L),
  "uint8" = list(base_type = "uint", nbytes = 1L),
  "uint16" = list(base_type = "uint", nbytes = 2L),
  "uint32" = list(base_type = "uint", nbytes = 4L),
  "uint64" = list(base_type = "uint", nbytes = 8L),
  "float16" = list(base_type = "float", nbytes = 2L),
  "float32" = list(base_type = "float", nbytes = 4L),
  "float64" = list(base_type = "float", nbytes = 8L),
  "string" = list(base_type = "string", nbytes = NA_integer_)
) |>
  list2env(hash = TRUE, parent = emptyenv())

usethis::use_data(
  SUPPORTED_V3_TYPES,
  CODEC_BYTES_BYTES,
  CODEC_ARRAY_ARRAY,
  CODEC_ARRAY_BYTES,
  internal = TRUE,
  overwrite = TRUE
)

test_that("creation of deeply nested group", {
  zarr_v2 <- withr::local_tempfile(fileext = ".zarr")
  expect_no_condition(
    write_zarr_group(zarr_v2, "test/deep/nested/group", zarr_version = 2L)
  )
  expect_true(dir.exists(file.path(zarr_v2, "test/deep/nested/group")))
  expect_identical(
    list.files(zarr_v2, recursive = TRUE, all.files = TRUE),
    c(
      ".zgroup",
      "test/.zgroup",
      "test/deep/.zgroup",
      "test/deep/nested/.zgroup",
      "test/deep/nested/group/.zgroup"
    )
  )
})

---
title: "Supported Zarr features in Rarr"
date: "`r BiocStyle::doc_date()`"
package: "`r BiocStyle::pkg_ver('Rarr')`"
vignette: >
  %\VignetteIndexEntry{"Supported Zarr features in Rarr"}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document
---

```{css, echo=FALSE}
.main-container {
  max-width: 1600px;
  margin-left: auto;
  margin-right: auto;
}
```

## Zarr version

Rarr supports both Zarr version 2 and Zarr version 3.

## Reading and Writing

Reading Zarr arrays is reasonably well supported. Writing is available, but is more limited.  Both aspects are under active development.

### Stores

| Store Type       | Status<br/>(reading / writing) | Notes |
|------------------|:------------------------------:|-------|
|`FilesystemStore` | &#x2714; / &#x2714;            |       |
|`S3Store`         | &#x2714; / &#x274C;            | Read more in the [S3 vignette](vignettes/S3.html) |

### Data Types

Currently there is only support for reading and writing a subset of the possible datatypes
that can be found in a Zarr array.  In some instances there are also limitations on the
datatypes natively supported by R, requiring conversion from the Zarr datatype.  The table below summarises the current status of
datatype support.  It will be updated as progress is made.

| Zarr Data Type | Status<br/>(reading / writing) | Notes |
|----------------|:------------------------------:|-------|
|`boolean`       | &#x2714; / &#x2714;            | (via `{grumpy}`) |
|`int8`          | &#x2714; / &#x2714;            | (via `{grumpy}`) |
|`uint8`         | &#x2714; / &#x2714;            | (via `{grumpy}`) |
|`int16`         | &#x2714; / &#x2714;            | (via `{grumpy}`) |
|`uint16`        | &#x2714; / &#x2714;            | (via `{grumpy}`) |
|`int32`         | &#x2714; / &#x2714;            | (via `{grumpy}`) |
|`uint32`        | &#x2714; / &#x2714;            | (via `{grumpy}`) |
|`int64`         | &#x2714; / &#x2714;            | (via `{grumpy}`) |
|`uint64`        | &#x2714; / &#x2714;            | (via `{grumpy}`) |
|`half` / `float16`   | &#x2714; / &#x274C; | (via `{grumpy}`) |
|`single` / `float32` | &#x2714; / &#x2714; | (via `{grumpy}`) |
|`double` / `float64` | &#x2714; / &#x2714; | (via `{grumpy}`) |
|`complex`            | &#x274C; / &#x274C; | |
|`timedelta`          | &#x274C; / &#x274C; | |
|`datetime`           | &#x274C; / &#x274C; | |
| V2 `string`             | &#x2714; / &#x2714; | (via `{grumpy}`) |
| V3 `null_terminated_bytes` | &#x2714; / &#x2714; | (via `{grumpy}`) |
| V2 `Unicode`            | &#x2714; / &#x2714; | (via `{grumpy}`) |
| V3 `fixed_length_utf32` | &#x2714; / &#x2714; | (via `{grumpy}`) |
|`void *`             | &#x274C; / &#x274C; | |
| V2 structured data types | &#x2714; / &#x274C; | (via `{grumpy}`) |
| V3 structured data types | &#x2714; / &#x274C; | This type is superseded by V3 struct data types.and the specification says implementation MUST NOT implement writing. (via `{grumpy}`) |
| V3 struct data types | &#x2714; / &#x274C; | (via `{grumpy}`) |

### Codecs

#### Compression codecs

| Library     | Status<br/>(reading / writing) | Notes |
|-------------|:-------------------:|-------|
|`zlib / gzip`| &#x2714; / &#x2714; | |
|`bzip2`      | &#x2714; / &#x2714; | |
|`blosc`      | &#x2714; / &#x2714; | |
|`LZMA `      | &#x2714; / &#x2714; | |
|`LZ4`        | &#x2714; / &#x2714; | |
|`Zstd`       | &#x2714; / &#x2714; | |

Please open an [issue](https://github.com/Huber-group-EMBL/Rarr/issues) if support for a required compression codec is missing.

#### Other codecs

| Codec     | Status<br/>(reading / writing) | Notes |
|-----------|:------------------------------:|-------|
|`endian`   | &#x2714; / &#x274C;            | (via `{grumpy}`) |
|`transpose`| &#x2714; / &#x2714;            |       |
|`VLen-UTF8`| &#x2714; / &#x2714; *          | Writing supported only via `update_zarr_array()` |
|`sharding_indexed` | &#x2714; / &#x274C;    |  |

### Optional fields or features

| Feature    | Status<br/>(reading / writing) | Notes |
|------------|:------------------------------:|-------|
|`dimension_names` | &#x2714; / &#x2714; | Corresponds to `names(dimnames(x))` in R |

## Installation and setup

If you want to quickly get started reading an existing Zarr array with the
package, this section should have the essentials covered.  First, we need to
install **Rarr**^[you only need to do the installation step once] with the
commands below.

```{r, installation, eval = FALSE}
## we need BiocManager to perform the installation
if (!require("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}
## install Rarr
BiocManager::install("Rarr")
```

Once **Rarr** is installed, we have to load it into our R session:

```{r message=FALSE}
library(Rarr)
```

**Rarr** can be used to read files either on local disk or on remote S3 storage
systems.  First lets take a look at reading from a local file.

## Reading a from a local Zarr array
To demonstrate reading a local file, we'll pick the example file containing
32-bit integers arranged in the "column first" ordering.

```{r, use-example}
zarr_example <- system.file(
  "extdata",
  "zarr_examples",
  "column-first",
  "int32.zarr",
  package = "Rarr"
)
```

### Exploring the data

We can get an summary of the array properties, such as its shape and datatype, or group properties,
using the function `zarr_overview()`^[This is essentially reading and formatting
the array metadata that accompanies any Zarr array, or the consolidated metadata if present in the case of a Zarr group.].

```{r, overview-1}
zarr_overview(zarr_example)
```

You can use this to check that the location is a valid Zarr array, and that the
shape and datatype of the array content are what you are expecting.  For
example, we can see in the output above that the data type (`int32`) corresponds
to what we expect.

### Reading the Zarr array

The summary information retrieved above is required, as to read the array with
**Rarr** you need to know the shape and size of the array (unless you want to
read the entire array).  From the previous output we can see our example array
has three dimensions of size 30 x 20 x 10.  We can select the subset we want to
extract using a `list`. The list must have the same length as the number of
dimensions in our array, with each element of the list corresponding to the
indices you want to extract in that dimension.

```{r}
index <- list(1:4, 1:2, 1)
```

We then extract the subset using `read_zarr_array()`:

```{r, extract-subset}
read_zarr_array(zarr_example, index = index)
```

## Reading from S3 storage {#read-s3}

Read the dedicated ["Working with **remote** Zarr arrays in R" vignette](https://huber-group-embl.github.io/Rarr/articles/S3.html) for more information on reading Zarr arrays from S3 storage.

## Writing to a Zarr array

Up until now we've only covered reading existing Zarr array into R.  However,
**Rarr** can also be used to write R data to disk following the Zarr
specification.  To explore this, lets create an example array we want to save as
a Zarr.  In this case it's going to be a three dimensional array and store the
values 1 to 600.

```{r, example-array}
x <- array(1:600, dim = c(10, 10, 6))
```

```{r}
path_to_new_zarr <- file.path(tempdir(), "new.zarr")
write_zarr_array(
  x = x,
  zarr_array_path = path_to_new_zarr,
  chunk_dim = c(10, 5, 1)
)
```

We can check that the contents of the Zarr array is what we're expecting.  Since
the contents of the whole array will be too large to display here, we use the
`index` argument to extract rows 6 to 10, from the 10th column and 1st slice.
That should be the values `r x[6:10, 10, 1]`, but retaining the 3-dimensional
array structure of the original array.  The second line below uses `identical()`
to confirm that reading the whole Zarr returns something equivalent to our
original input `x`.

```{r, check-written-array}
read_zarr_array(zarr_array_path = path_to_new_zarr, index = list(6:10, 10, 1))
identical(read_zarr_array(zarr_array_path = path_to_new_zarr), x)
```

---
title: "Design principles for the Rarr package"
date: "`r BiocStyle::doc_date()`"
author:
  - name: Hugo Gruson
affiliation:
  - EMBL Heidelberg
email: hugo.gruson@embl.de
package: "`r BiocStyle::pkg_ver('Rarr')`"
vignette: >
  %\VignetteIndexEntry{Design principles for the Rarr package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document
---

## Guiding principles

- **performant**. We focus on speed & memory efficiency. This results in Rarr being among the most performant Zarr implementations. This is ensured by the use of an extensive [continuous benchmarking](https://lorenzwalthert.github.io/touchstone/index.html) suite. Performance critical steps are written in C.
- **maintainable and extensible**. Additional codecs can easily be supported as the entire codec codebase is decoupled from the rest of the codebase.

## Scope

We aim for full support of the Zarr specification.

There is currently no clear decision process regarding support for [Zarr extensions](https://github.com/zarr-developers/zarr-extensions). Please reach out if you have a specific use case that relies on a Zarr extension.

## Zarr version

Rarr is a "Zarr version 3 first" implementation.

There is full support for both version 2 and version 3 Zarr arrays but:

- the package API is modelled on the version 3 specification
- we backport some backward compatible feature from version 3 to version 2. For example, the `dimension_names` field is supported for both version 2 and version 3 arrays. This is neither strictly defined nor forbidden in the version 2 specification, but it is a feature that we have chosen to support in both versions for consistency.
- if we ever had to make a decision on a tradeoff (e.g., performance tradeoff) between version 2 and version 3, we would prioritise version 3.

## Functional programming and API design

- Reading and writing Zarr arrays should be as easy as reading and writing `.csv` files. In other words, writing `read_zarr_array("my_array.zarr")` should be enough to read a Zarr array, and writing `write_zarr_array(my_array, "my_array.zarr")` should be enough to write a Zarr array. This has several consequences:
  - it is not necessary to explicitly manipulate custom objects for Zarr stores or groups. Passing a file path as a string should work out of the box.
  - when writing, we provide sensible defaults for:
    - data type (derived from `storage.mode` of the input array)
    - chunk size (TBD)
    - compression (Zstd with default compression level, for a good balance between speed and compression ratio)
    - dimension names (taken from the dimension names of the input array, if they exist)

---
title: "Working with **remote** Zarr arrays in R"
date: "23 June 2026"
author:
  - name: Hugo Gruson
    affiliation:
      - EMBL Heidelberg
    email: hugo.gruson@embl.de
  - name: Mike Smith
    affiliation:
      - EMBL Heidelberg
package: "Rarr 2.1.18"
vignette: >
  %\VignetteIndexEntry{"Working with **remote** Zarr arrays in R"}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document
---

<!-- 
# Pre-render with 
knitr::knit("vignettes/_S3.Rmd", output = "vignettes/S3.Rmd")
-->




``` r
library(Rarr)
```

It is recommended you read the [general introduction "Working with Zarr arrays in R"](https://huber-group-embl.github.io/Rarr/articles/Rarr.html) before reading this vignette.

Reading files in S3 storage works in a very similar fashion to local disk.  This
time the path needs to be a URL to the Zarr array.
We can again use `zarr_overview()` to quickly retrieve the array metadata.


``` r
s3_address <- "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0076A/10501752.zarr/0"
zarr_overview(s3_address)
```

```
## Type: Array
## Path: https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0076A/10501752.zarr/0
## Shape: 50 x 494 x 464
## Chunk Shape: 1 x 494 x 464
## No. of Chunks: 50 (50 x 1 x 1)
## Data Type: float64
## Endianness: little
## Compressor: blosc
```

You can also pass an S3 client to the function, which is useful if you need to set credentials or other options for accessing the bucket. See the section \@ref(s3-client) for more details.
If absent, *[Rarr](https://bioconductor.org/packages/3.24/Rarr)* will try to find credentials and other settings on its own, which may not always be successful.
This is equivalent to the previous code block:


``` r
s3_address <- "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0076A/10501752.zarr/0"
s3_client <- paws.storage::s3(
  config = list(
    credentials = list(anonymous = TRUE),
    region = "auto",
    endpoint = "https://uk1s3.embassy.ebi.ac.uk"
  )
)
zarr_overview(s3_address, s3_client = s3_client)
```

```
## Type: Array
## Path: https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0076A/10501752.zarr/0
## Shape: 50 x 494 x 464
## Chunk Shape: 1 x 494 x 464
## No. of Chunks: 50 (50 x 1 x 1)
## Data Type: float64
## Endianness: little
## Compressor: blosc
```

The output above indicates that the array is stored in 50 chunks, each
containing a slice of the overall data. In the example below we use the `index`
argument to extract the first and tenth slices from the array. Choosing to read only
2 of the 50 slices is much faster than if we opted to download the entire array
before accessing the data.


``` r
z2 <- read_zarr_array(
  s3_address,
  index = list(c(1, 10), NULL, NULL)
)
```

We then plot our two slices on top of one another using the `image()` function.


``` r
## plot the first slice in blue
image(
  log2(z2[1, , ]),
  col = hsv(h = 0.6, v = 1, s = 1, alpha = 0:100 / 100),
  asp = dim(z2)[2] / dim(z2)[3],
  axes = FALSE
)
## overlay the tenth slice in green
image(
  log2(z2[2, , ]),
  col = hsv(h = 0.3, v = 1, s = 1, alpha = 0:100 / 100),
  asp = dim(z2)[2] / dim(z2)[3],
  axes = FALSE,
  add = TRUE
)
```

![plot of chunk plot-raster](figure/plot-raster-1.png)

**Note:** if you receive the error message
`"Error in stop(aws_error(request$error)) : bad error message"` it is likely you
have some AWS credentials available in to your R session, which are being
inappropriately used to access this public bucket.  Please see the section
\@ref(s3-client) for details on how to set credentials for a specific
request.

## Using credentials to access S3 buckets {#s3-credentials}

If you're accessing data in a private S3 bucket, you can set the environment
variables `AWS_ACCESS_KEY_ID` and `AWS_SECRET_ACCESS_KEY` to store your
credentials.  For example, lets try reading a file in a private S3 bucket:


``` r
zarr_overview("https://s3.embl.de/rarr-testing/bzip2.zarr")
```

```
## Error:
## ! AccessDenied (HTTP 403). Access Denied.
```

We can see the "Access Denied" message in our output, indicating that we don't
have permission to access this resource as an anonymous user.  However, if we use the key pair
below, which gives read-only access to the objects in the `rarr-testing` bucket,
we're now able to interrogate the files with functions in *Rarr*.


``` r
Sys.setenv(
  "AWS_ACCESS_KEY_ID" = "bYUBYVg1AsEreuDgtg5K",
  "AWS_SECRET_ACCESS_KEY" = "r8FrLXc9dseD6V1P3htsu7ZBzP7Gszsd3sM1G4KX"
)
zarr_overview("https://s3.embl.de/rarr-testing/bzip2.zarr")
```

```
## Type: Array
## Path: https://s3.embl.de/rarr-testing/bzip2.zarr
## Shape: 20 x 10
## Chunk Shape: 10 x 10
## No. of Chunks: 2 (2 x 1)
## Data Type: int32
## Endianness: little
## Compressor: bz2
```

Behind the scenes **Rarr** makes use of the **paws** suite of packages
(https://paws-r.github.io/) to interact with S3 storage.  A comprehensive
overview of the multiple ways credentials can be set and used by **paws** can be
found at https://github.com/paws-r/paws/blob/main/docs/credentials.md.  If
setting environment variables as above doesn't work or is inappropriate for
your use case please refer to that document for other options.

## Creating an S3 client {#s3-client}

Although **Rarr** will try its best to find appropriate credentials and settings
to access a bucket, it is not always successful. Once such example is when you
have AWS credentials set somewhere and you try to access a public bucket.  We
can see an example of this below, where we access the same public bucket used in
\@ref(read-s3), but it now fails because we have set the `AWS_ACCESS_KEY_ID`
environment variable in the previous section.



``` r
s3_address <- "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0076A/10501752.zarr/0"
zarr_overview(s3_address)
```

```
## 
```

You might encounter similar problems if you're trying to access multiple buckets
each of which require different credentials.  The solution here is to create an
"s3_client" using `paws.storage::s3()`, which contains all the required details
for accessing a particular bucket.  Doing so will prevent **Rarr** from trying
to determine things on its own, and gives you complete control over the settings
used to communicate with the S3 bucket. Here's an example that will let us
access the failing bucket by creating a client with anonymous credentials.


``` r
s3_client <- paws.storage::s3(
  config = list(
    credentials = list(anonymous = TRUE),
    region = "auto",
    endpoint = "https://uk1s3.embassy.ebi.ac.uk"
  )
)
```

If you're accessing a public bucket, the most important step is to provide a
`credentials` list with `anonymous = TRUE`.  Doing so ensures that no attempts
to find other credentials are made, and prevents the problems seen above.  If
you're using files on Amazon AWS storage you'll need to set the `region` to
whatever is appropriate for your data e.g. `"us-east-2"`, `"eu-west-3"`, etc.
For other S3 providers that don't have regions use the value `"auto"` as in the
example below.  Finally the `endpoint` argument is the full hostname of the
server where your files can be found.  For more information on creating an S3
client see the [**paws.storage**
documentation](https://paws-r.github.io/docs/s3/).

We can then pass our s3_client to `zarr_overview()` and it now works
successfully.


``` r
zarr_overview(s3_address, s3_client = s3_client)
```

```
## Type: Array
## Path: https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0076A/10501752.zarr/0
## Shape: 50 x 494 x 464
## Chunk Shape: 1 x 494 x 464
## No. of Chunks: 50 (50 x 1 x 1)
## Data Type: float64
## Endianness: little
## Compressor: blosc
```

Most functions in **Rarr** have the `s3_client` argument and it
can be applied in the same way.

---
title: "Working with Zarr arrays in R"
date: "`r BiocStyle::doc_date()`"
author:
  - name: Mike L. Smith
affiliation:
  - EMBL Heidelberg
email: mike.smith@embl.de
package: "`r BiocStyle::pkg_ver('Rarr')`"
vignette: >
  %\VignetteIndexEntry{"Working with Zarr arrays in R"}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output:
  BiocStyle::html_document
---

```{css, echo=FALSE}
.main-container {
  max-width: 1600px;
  margin-left: auto;
  margin-right: auto;
}
```

# Introduction

The Zarr specification defines a format for chunked, compressed, N-dimensional
arrays.  It's design allows efficient access to subsets of the stored array, and
supports both local and cloud storage systems. Zarr is experiencing increasing
adoption in a number of scientific fields, where multi-dimensional data are
prevalent. In particular as a back-end to the The Open Microscopy Environment's
[OME-NGFF format](https://ngff.openmicroscopy.org/latest/) for storing
bioimaging data in the cloud.

**Rarr** is intended to be a simple interface to reading and writing individual
Zarr arrays. It is developed
in R and C with no reliance on external libraries or APIs for interfacing with
the Zarr arrays. Additional compression libraries (e.g. blosc) are bundled with
**Rarr** to provide support for datasets compressed using these tools.

## Limitations with **Rarr**

If you know about Zarr arrays already, you'll probably be aware they can be
stored in hierarchical groups, where additional meta data can explain the
relationship between the arrays.  Currently, **Rarr** is not designed to be
aware of these hierarchical Zarr array collections. However, the component
arrays can be read individually by providing the path to them directly.

Currently, there are also limitations on the Zarr datatypes that can be accessed
using **Rarr**.  For now most numeric types can be read into R, although in some
instances e.g. 64-bit integers there is potential for loss of information.

## Example data

The are some example Zarr arrays included with the package.  These were created
using the Zarr Python implementation and are primarily intended for testing the
functionality of **Rarr**.  You can use the code below to list the complete set
on your system, however it's a long list so we don't show the output here.

```{r, local-examples, echo = TRUE, results = "hide"}
list.dirs(
  system.file("extdata", "zarr_examples", package = "Rarr"),
  recursive = TRUE
) |>
  grep(pattern = "zarr$", value = TRUE)
```

# Quick start guide

```{r, child=system.file("rmd", "quick_start.Rmd", package="Rarr")}
```

# Additional details

## Working with Zarr metadata

By default the `zarr_overview()` function prints a summary of the array to
screen, so you can get a quick idea of the array properties.  However, there are
times when it might be useful to compute on that metadata, in which case
printing to screen isn't very helpful.  If his is the case the function also has
the argument `as_data_frame` which toggles whether to print the output to
screen, as seen before above, or to return a `data.frame` containing the array
details.

```{r, overview-2}
zarr_overview(zarr_example, as_data_frame = TRUE)
```

## Writing subsets of data

One of the key features of the Zarr specification is that the arrays are
chunked, allowing rapid access to the required data without needed to read or
write everything else.  If you want to modify a subset of a Zarr array, it is
very inefficient to write all chunks to disk, which is what `write_zarr_array()`
does.  Instead, **Rarr** provides two functions for reducing the amount of
writing required if the circumstances allow: `create_empty_zarr_array()` and
`update_zarr_array()`.

### Creating an "empty" array

Despite the name, you can actually think of `create_empty_zarr_array()` as
creating an array where all the values are the same.  The Zarr specification
allows for "uninitialized" chunks, which are not actually present on disk.  In
this case, any reading application assumes the entirety of the chunk is filled
with a single value, which is found in the array metadata.  This allows for very
efficient creation of the new array, since only a small metadata file is
actually written.  However it is necessary to provide some additional details,
such as the shape of the array, since there's no R array to infer these from.
Let's look at an example:

```{r, create-empty-array}
path <- tempfile()
create_empty_zarr_array(
  zarr_array_path = path,
  dim = c(50, 20),
  chunk_dim = c(5, 10),
  data_type = "integer",
  fill_value = 7L
)
```

First we have to provide a location for the array to be created using the
`zarr_array_path` argument.  Then we provide the dimensions of the new array,
and the shape of the chunks it should be split into.  These two arguments must
be compatible with one another i.e. have the same number of dimensions and no
value in `chunk_dim` should exceed the corresponding value in `dim`.  The
`data_type` argument defines what type of values will be stored in the array.
Finally we use the `fill_value` argument to provide our default value for the
uninitialized chunks.
The next few lines check what's actually been created on our file system.
First, we use `list.files()` to confirm that that only file that's been created
is the `zarr.json` metadata; there are no chunk files.  Then we use `table()` to
check the contents of the array, and confirm that when it's read the resulting
array in R is full of 7s, our fill value.

```{r, check-empty-array}
list.files(path, all.files = TRUE, no.. = TRUE)
table(read_zarr_array(path))
```

### Updating a subset of an existing array

Lets assume we want to update the first row of our array to contain the sequence
of integers from 1 to 20.  In the code below we create an example vector
containing the new data.  We then use `update_zarr_array()`, passing the
location of the Zarr and the new values to be inserted.  Finally, we provide the
`index` argument which defines which elements in the Zarr array should be
updated.  It's important that the shape and number of values in `x` corresponds
to the total count of points in the Zarr array we want to update e.g. in this
case we're updating a single row of 20 values.

```{r, update-array}
x <- 1:20
update_zarr_array(
  zarr_array_path = path,
  x = x,
  index = list(1, 1:20)
)
```

As before, we can take a look at what's happened on disk and confirm the values
are present in the array if we read it into R.

```{r, check-update}
list.files(path, all.files = TRUE, no.. = TRUE)
read_zarr_array(path, index = list(1:2, 1:5))
table(read_zarr_array(path))
```

Here `list.files()` confirms that there's now two chunk files that have been
created.  When we first created the Zarr we specified that the chunks should be
10 columns wide, so by modifying 20 columns we'd expect at least two chunks to
be realized on disk.  We use `read_zarr_array()` to confirm visually that the
first row contains our sequence of values, whilst the second row is still all 7.
We use `table()` to confirm that the total contents is as expected.

# Appendix

## Session info {.unnumbered}
```{r, session-info, echo = FALSE}
sessionInfo()
```

# Authenticated S3 access works

    Code
      zarr_overview("https://s3.embl.de/rarr-testing/bzip2.zarr")
    Output
      Type: Array
      Path: https://s3.embl.de/rarr-testing/bzip2.zarr
      Shape: 20 x 10
      Chunk Shape: 10 x 10
      No. of Chunks: 2 (2 x 1)
      Data Type: int32
      Endianness: little
      Compressor: bz2

---

    Code
      read_zarr_array("https://s3.embl.de/rarr-testing/bzip2.zarr")
    Output
            [,1] [,2] [,3] [,4] [,5] [,6] [,7] [,8] [,9] [,10]
       [1,]    1    2    3    4    5    6    7    8    9    10
       [2,]    4    0    0    0    0    0    0    0    0     0
       [3,]    7    0    0    0    0    0    0    0    0     0
       [4,]   10    0    0    0    0    0    0    0    0     0
       [5,]   13    0    0    0    0    0    0    0    0     0
       [6,]   16    0    0    0    0    0    0    0    0     0
       [7,]   19    0    0    0    0    0    0    0    0     0
       [8,]   22    0    0    0    0    0    0    0    0     0
       [9,]   25    0    0    0    0    0    0    0    0     0
      [10,]   28    0    0    0    0    0    0    0    0     0
      [11,]   31    0    0    0    0    0    0    0    0     0
      [12,]   34    0    0    0    0    0    0    0    0     0
      [13,]   37    0    0    0    0    0    0    0    0     0
      [14,]   40    0    0    0    0    0    0    0    0     0
      [15,]   43    0    0    0    0    0    0    0    0     0
      [16,]   46    0    0    0    0    0    0    0    0     0
      [17,]   49    0    0    0    0    0    0    0    0     0
      [18,]   52    0    0    0    0    0    0    0    0     0
      [19,]   55    0    0    0    0    0    0    0    0     0
      [20,]   58    0    0    0    0    0    0    0    0     0

# JSON metadata is not unboxed

    Code
      zarr_overview(
        "https://livingobjects.ebi.ac.uk/idr/share/ome2024-ngff-challenge/idr0066/ExpA_VIP_ASLM_on_XZ_slice1088.zarr/0")
    Output
      Type: Array
      Path: https://livingobjects.ebi.ac.uk/idr/share/ome2024-ngff-challenge/idr0066/ExpA_VIP_ASLM_on_XZ_slice1088.zarr/0
      Shape: 1478 x 2048
      Chunk Shape: 1536 x 2048
      No. of Chunks: 1 (1 x 1)
      Data Type: uint16
      Endianness: NA
      Compressor: None
      Attributes: yes

# Empty zarr arrays have expected metadata

    Code
      jsonlite::read_json(file.path(empty_zarr, ".zarray"))
    Output
      $shape
      $shape[[1]]
      [1] 10
      
      $shape[[2]]
      [1] 30
      
      
      $chunks
      $chunks[[1]]
      [1] 5
      
      $chunks[[2]]
      [1] 3
      
      
      $dtype
      [1] "<i4"
      
      $fill_value
      [1] 0
      
      $dimension_separator
      [1] "."
      
      $order
      [1] "F"
      
      $zarr_format
      [1] 2
      
      $filters
      NULL
      
      $compressor
      $compressor$id
      [1] "blosc"
      
      $compressor$cname
      [1] "blosclz"
      
      $compressor$clevel
      [1] 9
      
      $compressor$shuffle
      [1] "bitshuffle"
      
      $compressor$typesize
      [1] 4
      
      $compressor$blocksize
      [1] 0
      
      

# zarr_consolidate_metadata() on v2

    Code
      zarr_consolidate_metadata(zarr_v2, action = "return")
    Output
      $zarr_consolidated_format
      [1] 1
      
      $metadata
      $metadata$.zgroup
      $metadata$.zgroup$zarr_format
      $metadata$.zgroup$zarr_format[[1]]
      [1] 2
      
      
      
      $metadata$`array1/.zarray`
      $metadata$`array1/.zarray`$shape
      $metadata$`array1/.zarray`$shape[[1]]
      [1] 2
      
      $metadata$`array1/.zarray`$shape[[2]]
      [1] 2
      
      
      $metadata$`array1/.zarray`$chunks
      $metadata$`array1/.zarray`$chunks[[1]]
      [1] 1
      
      $metadata$`array1/.zarray`$chunks[[2]]
      [1] 2
      
      
      $metadata$`array1/.zarray`$dtype
      [1] "|u1"
      
      $metadata$`array1/.zarray`$fill_value
      [1] 0
      
      $metadata$`array1/.zarray`$dimension_separator
      [1] "."
      
      $metadata$`array1/.zarray`$order
      [1] "F"
      
      $metadata$`array1/.zarray`$zarr_format
      [1] 2
      
      $metadata$`array1/.zarray`$filters
      NULL
      
      $metadata$`array1/.zarray`$compressor
      $metadata$`array1/.zarray`$compressor$id
      [1] "zstd"
      
      $metadata$`array1/.zarray`$compressor$level
      [1] 0
      
      
      
      $metadata$`array1/.zattrs`
      $metadata$`array1/.zattrs`$description
      [1] "This is array 1"
      
      
      $metadata$`array2/.zarray`
      $metadata$`array2/.zarray`$shape
      $metadata$`array2/.zarray`$shape[[1]]
      [1] 2
      
      $metadata$`array2/.zarray`$shape[[2]]
      [1] 2
      
      
      $metadata$`array2/.zarray`$chunks
      $metadata$`array2/.zarray`$chunks[[1]]
      [1] 1
      
      $metadata$`array2/.zarray`$chunks[[2]]
      [1] 2
      
      
      $metadata$`array2/.zarray`$dtype
      [1] "<f8"
      
      $metadata$`array2/.zarray`$fill_value
      [1] 0
      
      $metadata$`array2/.zarray`$dimension_separator
      [1] "."
      
      $metadata$`array2/.zarray`$order
      [1] "F"
      
      $metadata$`array2/.zarray`$zarr_format
      [1] 2
      
      $metadata$`array2/.zarray`$filters
      NULL
      
      $metadata$`array2/.zarray`$compressor
      $metadata$`array2/.zarray`$compressor$id
      [1] "zstd"
      
      $metadata$`array2/.zarray`$compressor$level
      [1] 0
      
      
      
      

# zarr_consoliate_metadata() on v3

    Code
      zarr_consolidate_metadata(zarr_v3, action = "return")
    Output
      $zarr_format
      [1] 3
      
      $node_type
      [1] "group"
      
      $attributes
      list()
      
      $consolidated_metadata
      $consolidated_metadata$kind
      [1] "inline"
      
      $consolidated_metadata$must_understand
      [1] FALSE
      
      $consolidated_metadata$metadata
      $consolidated_metadata$metadata$array1
      $consolidated_metadata$metadata$array1$node_type
      [1] "array"
      
      $consolidated_metadata$metadata$array1$zarr_format
      [1] 3
      
      $consolidated_metadata$metadata$array1$data_type
      [1] "uint8"
      
      $consolidated_metadata$metadata$array1$shape
      $consolidated_metadata$metadata$array1$shape[[1]]
      [1] 2
      
      $consolidated_metadata$metadata$array1$shape[[2]]
      [1] 2
      
      
      $consolidated_metadata$metadata$array1$chunk_grid
      $consolidated_metadata$metadata$array1$chunk_grid$name
      [1] "regular"
      
      $consolidated_metadata$metadata$array1$chunk_grid$configuration
      $consolidated_metadata$metadata$array1$chunk_grid$configuration$chunk_shape
      $consolidated_metadata$metadata$array1$chunk_grid$configuration$chunk_shape[[1]]
      [1] 1
      
      $consolidated_metadata$metadata$array1$chunk_grid$configuration$chunk_shape[[2]]
      [1] 2
      
      
      
      
      $consolidated_metadata$metadata$array1$chunk_key_encoding
      $consolidated_metadata$metadata$array1$chunk_key_encoding$name
      [1] "default"
      
      $consolidated_metadata$metadata$array1$chunk_key_encoding$configuration
      $consolidated_metadata$metadata$array1$chunk_key_encoding$configuration$separator
      [1] "/"
      
      
      
      $consolidated_metadata$metadata$array1$fill_value
      [1] 0
      
      $consolidated_metadata$metadata$array1$codecs
      $consolidated_metadata$metadata$array1$codecs[[1]]
      $consolidated_metadata$metadata$array1$codecs[[1]]$name
      [1] "transpose"
      
      $consolidated_metadata$metadata$array1$codecs[[1]]$configuration
      $consolidated_metadata$metadata$array1$codecs[[1]]$configuration$order
      $consolidated_metadata$metadata$array1$codecs[[1]]$configuration$order[[1]]
      [1] 1
      
      $consolidated_metadata$metadata$array1$codecs[[1]]$configuration$order[[2]]
      [1] 0
      
      
      
      
      $consolidated_metadata$metadata$array1$codecs[[2]]
      $consolidated_metadata$metadata$array1$codecs[[2]]$name
      [1] "bytes"
      
      $consolidated_metadata$metadata$array1$codecs[[2]]$configuration
      $consolidated_metadata$metadata$array1$codecs[[2]]$configuration$endian
      NULL
      
      
      
      $consolidated_metadata$metadata$array1$codecs[[3]]
      $consolidated_metadata$metadata$array1$codecs[[3]]$name
      [1] "zstd"
      
      $consolidated_metadata$metadata$array1$codecs[[3]]$configuration
      $consolidated_metadata$metadata$array1$codecs[[3]]$configuration$level
      [1] 0
      
      
      
      
      $consolidated_metadata$metadata$array1$attributes
      $consolidated_metadata$metadata$array1$attributes$description
      [1] "This is array 1"
      
      
      
      $consolidated_metadata$metadata$array2
      $consolidated_metadata$metadata$array2$node_type
      [1] "array"
      
      $consolidated_metadata$metadata$array2$zarr_format
      [1] 3
      
      $consolidated_metadata$metadata$array2$data_type
      [1] "float64"
      
      $consolidated_metadata$metadata$array2$shape
      $consolidated_metadata$metadata$array2$shape[[1]]
      [1] 2
      
      $consolidated_metadata$metadata$array2$shape[[2]]
      [1] 2
      
      
      $consolidated_metadata$metadata$array2$chunk_grid
      $consolidated_metadata$metadata$array2$chunk_grid$name
      [1] "regular"
      
      $consolidated_metadata$metadata$array2$chunk_grid$configuration
      $consolidated_metadata$metadata$array2$chunk_grid$configuration$chunk_shape
      $consolidated_metadata$metadata$array2$chunk_grid$configuration$chunk_shape[[1]]
      [1] 1
      
      $consolidated_metadata$metadata$array2$chunk_grid$configuration$chunk_shape[[2]]
      [1] 2
      
      
      
      
      $consolidated_metadata$metadata$array2$chunk_key_encoding
      $consolidated_metadata$metadata$array2$chunk_key_encoding$name
      [1] "default"
      
      $consolidated_metadata$metadata$array2$chunk_key_encoding$configuration
      $consolidated_metadata$metadata$array2$chunk_key_encoding$configuration$separator
      [1] "/"
      
      
      
      $consolidated_metadata$metadata$array2$fill_value
      [1] 0
      
      $consolidated_metadata$metadata$array2$codecs
      $consolidated_metadata$metadata$array2$codecs[[1]]
      $consolidated_metadata$metadata$array2$codecs[[1]]$name
      [1] "transpose"
      
      $consolidated_metadata$metadata$array2$codecs[[1]]$configuration
      $consolidated_metadata$metadata$array2$codecs[[1]]$configuration$order
      $consolidated_metadata$metadata$array2$codecs[[1]]$configuration$order[[1]]
      [1] 1
      
      $consolidated_metadata$metadata$array2$codecs[[1]]$configuration$order[[2]]
      [1] 0
      
      
      
      
      $consolidated_metadata$metadata$array2$codecs[[2]]
      $consolidated_metadata$metadata$array2$codecs[[2]]$name
      [1] "bytes"
      
      $consolidated_metadata$metadata$array2$codecs[[2]]$configuration
      $consolidated_metadata$metadata$array2$codecs[[2]]$configuration$endian
      [1] "little"
      
      
      
      $consolidated_metadata$metadata$array2$codecs[[3]]
      $consolidated_metadata$metadata$array2$codecs[[3]]$name
      [1] "zstd"
      
      $consolidated_metadata$metadata$array2$codecs[[3]]$configuration
      $consolidated_metadata$metadata$array2$codecs[[3]]$configuration$level
      [1] 0
      
      
      
      
      
      $consolidated_metadata$metadata$.
      $consolidated_metadata$metadata$.$zarr_format
      [1] 3
      
      $consolidated_metadata$metadata$.$node_type
      [1] "group"
      
      
      
      

# zarr_overview console output matches snapshot for single array

    Code
      zarr_overview(zarr_c, as_data_frame = FALSE)
    Output
      Type: Array
      Path: <path>
      Shape: 30 x 20 x 10
      Chunk Shape: 10 x 10 x 5
      No. of Chunks: 12 (3 x 2 x 2)
      Data Type: int64
      Endianness: little
      Compressor: zstd
      Attributes: no

# zarr_overview works with consolidated metadata store

    Code
      metadata_df
    Output
                                                        path          data_type
      1                           /conditions/geometry/angle fixed_length_utf32
      2                            /conditions/geometry/band fixed_length_utf32
      3                        /conditions/geometry/detector              int64
      4                 /conditions/geometry/mean_sun_angles            float64
      5   /conditions/geometry/mean_viewing_incidence_angles            float64
      6                      /conditions/geometry/sun_angles            float64
      7        /conditions/geometry/viewing_incidence_angles            float64
      8                               /conditions/geometry/x              int64
      9                               /conditions/geometry/y              int64
      10        /conditions/mask/detector_footprint/r10m/b02              uint8
      11        /conditions/mask/detector_footprint/r10m/b03              uint8
      12        /conditions/mask/detector_footprint/r10m/b04              uint8
      13        /conditions/mask/detector_footprint/r10m/b08              uint8
      14          /conditions/mask/detector_footprint/r10m/x              int64
      15          /conditions/mask/detector_footprint/r10m/y              int64
      16        /conditions/mask/detector_footprint/r20m/b05              uint8
      17        /conditions/mask/detector_footprint/r20m/b06              uint8
      18        /conditions/mask/detector_footprint/r20m/b07              uint8
      19        /conditions/mask/detector_footprint/r20m/b11              uint8
      20        /conditions/mask/detector_footprint/r20m/b12              uint8
      21        /conditions/mask/detector_footprint/r20m/b8a              uint8
      22          /conditions/mask/detector_footprint/r20m/x              int64
      23          /conditions/mask/detector_footprint/r20m/y              int64
      24        /conditions/mask/detector_footprint/r60m/b01              uint8
      25        /conditions/mask/detector_footprint/r60m/b09              uint8
      26        /conditions/mask/detector_footprint/r60m/b10              uint8
      27          /conditions/mask/detector_footprint/r60m/x              int64
      28          /conditions/mask/detector_footprint/r60m/y              int64
      29        /conditions/mask/l1c_classification/r60m/b00              uint8
      30          /conditions/mask/l1c_classification/r60m/x              int64
      31          /conditions/mask/l1c_classification/r60m/y              int64
      32        /conditions/mask/l2a_classification/r20m/scl              uint8
      33          /conditions/mask/l2a_classification/r20m/x              int64
      34          /conditions/mask/l2a_classification/r20m/y              int64
      35        /conditions/mask/l2a_classification/r60m/scl              uint8
      36          /conditions/mask/l2a_classification/r60m/x              int64
      37          /conditions/mask/l2a_classification/r60m/y              int64
      38                /conditions/meteorology/cams/aod1240            float32
      39                 /conditions/meteorology/cams/aod469            float32
      40                 /conditions/meteorology/cams/aod550            float32
      41                 /conditions/meteorology/cams/aod670            float32
      42                 /conditions/meteorology/cams/aod865            float32
      43               /conditions/meteorology/cams/bcaod550            float32
      44               /conditions/meteorology/cams/duaod550            float32
      45          /conditions/meteorology/cams/isobaricInhPa            float64
      46               /conditions/meteorology/cams/latitude            float64
      47              /conditions/meteorology/cams/longitude            float64
      48                 /conditions/meteorology/cams/number              int64
      49               /conditions/meteorology/cams/omaod550            float32
      50               /conditions/meteorology/cams/ssaod550            float32
      51                   /conditions/meteorology/cams/step              int64
      52               /conditions/meteorology/cams/suaod550            float32
      53                /conditions/meteorology/cams/surface            float64
      54                   /conditions/meteorology/cams/time              int64
      55             /conditions/meteorology/cams/valid_time              int64
      56                      /conditions/meteorology/cams/z            float32
      57         /conditions/meteorology/ecmwf/isobaricInhPa            float64
      58              /conditions/meteorology/ecmwf/latitude            float64
      59             /conditions/meteorology/ecmwf/longitude            float64
      60                   /conditions/meteorology/ecmwf/msl            float32
      61                /conditions/meteorology/ecmwf/number              int64
      62                     /conditions/meteorology/ecmwf/r            float32
      63                  /conditions/meteorology/ecmwf/step              int64
      64               /conditions/meteorology/ecmwf/surface            float64
      65                  /conditions/meteorology/ecmwf/tco3            float32
      66                  /conditions/meteorology/ecmwf/tcwv            float32
      67                  /conditions/meteorology/ecmwf/time              int64
      68                   /conditions/meteorology/ecmwf/u10            float32
      69                   /conditions/meteorology/ecmwf/v10            float32
      70            /conditions/meteorology/ecmwf/valid_time              int64
      71                  /measurements/reflectance/r10m/b02             uint16
      72                  /measurements/reflectance/r10m/b03             uint16
      73                  /measurements/reflectance/r10m/b04             uint16
      74                  /measurements/reflectance/r10m/b08             uint16
      75                    /measurements/reflectance/r10m/x              int64
      76                    /measurements/reflectance/r10m/y              int64
      77                  /measurements/reflectance/r20m/b01             uint16
      78                  /measurements/reflectance/r20m/b02             uint16
      79                  /measurements/reflectance/r20m/b03             uint16
      80                  /measurements/reflectance/r20m/b04             uint16
      81                  /measurements/reflectance/r20m/b05             uint16
      82                  /measurements/reflectance/r20m/b06             uint16
      83                  /measurements/reflectance/r20m/b07             uint16
      84                  /measurements/reflectance/r20m/b11             uint16
      85                  /measurements/reflectance/r20m/b12             uint16
      86                  /measurements/reflectance/r20m/b8a             uint16
      87                    /measurements/reflectance/r20m/x              int64
      88                    /measurements/reflectance/r20m/y              int64
      89                  /measurements/reflectance/r60m/b01             uint16
      90                  /measurements/reflectance/r60m/b02             uint16
      91                  /measurements/reflectance/r60m/b03             uint16
      92                  /measurements/reflectance/r60m/b04             uint16
      93                  /measurements/reflectance/r60m/b05             uint16
      94                  /measurements/reflectance/r60m/b06             uint16
      95                  /measurements/reflectance/r60m/b07             uint16
      96                  /measurements/reflectance/r60m/b09             uint16
      97                  /measurements/reflectance/r60m/b11             uint16
      98                  /measurements/reflectance/r60m/b12             uint16
      99                  /measurements/reflectance/r60m/b8a             uint16
      100                   /measurements/reflectance/r60m/x              int64
      101                   /measurements/reflectance/r60m/y              int64
      102                       /quality/atmosphere/r10m/aot             uint16
      103                       /quality/atmosphere/r10m/wvp             uint16
      104                         /quality/atmosphere/r10m/x              int64
      105                         /quality/atmosphere/r10m/y              int64
      106                       /quality/atmosphere/r20m/aot             uint16
      107                       /quality/atmosphere/r20m/wvp             uint16
      108                         /quality/atmosphere/r20m/x              int64
      109                         /quality/atmosphere/r20m/y              int64
      110                       /quality/atmosphere/r60m/aot             uint16
      111                       /quality/atmosphere/r60m/wvp             uint16
      112                         /quality/atmosphere/r60m/x              int64
      113                         /quality/atmosphere/r60m/y              int64
      114                   /quality/l2a_quicklook/r10m/band              int64
      115                    /quality/l2a_quicklook/r10m/tci              uint8
      116                      /quality/l2a_quicklook/r10m/x              int64
      117                      /quality/l2a_quicklook/r10m/y              int64
      118                   /quality/l2a_quicklook/r20m/band              int64
      119                    /quality/l2a_quicklook/r20m/tci              uint8
      120                      /quality/l2a_quicklook/r20m/x              int64
      121                      /quality/l2a_quicklook/r20m/y              int64
      122                   /quality/l2a_quicklook/r60m/band              int64
      123                    /quality/l2a_quicklook/r60m/tci              uint8
      124                      /quality/l2a_quicklook/r60m/x              int64
      125                      /quality/l2a_quicklook/r60m/y              int64
      126                             /quality/mask/r10m/b02              uint8
      127                             /quality/mask/r10m/b03              uint8
      128                             /quality/mask/r10m/b04              uint8
      129                             /quality/mask/r10m/b08              uint8
      130                               /quality/mask/r10m/x              int64
      131                               /quality/mask/r10m/y              int64
      132                             /quality/mask/r20m/b05              uint8
      133                             /quality/mask/r20m/b06              uint8
      134                             /quality/mask/r20m/b07              uint8
      135                             /quality/mask/r20m/b11              uint8
      136                             /quality/mask/r20m/b12              uint8
      137                             /quality/mask/r20m/b8a              uint8
      138                               /quality/mask/r20m/x              int64
      139                               /quality/mask/r20m/y              int64
      140                             /quality/mask/r60m/b01              uint8
      141                             /quality/mask/r60m/b09              uint8
      142                             /quality/mask/r60m/b10              uint8
      143                               /quality/mask/r60m/x              int64
      144                               /quality/mask/r60m/y              int64
      145                     /quality/probability/r20m/band              int64
      146                      /quality/probability/r20m/cld              uint8
      147                      /quality/probability/r20m/snw              uint8
      148                        /quality/probability/r20m/x              int64
      149                        /quality/probability/r20m/y              int64
          endianness compressor              dim       chunk_dim       nchunks
      1       little      blosc                2               2             1
      2       little      blosc               13              13             1
      3       little      blosc                6               6             1
      4       little      blosc                2               2             1
      5       little      blosc            13, 2           13, 2          1, 1
      6       little      blosc        2, 23, 23       2, 23, 23       1, 1, 1
      7       little      blosc 13, 6, 2, 23, 23 7, 3, 2, 23, 23 2, 2, 1, 1, 1
      8       little      blosc               23              23             1
      9       little      blosc               23              23             1
      10        <NA>      blosc     10980, 10980      1830, 1830          6, 6
      11        <NA>      blosc     10980, 10980      1830, 1830          6, 6
      12        <NA>      blosc     10980, 10980      1830, 1830          6, 6
      13        <NA>      blosc     10980, 10980      1830, 1830          6, 6
      14      little      blosc            10980           10980             1
      15      little      blosc            10980           10980             1
      16        <NA>      blosc       5490, 5490        915, 915          6, 6
      17        <NA>      blosc       5490, 5490        915, 915          6, 6
      18        <NA>      blosc       5490, 5490        915, 915          6, 6
      19        <NA>      blosc       5490, 5490        915, 915          6, 6
      20        <NA>      blosc       5490, 5490        915, 915          6, 6
      21        <NA>      blosc       5490, 5490        915, 915          6, 6
      22      little      blosc             5490            5490             1
      23      little      blosc             5490            5490             1
      24        <NA>      blosc       1830, 1830        305, 305          6, 6
      25        <NA>      blosc       1830, 1830        305, 305          6, 6
      26        <NA>      blosc       1830, 1830        305, 305          6, 6
      27      little      blosc             1830            1830             1
      28      little      blosc             1830            1830             1
      29        <NA>      blosc       1830, 1830        305, 305          6, 6
      30      little      blosc             1830            1830             1
      31      little      blosc             1830            1830             1
      32        <NA>      blosc       5490, 5490        915, 915          6, 6
      33      little      blosc             5490            5490             1
      34      little      blosc             5490            5490             1
      35        <NA>      blosc       1830, 1830        305, 305          6, 6
      36      little      blosc             1830            1830             1
      37      little      blosc             1830            1830             1
      38      little      blosc             9, 9            9, 9          1, 1
      39      little      blosc             9, 9            9, 9          1, 1
      40      little      blosc             9, 9            9, 9          1, 1
      41      little      blosc             9, 9            9, 9          1, 1
      42      little      blosc             9, 9            9, 9          1, 1
      43      little      blosc             9, 9            9, 9          1, 1
      44      little      blosc             9, 9            9, 9          1, 1
      45      little       <NA>                1               1             1
      46      little      blosc                9               9             1
      47      little      blosc                9               9             1
      48      little       <NA>                1               1             1
      49      little      blosc             9, 9            9, 9          1, 1
      50      little      blosc             9, 9            9, 9          1, 1
      51      little       <NA>                1               1             1
      52      little      blosc             9, 9            9, 9          1, 1
      53      little       <NA>                1               1             1
      54      little       <NA>                1               1             1
      55      little       <NA>                1               1             1
      56      little      blosc             9, 9            9, 9          1, 1
      57      little       <NA>                1               1             1
      58      little      blosc                9               9             1
      59      little      blosc                9               9             1
      60      little      blosc             9, 9            9, 9          1, 1
      61      little       <NA>                1               1             1
      62      little      blosc             9, 9            9, 9          1, 1
      63      little       <NA>                1               1             1
      64      little       <NA>                1               1             1
      65      little      blosc             9, 9            9, 9          1, 1
      66      little      blosc             9, 9            9, 9          1, 1
      67      little       <NA>                1               1             1
      68      little      blosc             9, 9            9, 9          1, 1
      69      little      blosc             9, 9            9, 9          1, 1
      70      little       <NA>                1               1             1
      71      little      blosc     10980, 10980      1830, 1830          6, 6
      72      little      blosc     10980, 10980      1830, 1830          6, 6
      73      little      blosc     10980, 10980      1830, 1830          6, 6
      74      little      blosc     10980, 10980      1830, 1830          6, 6
      75      little      blosc            10980           10980             1
      76      little      blosc            10980           10980             1
      77      little      blosc       5490, 5490        915, 915          6, 6
      78      little      blosc       5490, 5490        915, 915          6, 6
      79      little      blosc       5490, 5490        915, 915          6, 6
      80      little      blosc       5490, 5490        915, 915          6, 6
      81      little      blosc       5490, 5490        915, 915          6, 6
      82      little      blosc       5490, 5490        915, 915          6, 6
      83      little      blosc       5490, 5490        915, 915          6, 6
      84      little      blosc       5490, 5490        915, 915          6, 6
      85      little      blosc       5490, 5490        915, 915          6, 6
      86      little      blosc       5490, 5490        915, 915          6, 6
      87      little      blosc             5490            5490             1
      88      little      blosc             5490            5490             1
      89      little      blosc       1830, 1830        305, 305          6, 6
      90      little      blosc       1830, 1830        305, 305          6, 6
      91      little      blosc       1830, 1830        305, 305          6, 6
      92      little      blosc       1830, 1830        305, 305          6, 6
      93      little      blosc       1830, 1830        305, 305          6, 6
      94      little      blosc       1830, 1830        305, 305          6, 6
      95      little      blosc       1830, 1830        305, 305          6, 6
      96      little      blosc       1830, 1830        305, 305          6, 6
      97      little      blosc       1830, 1830        305, 305          6, 6
      98      little      blosc       1830, 1830        305, 305          6, 6
      99      little      blosc       1830, 1830        305, 305          6, 6
      100     little      blosc             1830            1830             1
      101     little      blosc             1830            1830             1
      102     little      blosc     10980, 10980      1830, 1830          6, 6
      103     little      blosc     10980, 10980      1830, 1830          6, 6
      104     little      blosc            10980           10980             1
      105     little      blosc            10980           10980             1
      106     little      blosc       5490, 5490        915, 915          6, 6
      107     little      blosc       5490, 5490        915, 915          6, 6
      108     little      blosc             5490            5490             1
      109     little      blosc             5490            5490             1
      110     little      blosc       1830, 1830        305, 305          6, 6
      111     little      blosc       1830, 1830        305, 305          6, 6
      112     little      blosc             1830            1830             1
      113     little      blosc             1830            1830             1
      114     little      blosc                3               3             1
      115       <NA>      blosc  3, 10980, 10980   1, 1830, 1830       3, 6, 6
      116     little      blosc            10980           10980             1
      117     little      blosc            10980           10980             1
      118     little      blosc                3               3             1
      119       <NA>      blosc    3, 5490, 5490     1, 915, 915       3, 6, 6
      120     little      blosc             5490            5490             1
      121     little      blosc             5490            5490             1
      122     little      blosc                3               3             1
      123       <NA>      blosc    3, 1830, 1830     1, 305, 305       3, 6, 6
      124     little      blosc             1830            1830             1
      125     little      blosc             1830            1830             1
      126       <NA>      blosc     10980, 10980      1830, 1830          6, 6
      127       <NA>      blosc     10980, 10980      1830, 1830          6, 6
      128       <NA>      blosc     10980, 10980      1830, 1830          6, 6
      129       <NA>      blosc     10980, 10980      1830, 1830          6, 6
      130     little      blosc            10980           10980             1
      131     little      blosc            10980           10980             1
      132       <NA>      blosc       5490, 5490        915, 915          6, 6
      133       <NA>      blosc       5490, 5490        915, 915          6, 6
      134       <NA>      blosc       5490, 5490        915, 915          6, 6
      135       <NA>      blosc       5490, 5490        915, 915          6, 6
      136       <NA>      blosc       5490, 5490        915, 915          6, 6
      137       <NA>      blosc       5490, 5490        915, 915          6, 6
      138     little      blosc             5490            5490             1
      139     little      blosc             5490            5490             1
      140       <NA>      blosc       1830, 1830        305, 305          6, 6
      141       <NA>      blosc       1830, 1830        305, 305          6, 6
      142       <NA>      blosc       1830, 1830        305, 305          6, 6
      143     little      blosc             1830            1830             1
      144     little      blosc             1830            1830             1
      145     little       <NA>                1               1             1
      146       <NA>      blosc       5490, 5490        915, 915          6, 6
      147       <NA>      blosc       5490, 5490        915, 915          6, 6
      148     little      blosc             5490            5490             1
      149     little      blosc             5490            5490             1
          attributes
      1         TRUE
      2         TRUE
      3         TRUE
      4         TRUE
      5         TRUE
      6         TRUE
      7         TRUE
      8         TRUE
      9         TRUE
      10        TRUE
      11        TRUE
      12        TRUE
      13        TRUE
      14        TRUE
      15        TRUE
      16        TRUE
      17        TRUE
      18        TRUE
      19        TRUE
      20        TRUE
      21        TRUE
      22        TRUE
      23        TRUE
      24        TRUE
      25        TRUE
      26        TRUE
      27        TRUE
      28        TRUE
      29        TRUE
      30        TRUE
      31        TRUE
      32        TRUE
      33        TRUE
      34        TRUE
      35        TRUE
      36        TRUE
      37        TRUE
      38        TRUE
      39        TRUE
      40        TRUE
      41        TRUE
      42        TRUE
      43        TRUE
      44        TRUE
      45        TRUE
      46        TRUE
      47        TRUE
      48        TRUE
      49        TRUE
      50        TRUE
      51        TRUE
      52        TRUE
      53        TRUE
      54        TRUE
      55        TRUE
      56        TRUE
      57        TRUE
      58        TRUE
      59        TRUE
      60        TRUE
      61        TRUE
      62        TRUE
      63        TRUE
      64        TRUE
      65        TRUE
      66        TRUE
      67        TRUE
      68        TRUE
      69        TRUE
      70        TRUE
      71        TRUE
      72        TRUE
      73        TRUE
      74        TRUE
      75        TRUE
      76        TRUE
      77        TRUE
      78        TRUE
      79        TRUE
      80        TRUE
      81        TRUE
      82        TRUE
      83        TRUE
      84        TRUE
      85        TRUE
      86        TRUE
      87        TRUE
      88        TRUE
      89        TRUE
      90        TRUE
      91        TRUE
      92        TRUE
      93        TRUE
      94        TRUE
      95        TRUE
      96        TRUE
      97        TRUE
      98        TRUE
      99        TRUE
      100       TRUE
      101       TRUE
      102       TRUE
      103       TRUE
      104       TRUE
      105       TRUE
      106       TRUE
      107       TRUE
      108       TRUE
      109       TRUE
      110       TRUE
      111       TRUE
      112       TRUE
      113       TRUE
      114       TRUE
      115       TRUE
      116       TRUE
      117       TRUE
      118       TRUE
      119       TRUE
      120       TRUE
      121       TRUE
      122       TRUE
      123       TRUE
      124       TRUE
      125       TRUE
      126       TRUE
      127       TRUE
      128       TRUE
      129       TRUE
      130       TRUE
      131       TRUE
      132       TRUE
      133       TRUE
      134       TRUE
      135       TRUE
      136       TRUE
      137       TRUE
      138       TRUE
      139       TRUE
      140       TRUE
      141       TRUE
      142       TRUE
      143       TRUE
      144       TRUE
      145       TRUE
      146       TRUE
      147       TRUE
      148       TRUE
      149       TRUE

# zarr_overview console output matches snapshot for consolidated store

    Code
      zarr_overview(zarr_store_consolidated, as_data_frame = FALSE)
    Output
      Type: Group of Arrays
      Path: <path>
      Arrays:
      ---
      Path: <path>
        Shape: 1
        Chunk Shape: 1
        No. of Chunks: 1 (1)
        Data Type: float64
        Endianness: little
        Compressor: blosc
        Attributes: no
      ---
      Path: <path>
        Shape: 2 x 2
        Chunk Shape: 2 x 2
        No. of Chunks: 1 (1 x 1)
        Data Type: float64
        Endianness: little
        Compressor: blosc
        Attributes: no
      ---
      Path: <path>
        Shape: 3 x 3 x 3
        Chunk Shape: 3 x 3 x 3
        No. of Chunks: 1 (1 x 1 x 1)
        Data Type: float64
        Endianness: little
        Compressor: blosc
        Attributes: no

# zarr_overview works with consolidated v3 metadata

    Code
      zarr_overview(zarr_store_consolidated_v3, as_data_frame = FALSE)
    Output
      Type: Group of Arrays
      Path: <path>
      Arrays:
      ---
      Path: <path>
        Shape: 3 x 45087 x 11580
        Chunk Shape: 1 x 4096 x 4096
        No. of Chunks: 108 (3 x 12 x 3)
        Data Type: uint8
        Endianness: NA
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 3 x 22543 x 5790
        Chunk Shape: 1 x 4096 x 4096
        No. of Chunks: 36 (3 x 6 x 2)
        Data Type: uint8
        Endianness: NA
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 3 x 11271 x 2895
        Chunk Shape: 1 x 4096 x 2895
        No. of Chunks: 9 (3 x 3 x 1)
        Data Type: uint8
        Endianness: NA
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 3 x 5635 x 1447
        Chunk Shape: 1 x 4096 x 1447
        No. of Chunks: 6 (3 x 2 x 1)
        Data Type: uint8
        Endianness: NA
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 3 x 2817 x 723
        Chunk Shape: 1 x 2817 x 723
        No. of Chunks: 3 (3 x 1 x 1)
        Data Type: uint8
        Endianness: NA
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 4 x 17098 x 51187
        Chunk Shape: 1 x 4096 x 4096
        No. of Chunks: 260 (4 x 5 x 13)
        Data Type: uint16
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 4 x 8549 x 25593
        Chunk Shape: 1 x 4096 x 4096
        No. of Chunks: 84 (4 x 3 x 7)
        Data Type: uint16
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 4 x 4274 x 12796
        Chunk Shape: 1 x 4096 x 4096
        No. of Chunks: 32 (4 x 2 x 4)
        Data Type: uint16
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 4 x 2137 x 6398
        Chunk Shape: 1 x 2137 x 4096
        No. of Chunks: 8 (4 x 1 x 2)
        Data Type: uint16
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 4 x 1068 x 3199
        Chunk Shape: 1 x 1068 x 3199
        No. of Chunks: 4 (4 x 1 x 1)
        Data Type: uint16
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 17098 x 51187
        Chunk Shape: 4096 x 4096
        No. of Chunks: 65 (5 x 13)
        Data Type: uint32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 8549 x 25593
        Chunk Shape: 4096 x 4096
        No. of Chunks: 21 (3 x 7)
        Data Type: uint32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 4274 x 12796
        Chunk Shape: 4096 x 4096
        No. of Chunks: 8 (2 x 4)
        Data Type: uint32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 2137 x 6398
        Chunk Shape: 2137 x 4096
        No. of Chunks: 2 (1 x 2)
        Data Type: uint32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 1068 x 3199
        Chunk Shape: 1068 x 3199
        No. of Chunks: 1 (1 x 1)
        Data Type: uint32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 17098 x 51187
        Chunk Shape: 4096 x 4096
        No. of Chunks: 65 (5 x 13)
        Data Type: uint32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 8549 x 25593
        Chunk Shape: 4096 x 4096
        No. of Chunks: 21 (3 x 7)
        Data Type: uint32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 4274 x 12796
        Chunk Shape: 4096 x 4096
        No. of Chunks: 8 (2 x 4)
        Data Type: uint32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 2137 x 6398
        Chunk Shape: 2137 x 4096
        No. of Chunks: 2 (1 x 2)
        Data Type: uint32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 1068 x 3199
        Chunk Shape: 1068 x 3199
        No. of Chunks: 1 (1 x 1)
        Data Type: uint32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 
        Chunk Shape: 
        No. of Chunks: 1 ()
        Data Type: bool
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 162254
        No. of Chunks: 1 (1)
        Data Type: string
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: float64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 162254
        No. of Chunks: 1 (1)
        Data Type: string
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: int64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: int64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: int64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: int64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: float64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: float64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: int64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: int64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: int64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 40564
        No. of Chunks: 4 (4)
        Data Type: float64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254 x 2
        Chunk Shape: 40564 x 1
        No. of Chunks: 8 (4 x 2)
        Data Type: float64
        Endianness: little
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 377
        Chunk Shape: 377
        No. of Chunks: 1 (1)
        Data Type: string
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 377
        Chunk Shape: 377
        No. of Chunks: 1 (1)
        Data Type: string
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 377
        Chunk Shape: 377
        No. of Chunks: 1 (1)
        Data Type: string
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 377
        Chunk Shape: 377
        No. of Chunks: 1 (1)
        Data Type: string
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 4900367
        Chunk Shape: 153137
        No. of Chunks: 32 (32)
        Data Type: float32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 4900367
        Chunk Shape: 153137
        No. of Chunks: 32 (32)
        Data Type: int32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 162255
        Chunk Shape: 81128
        No. of Chunks: 2 (2)
        Data Type: int32
        Endianness: little
        Compressor: zstd
        Attributes: no
      ---
      Path: <path>
        Shape: 1
        Chunk Shape: 1
        No. of Chunks: 1 (1)
        Data Type: string
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 162254
        Chunk Shape: 162254
        No. of Chunks: 1 (1)
        Data Type: int8
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 
        Chunk Shape: 
        No. of Chunks: 1 ()
        Data Type: string
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 
        Chunk Shape: 
        No. of Chunks: 1 ()
        Data Type: string
        Endianness: NA
        Compressor: zstd
        Attributes: yes
      ---
      Path: <path>
        Shape: 
        Chunk Shape: 
        No. of Chunks: 1 ()
        Data Type: string
        Endianness: NA
        Compressor: zstd
        Attributes: yes

# zarr_overview console output matches snapshot for v3 metadata

    Code
      zarr_overview(zarr_v3, as_data_frame = FALSE)
    Output
      Type: Array
      Path: <path>
      Shape: 3 x 4 x 5
      Chunk Shape: 3 x 4 x 5
      No. of Chunks: 1 (1 x 1 x 1)
      Data Type: int64
      Endianness: little
      Compressor: zstd
      Attributes: no

# zarr_overview throws error for mixed v2/v3 zarr

    The path contains both `.zarray` (Zarr V2 specification) and `zarr.json` (Zarr V3 specification) metadata files.
    An array or group must conform to either the Zarr V2 or V3 specification.

---

    The path contains both `.zmetadata` (Zarr V2 specification) and `zarr.json` (Zarr V3 specification) consolidated metadata files.
    An array or group must conform to either the Zarr V2 or V3 specification.

# read_zarr_attributes errors clearly for invalid mixed arrays

    The path contains both `.zattrs` (v2 Zarr specification) and `zarr.json` (v3 Zarr specification) files, which is not allowed.

# read_zarr_attributes from s3

    Code
      read_zarr_attributes(
        "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.3/9836842.zarr/")
    Output
      $`_creator`
      $`_creator`$name
      [1] "omero-zarr"
      
      $`_creator`$version
      [1] "0.1.dev219+g541c88e"
      
      
      $multiscales
      $multiscales[[1]]
      $multiscales[[1]]$axes
      $multiscales[[1]]$axes[[1]]
      [1] "c"
      
      $multiscales[[1]]$axes[[2]]
      [1] "y"
      
      $multiscales[[1]]$axes[[3]]
      [1] "x"
      
      
      $multiscales[[1]]$datasets
      $multiscales[[1]]$datasets[[1]]
      $multiscales[[1]]$datasets[[1]]$path
      [1] "0"
      
      
      $multiscales[[1]]$datasets[[2]]
      $multiscales[[1]]$datasets[[2]]$path
      [1] "1"
      
      
      $multiscales[[1]]$datasets[[3]]
      $multiscales[[1]]$datasets[[3]]$path
      [1] "2"
      
      
      $multiscales[[1]]$datasets[[4]]
      $multiscales[[1]]$datasets[[4]]$path
      [1] "3"
      
      
      $multiscales[[1]]$datasets[[5]]
      $multiscales[[1]]$datasets[[5]]$path
      [1] "4"
      
      
      $multiscales[[1]]$datasets[[6]]
      $multiscales[[1]]$datasets[[6]]$path
      [1] "5"
      
      
      
      $multiscales[[1]]$version
      [1] "0.3"
      
      
      
      $omero
      $omero$channels
      $omero$channels[[1]]
      $omero$channels[[1]]$active
      [1] TRUE
      
      $omero$channels[[1]]$coefficient
      [1] 1
      
      $omero$channels[[1]]$color
      [1] "FF0000"
      
      $omero$channels[[1]]$family
      [1] "linear"
      
      $omero$channels[[1]]$inverted
      [1] FALSE
      
      $omero$channels[[1]]$label
      [1] "Cam2-T1"
      
      $omero$channels[[1]]$window
      $omero$channels[[1]]$window$end
      [1] 5090
      
      $omero$channels[[1]]$window$max
      [1] 65535
      
      $omero$channels[[1]]$window$min
      [1] 0
      
      $omero$channels[[1]]$window$start
      [1] 198
      
      
      
      $omero$channels[[2]]
      $omero$channels[[2]]$active
      [1] TRUE
      
      $omero$channels[[2]]$coefficient
      [1] 1
      
      $omero$channels[[2]]$color
      [1] "00C000"
      
      $omero$channels[[2]]$family
      [1] "linear"
      
      $omero$channels[[2]]$inverted
      [1] FALSE
      
      $omero$channels[[2]]$label
      [1] "Cam1-T2"
      
      $omero$channels[[2]]$window
      $omero$channels[[2]]$window$end
      [1] 5736
      
      $omero$channels[[2]]$window$max
      [1] 65535
      
      $omero$channels[[2]]$window$min
      [1] 0
      
      $omero$channels[[2]]$window$start
      [1] 198
      
      
      
      $omero$channels[[3]]
      $omero$channels[[3]]$active
      [1] TRUE
      
      $omero$channels[[3]]$coefficient
      [1] 1
      
      $omero$channels[[3]]$color
      [1] "FFFFFF"
      
      $omero$channels[[3]]$family
      [1] "linear"
      
      $omero$channels[[3]]$inverted
      [1] FALSE
      
      $omero$channels[[3]]$label
      [1] "2"
      
      $omero$channels[[3]]$window
      $omero$channels[[3]]$window$end
      [1] 5312
      
      $omero$channels[[3]]$window$max
      [1] 65535
      
      $omero$channels[[3]]$window$min
      [1] 0
      
      $omero$channels[[3]]$window$start
      [1] 196
      
      
      
      $omero$channels[[4]]
      $omero$channels[[4]]$active
      [1] FALSE
      
      $omero$channels[[4]]$coefficient
      [1] 1
      
      $omero$channels[[4]]$color
      [1] "FFFFFF"
      
      $omero$channels[[4]]$family
      [1] "linear"
      
      $omero$channels[[4]]$inverted
      [1] FALSE
      
      $omero$channels[[4]]$label
      [1] "3"
      
      $omero$channels[[4]]$window
      $omero$channels[[4]]$window$end
      [1] 4128
      
      $omero$channels[[4]]$window$max
      [1] 65535
      
      $omero$channels[[4]]$window$min
      [1] 0
      
      $omero$channels[[4]]$window$start
      [1] 194
      
      
      
      
      $omero$id
      [1] 1
      
      $omero$rdefs
      $omero$rdefs$defaultT
      [1] 0
      
      $omero$rdefs$defaultZ
      [1] 0
      
      $omero$rdefs$model
      [1] "color"
      
      
      $omero$version
      [1] "0.3"
      
      

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_metadata.R
\name{.read_array_metadata}
\alias{.read_array_metadata}
\title{Read the \code{.zarray} or \code{zarr.json} metadata file associated with a Zarr array}
\usage{
.read_array_metadata(zarr_path, s3_client = NULL)
}
\arguments{
\item{zarr_path}{A character vector of length 1.  This provides the
path to a Zarr array or group of arrays. This can either be on a local file
system or on S3 storage.}

\item{s3_client}{A list representing an S3 client.  This should be produced
by \code{\link[paws.storage:s3]{paws.storage::s3()}}.}
}
\value{
A list containing the array metadata
}
\description{
Read the \code{.zarray} or \code{zarr.json} metadata file associated with a Zarr array
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_utils.R
\name{.prepare_write_data}
\alias{.prepare_write_data}
\title{Truncate overflow values and warn about bool NA values before writing}
\usage{
.prepare_write_data(x, metadata)
}
\arguments{
\item{x}{The R array to be written.}

\item{metadata}{Array metadata list as returned by \code{\link[=.read_array_metadata]{.read_array_metadata()}}.}
}
\value{
\code{x}, possibly with overflow values clamped.
}
\description{
Applies the two pre-write data checks that are needed identically in both
\code{\link[=write_zarr_array]{write_zarr_array()}} and \code{\link[=update_zarr_array]{update_zarr_array()}}:
}
\details{
\enumerate{
\item If the target type is narrower than R's native representation (e.g.
\code{int8}, \code{int16}, \code{float32}), clamp out-of-range values via
\code{.truncate_overflow()}.
\item If the target type is \code{bool}, warn when \code{x} contains \code{NA} values (which
the Zarr bool type cannot represent).
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/store_utils.R
\name{.read_json_file}
\alias{.read_json_file}
\title{Read a JSON file from local disk or S3}
\usage{
.read_json_file(path, s3_client = NULL)
}
\arguments{
\item{path}{Full path (local or S3) to a JSON file.}

\item{s3_client}{An S3 client produced by \code{\link[paws.storage:s3]{paws.storage::s3()}}, or \code{NULL}
for local files.}
}
\value{
A list as returned by \code{\link[jsonlite:read_json]{jsonlite::read_json()}} /
\code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}}.
}
\description{
Read a JSON file from local disk or S3
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_metadata.R
\name{read_zarr_consolidated_metadata}
\alias{read_zarr_consolidated_metadata}
\title{Read consolidated metadata}
\usage{
read_zarr_consolidated_metadata(
  zarr_path,
  consolidate = c("never", "missing", "always"),
  s3_client = NULL
)
}
\arguments{
\item{zarr_path}{A character vector of length 1.  This provides the
path to a Zarr array or group of arrays. This can either be on a local file
system or on S3 storage.}

\item{consolidate}{One of \code{"never"} (default), \code{"missing"} or \code{"always"}.
This controls whether the consolidated metadata needs to be (re)generated
on the fly, in memory. If \code{"never"} then the consolidated metadata is only
read if it already exists, and \code{NULL} is returned otherwise. If \code{"missing"}
then the consolidated metadata is generated if it does not already exist.
If \code{"always"} then the consolidated metadata is always (re)generated, even
if it already exists.}

\item{s3_client}{A list representing an S3 client.  This should be produced
by \code{\link[paws.storage:s3]{paws.storage::s3()}}.}
}
\value{
A list containing the consolidated metadata, formatted as Zarr v3
consolidated metadata.
If \code{consolidate = "never"} and the consolidated metadata does not exist,
\code{NULL} is returned.
}
\description{
Read consolidated metadata
}
\details{
This is stored in the \code{.zmetadata} (Zarr v2) or \code{zarr.json} (Zarr v3) file at
the root of a Zarr store.

Note that it is not documented in the official Zarr specification, because
it is not (yet?) part of the standard.

In particular, it lists the location of all the metadata files for arrays in
the current group, so it is not necessary to crawl to discover them. It
leads to much better performance, in particular when reading arrays from S3
storage.
}
\examples{
z <- system.file("extdata", "zarr_examples", "metadata", "consolidated.zarr", package = "Rarr")
m <- read_zarr_consolidated_metadata(z)
head(m)

}
\references{
\url{https://zarr.readthedocs.io/en/latest/user-guide/consolidated_metadata.html}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_data.R
\name{create_empty_zarr_array}
\alias{create_empty_zarr_array}
\title{Create an (empty) Zarr array}
\usage{
create_empty_zarr_array(
  zarr_array_path,
  dim,
  chunk_dim,
  data_type,
  order = c("F", "C"),
  compressor = use_zstd(),
  fill_value = NULL,
  nchar = NULL,
  dimension_separator = if (zarr_version == 2L) "." else "/",
  dimension_names = NULL,
  zarr_version = 3L
)
}
\arguments{
\item{zarr_array_path}{Character vector of length 1 giving the path to the
new Zarr array.}

\item{dim}{Dimensions of the new array.  Should be a numeric vector with the
same length as the number of dimensions.}

\item{chunk_dim}{Dimensions of the array chunks. Should be a numeric vector
with the same length as the \code{dim} argument.}

\item{data_type}{Character vector giving the data type of the new array.
Valid options are: "integer", "double", "character", "logical", which are
based on standard R data types. You can also use the analogous Numpy
formats: "|i1", "<i2", "<i4", "<f4", "<f8", "|S", "|b1".
If this argument isn't provided the \code{fill_value} will be used to determine
the datatype.}

\item{order}{Define the layout of the bytes within each chunk.  Valid options
are 'column', 'row', 'F' & 'C'.  'column' or 'F' will specify
"column-major" ordering, which is how R arrays are arranged in memory.
'row' or 'C' will specify "row-major" order.}

\item{compressor}{What (if any) compression tool should be applied to the
array chunks.  The default is to use \code{zstd} compression. Supplying \code{NULL}
will disable chunk compression. See \link{compressors} for more details.}

\item{fill_value}{The default value for uninitialized portions of the array.
Does not have to be provided, in which case the default for the specified
data type will be used.}

\item{nchar}{For \code{datatype = "character"} this parameter gives the maximum
length of the stored strings. It is an error not to specify this for a
character array, but it is ignored for other data types.}

\item{dimension_separator}{The character used to to separate the dimensions
in the names of the chunk files.  Valid options are limited to "." and "/".}

\item{dimension_names}{Optional character vector with the same length as
\code{dim}.}

\item{zarr_version}{The version of the Zarr specification to use. Currently,
either \code{2} or \code{3}. The default is \code{3}.}
}
\value{
This function is primarily called for the side effect of
initialising a Zarr array location and creating the \code{.zarray} or
\code{zarr.json} metadata file.
Returns (invisibly) the normalized path it wrote the metadata to.
}
\description{
Create an (empty) Zarr array
}
\examples{

new_zarr_array <- file.path(tempdir(), "temp.zarr")
create_empty_zarr_array(new_zarr_array,
  dim = c(10, 20), chunk_dim = c(2, 5),
  data_type = "integer"
)

}
\seealso{
\code{\link[=write_zarr_array]{write_zarr_array()}}, \code{\link[=update_zarr_array]{update_zarr_array()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/codecs_utils.R
\name{.configure_codecs}
\alias{.configure_codecs}
\title{Convert the metadata \code{codecs} elements into functions}
\usage{
.configure_codecs(codecs, operation = c("encode", "decode"))
}
\arguments{
\item{codecs}{A list containing the Zarr v3 codecs}

\item{operation}{One of "encode" or "decode"}
}
\value{
An list of 3 environments containing functions:
\itemize{
\item \code{bytes_bytes_codecs}: functions to encode/decode raw bytes
\item \code{array_array_codecs}: functions to encode/decode R arrays
\item \code{array_bytes_codecs}: functions to encode/decode between R arrays and raw bytes
}
}
\description{
Convert the metadata \code{codecs} elements into functions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_metadata.R
\name{zarr_consolidate_metadata}
\alias{zarr_consolidate_metadata}
\title{Consolidate Zarr metadata files into a single file}
\usage{
zarr_consolidate_metadata(
  zarr_store_path,
  s3_client = NULL,
  action = c("write", "return"),
  overwrite = TRUE
)
}
\arguments{
\item{zarr_store_path}{A character vector of length 1. This provides the path to a Zarr store.}

\item{s3_client}{A list representing an S3 client.  This should be produced
by \code{\link[paws.storage:s3]{paws.storage::s3()}}.}

\item{action}{A character string specifying the action to take with the consolidated metadata.
If \code{"write"} (the default), the consolidated metadata will be written back to the Zarr store.
If \code{"return"}, the consolidated metadata will be returned as a list without writing it back to the store.
The latter is particularly useful for non-writable stores.}

\item{overwrite}{A logical value (default \code{TRUE}) indicating whether to overwrite existing
consolidated metadata when \code{action} is \code{"write"}. If \code{FALSE} and consolidated metadata
already exists, an error will be raised.}
}
\value{
If \code{action} is \code{"return"}, a list containing the consolidated metadata.
Otherwise, the function is called for its side effect and \code{NULL} is returned invisibly.
}
\description{
This function reads all the metadata files in a Zarr store and consolidates them into a single file.
Thanks to this, a single request can be made to retrieve all the elements and their related metadata for a Zarr store,
which is especially beneficial for remote stores like S3.
}
\examples{
# v2
zarr_v2 <- withr::local_tempfile(fileext = ".zarr")
dir.create(zarr_v2)
jsonlite::write_json(
  list("zarr_format" = 2L),
  file.path(zarr_v2, ".zgroup")
)
write_zarr_array(
  array(1:4, dim = c(2, 2)),
  file.path(zarr_v2, "array1"),
  chunk_dim = c(1, 2),
  zarr_version = 2L
)
write_zarr_array(
  array(c(3.14, 42.42, 12.96, 7.89), dim = c(2, 2)),
  file.path(zarr_v2, "array2"),
  chunk_dim = c(1, 2),
  zarr_version = 2L
)
write_zarr_attributes(
 file.path(zarr_v2, "array1"),
 list(description = "This is array 1")
)
zarr_consolidate_metadata(zarr_v2, action = "return")

zarr_consolidate_metadata(zarr_v2, action = "write")
zarr_overview(zarr_v2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_metadata.R
\name{zarr_overview}
\alias{zarr_overview}
\title{Print a summary of a Zarr array or group}
\usage{
zarr_overview(zarr_array_path, s3_client = NULL, as_data_frame = FALSE)
}
\arguments{
\item{zarr_array_path}{A character vector of length 1.  This provides the
path to a Zarr array or group of arrays. This can either be on a local file
system or on S3 storage.}

\item{s3_client}{A list representing an S3 client.  This should be produced
by \code{\link[paws.storage:s3]{paws.storage::s3()}}.}

\item{as_data_frame}{Logical determining whether the Zarr array details
should be printed to screen (\code{FALSE}) or returned as a \code{data.frame}
(\code{TRUE}) so they can be used computationally.}
}
\value{
If \code{as_data_frame = FALSE} the function invisible returns \code{TRUE} if
successful.  However it is primarily called for the side effect of printing
details of the Zarr array(s) to the screen.  If \code{as_data_frame = TRUE} then
a \code{data.frame} containing details of the array is returned.
}
\description{
When reading a Zarr array using \code{\link[=read_zarr_array]{read_zarr_array()}} it is necessary to know
it's shape and size. \code{zarr_overview()} can be used to get a quick overview of
the array shape and contents, based on the \code{.zarray} (Zarr v2) or \code{zarr.json}
(Zarr v3) metadata file each array contains.
}
\details{
The function currently prints the following information to the R console:
\itemize{
\item array path
\item array shape and size
\item chunk and size
\item the number of chunks
\item the datatype of the array
\item codec used for data compression (if any)
}

If given the path to a group of arrays the function will attempt to print the
details of all sub-arrays in the group.
}
\examples{

## Using a local file provided with the package
z1 <- system.file("extdata", "zarr_examples", "row-first",
  "int32.zarr",
  package = "Rarr"
)

## read the entire array
zarr_overview(zarr_array_path = z1)

## using a file on S3 storage
\donttest{
z2 <- "https://noaa-nwm-retro-v2-zarr-pds.s3.amazonaws.com/feature_id/"
zarr_overview(z2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Rarr-package.R
\docType{package}
\name{Rarr-package}
\alias{Rarr}
\alias{Rarr-package}
\title{Rarr: A Simple and Performant Native R Reader & Writer for Zarr Arrays}
\description{
The Zarr specification defines a format for chunked, compressed, N-dimensional arrays. Its design allows efficient access to subsets of the stored array, and supports both local and cloud storage systems. Rarr aims to implement this specification in R with minimal reliance on an external tools or libraries.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://huber-group-embl.github.io/Rarr/}
  \item \url{https://github.com/Huber-group-EMBL/Rarr}
  \item Report bugs at \url{https://github.com/Huber-group-EMBL/Rarr/issues}
}

}
\author{
\strong{Maintainer}: Hugo Gruson \email{hugo.gruson@embl.de} (\href{https://orcid.org/0000-0002-4094-1476}{ORCID})

Authors:
\itemize{
  \item Hugo Gruson \email{hugo.gruson@embl.de} (\href{https://orcid.org/0000-0002-4094-1476}{ORCID})
  \item Mike Smith (\href{https://orcid.org/0000-0002-7800-3848}{ORCID}) (Maintainer from 2022 to 2025.) [conceptor]
}

Other contributors:
\itemize{
  \item Artür Manukyan [contributor]
  \item Sharla Gelfand [contributor]
  \item German Network for Bioinformatics Infrastructure - de.NBI (\href{https://ror.org/01vmpm840}{ROR}) [funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.extract_chunk}
\alias{.extract_chunk}
\title{Subset extraction for an array with a variable number of dimensions.}
\usage{
.extract_chunk(x, idx)
}
\arguments{
\item{x}{Array to extract from.}

\item{idx}{List of index vectors, one per dimension.}
}
\value{
The extracted sub-array (with \code{drop = FALSE}).
}
\description{
Subset extraction for an array with a variable number of dimensions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_metadata.R
\name{read_zarr_attributes}
\alias{read_zarr_attributes}
\title{Read the attributes associated with a Zarr array or group}
\usage{
read_zarr_attributes(
  zarr_path,
  s3_client = NULL,
  missing = c("ignore", "warning", "error")
)
}
\arguments{
\item{zarr_path}{A character vector of length 1.  This provides the
path to a Zarr array or group of arrays. This can either be on a local file
system or on S3 storage.}

\item{s3_client}{A list representing an S3 client.  This should be produced
by \code{\link[paws.storage:s3]{paws.storage::s3()}}.}

\item{missing}{A character vector of length 1. This determines the behaviour
when no file containing attributes is found. This can be one of:
\itemize{
\item "ignore" (the default): an empty list is returned silently
\item "warning": a warning is issued and an empty list is returned.
\item "error": an error is raised.
}}
}
\value{
A list containing the attributes. If the file containing attributes
(\code{.zattrs} for Zarr v2 or \code{zarr.json} for Zarr v3) exists but no attributes
are provided, an empty list is returned.
}
\description{
Read the attributes associated with a Zarr array or group
}
\examples{
read_zarr_attributes(
  "https://uk1s3.embassy.ebi.ac.uk/idr/zarr/v0.4/idr0048A/9846152.zarr"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_metadata.R
\name{.normalize_v3_metadata}
\alias{.normalize_v3_metadata}
\title{Apply Zarr v3-specific fixups to raw array metadata}
\usage{
.normalize_v3_metadata(metadata)
}
\arguments{
\item{metadata}{A list as returned by \code{\link[=.read_json_file]{.read_json_file()}} for a Zarr v3
array node (i.e. \code{metadata$zarr_format == 3L} and
\code{metadata$node_type == "array"}).}
}
\value{
The modified \code{metadata} list.
}
\description{
Called after reading a \code{zarr.json} file for an array node.  Performs four
normalisation steps that are needed before the metadata can be used by the
rest of the package:
}
\details{
\enumerate{
\item Parse the \code{data_type} string into an R-friendly \code{datatype} list.
\item Name \code{codecs} by their \code{name} field for O(1) lookup.
\item Normalise scalar arrays (empty \code{shape}) to shape \code{1}.
\item Inject a default \code{transpose} codec when absent (R uses F-order so we
always need to know the intended order).
\item Replicate the \code{endian} field across all fields of struct datatypes so
that downstream chunk-reading code can treat it uniformly.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Rarr-deprecated.R
\name{ZarrArray}
\alias{ZarrArray}
\alias{writeZarrArray}
\title{Deprecated DelayedArray backend functions}
\usage{
ZarrArray(...)

writeZarrArray(...)
}
\arguments{
\item{...}{Passed to the new function in the ZarrArray package.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

The DelayedArray backend has moved to a dedicated package: the ZarrArray package
(\url{https://github.com/Bioconductor/ZarrArray}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_data.R
\name{update_zarr_array}
\alias{update_zarr_array}
\title{Update (a subset of) an existing Zarr array}
\usage{
update_zarr_array(zarr_array_path, x, index)
}
\arguments{
\item{zarr_array_path}{Character vector of length 1 giving the path to the
Zarr array that is to be modified.}

\item{x}{The R array (or object that can be coerced to an array) that will be
written to the Zarr array.}

\item{index}{A list with the same length as the number of dimensions of the
target array. This argument indicates which elements in the target array
should be updated.}
}
\value{
The function is primarily called for the side effect of writing to
disk. Returns (invisibly) \code{TRUE} if the array is successfully updated.
}
\description{
Update (a subset of) an existing Zarr array
}
\examples{

## first create a new, empty, Zarr array
new_zarry_array <- file.path(tempdir(), "new_array.zarr")
create_empty_zarr_array(
  zarr_array_path = new_zarry_array, dim = c(20, 10),
  chunk_dim = c(10, 5), data_type = "double"
)

## create a matrix smaller than our Zarr array
small_matrix <- matrix(runif(6), nrow = 3)

## insert the matrix into the first 3 rows, 2 columns of the Zarr array
update_zarr_array(new_zarry_array, x = small_matrix, index = list(1:3, 1:2))

## reading back a slightly larger subset,
## we can see only the top left corner has been changed
read_zarr_array(new_zarry_array, index = list(1:5, 1:5))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_data.R
\name{write_zarr_array}
\alias{write_zarr_array}
\title{Write an R array to Zarr}
\usage{
write_zarr_array(
  x,
  zarr_array_path,
  chunk_dim,
  data_type = storage.mode(x),
  order = c("F", "C"),
  compressor = use_zstd(),
  fill_value = NULL,
  nchar,
  dimension_separator = if (zarr_version == 2L) "." else "/",
  zarr_version = 3L
)
}
\arguments{
\item{x}{The R array that will be written to the Zarr array.}

\item{zarr_array_path}{Character vector of length 1 giving the path to the
new Zarr array.}

\item{chunk_dim}{Dimensions of the array chunks. Should be a numeric vector
with the same length as the \code{dim} argument.}

\item{data_type}{Character vector giving the data type of the new array.
Valid options are: "integer", "double", "character", "logical", which are
based on standard R data types. You can also use the analogous Numpy
formats: "|i1", "<i2", "<i4", "<f4", "<f8", "|S", "|b1".
If this argument isn't provided the \code{fill_value} will be used to determine
the datatype.}

\item{order}{Define the layout of the bytes within each chunk.  Valid options
are 'column', 'row', 'F' & 'C'.  'column' or 'F' will specify
"column-major" ordering, which is how R arrays are arranged in memory.
'row' or 'C' will specify "row-major" order.}

\item{compressor}{What (if any) compression tool should be applied to the
array chunks.  The default is to use \code{zstd} compression. Supplying \code{NULL}
will disable chunk compression. See \link{compressors} for more details.}

\item{fill_value}{The default value for uninitialized portions of the array.
Does not have to be provided, in which case the default for the specified
data type will be used.}

\item{nchar}{For character arrays this parameter gives the maximum length of
the stored strings. If this argument is not specified the array provided to
\code{x} will be checked and the length of the longest string found will be used
so no data are truncated. However this may be slow and providing a value to
\code{nchar} can provide a modest performance improvement.}

\item{dimension_separator}{The character used to to separate the dimensions
in the names of the chunk files.  Valid options are limited to "." and "/".}

\item{zarr_version}{The version of the Zarr specification to use. Currently,
either \code{2} or \code{3}. The default is \code{3}.}
}
\value{
The function is primarily called for the side effect of writing to
disk. Returns (invisibly) \code{TRUE} if the array is successfully written.
}
\description{
Write an R array to Zarr
}
\note{
If \code{x} has \code{dimnames}, \code{names(dimnames(x))} will be stored as the
\code{dimension_names} field in the Zarr metadata.
}
\examples{

new_zarr_array <- file.path(tempdir(), "integer.zarr")
x <- array(1:50, dim = c(10, 5))
write_zarr_array(
  x = x, zarr_array_path = new_zarr_array,
  chunk_dim = c(2, 5)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3.R
\name{.get_credentials}
\alias{.get_credentials}
\title{This is a modified version of paws.storage:::get_credentials().  It is
included to prevent using the \code{:::} operator.  Look at that function if
things stop working.}
\usage{
.get_credentials(credentials)
}
\arguments{
\item{credentials}{Content stored at \code{.internal$config$credentials} in
an object created by \code{paws.storage::s3()}.}
}
\value{
A credentials list to be reinserted into a \code{paws.storage} s3 object.
If no valid credentials are found this function will error, which is expected
and is caught by \code{.check_credentials}.
}
\description{
This is a modified version of paws.storage:::get_credentials().  It is
included to prevent using the \code{:::} operator.  Look at that function if
things stop working.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compressors.R
\name{compressors}
\alias{compressors}
\alias{use_blosc}
\alias{use_zlib}
\alias{use_gzip}
\alias{use_bz2}
\alias{use_lzma}
\alias{use_lz4}
\alias{use_zstd}
\title{Define compression tool and settings}
\usage{
use_blosc(
  cname = c("lz4", "lz4hc", "blosclz", "zstd", "zlib", "snappy"),
  clevel = 5L,
  shuffle = c("shuffle", "noshuffle", "bitshuffle"),
  typesize = NULL,
  blocksize = 0L
)

use_zlib(level = 6L)

use_gzip(level = 6L)

use_bz2(level = 6L)

use_lzma(level = 9L)

use_lz4()

use_zstd(level = 0L)
}
\arguments{
\item{cname}{Blosc is a 'meta-compressor' providing access to several
compression algorithms.  This argument defines which compression tool
should be used.  Valid options are: \code{"lz4"}, \code{"lz4hc"}, \code{"blosclz"},
\code{"zstd"}, \code{"zlib"}, \code{"snappy"}.}

\item{clevel}{An integer from 0 to 9 which controls the speed and level of
compression. A level of 1 is the fastest compression method and produces
the least compressions, while 9 is slowest and produces the most compression.
Compression is turned off completely when level is 0. Defaults to 5.}

\item{shuffle}{Specifies the type of shuffling to perform, if any, prior to
compression. Must be one of \code{"noshuffle"}, to indicate no shuffling;
\code{"shuffle"} (default), to indicate byte-wise shuffling; \code{"bitshuffle"}, to
indicate bit-wise shuffling.}

\item{typesize}{The data type size in bytes used by Blosc shuffling. If
\code{NULL} (default), this will be inferred from the array datatype. Ignored if
\code{shuffle = "noshuffle"}.}

\item{blocksize}{The requested size of the compressed blocks in bytes. Use 0
(default) to let Blosc choose automatically.}

\item{level}{Specify the compression level to use.  The range of possible
values is dependant on the compression tool being used.  For example, for
\code{use_zlib()} this argument can be between 1 & 9, while for \code{use_zstd()}the
valid range is 0 (default; level chosen automatically) to 22.}
}
\value{
A list containing the details of the selected compression tool. This
will be written to the .zarray metadata when the Zarr array is created.
}
\description{
These functions select a compression tool and its setting when writing a Zarr
file
}
\examples{

## define 2 compression filters for blosc (using snappy) and bzip2 (level 5)
blosc_with_snappy_compression <- use_blosc(cname = "snappy")
bzip2_compression <- use_bz2(level = 5)

## create an example array to write to a file
x <- array(runif(n = 1000, min = -10, max = 10), dim = c(10, 20, 5))

## write the array to two files using each compression filter
blosc_path <- tempfile()
bzip2_path <- tempfile()
write_zarr_array(
  x = x, zarr_array_path = blosc_path, chunk_dim = c(2, 5, 1),
  compressor = blosc_with_snappy_compression
)
write_zarr_array(
  x = x, zarr_array_path = bzip2_path, chunk_dim = c(2, 5, 1),
  compressor = bzip2_compression
)

## the contents of the two arrays should be the same
identical(read_zarr_array(blosc_path), read_zarr_array(bzip2_path))

## the size of the files on disk are not the same
sum(file.size(list.files(blosc_path, full.names = TRUE)))
sum(file.size(list.files(bzip2_path, full.names = TRUE)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_metadata.R
\name{write_zarr_attributes}
\alias{write_zarr_attributes}
\title{Read the .zattrs file associated with a Zarr array or group}
\usage{
write_zarr_attributes(
  zarr_path,
  new.zattrs = list(),
  overwrite = TRUE,
  zarr_version = if (has_metadata_v2) 2L else 3L
)
}
\arguments{
\item{zarr_path}{A character vector of length 1. This provides the
path to a Zarr array or group.}

\item{new.zattrs}{a list inserted to .zattrs at the \code{path}.}

\item{overwrite}{if \code{TRUE} (the default), existing .zattrs elements will be overwritten by \code{new.zattrs}.}

\item{zarr_version}{The version of the Zarr specification to use. If a
metadata file already exists, the version will be inferred from the file.
Otherwise, the default is \code{3}.}
}
\value{
Invisibly, the updated attributes as a named list.
This is equivalent to (but faster than) using \code{read_zarr_attributes()} after writing.
If no attributes were present before, this is identical to \code{new.zattrs}.
}
\description{
Read the .zattrs file associated with a Zarr array or group
}
\examples{
z1 <- withr::local_tempdir(fileext = ".zarr")
write_zarr_attributes(z1, list(date = "2025-01-01", author = "Jane Doe"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_metadata.R
\name{write_zarr_group}
\alias{write_zarr_group}
\title{Initialize a Zarr group}
\usage{
write_zarr_group(
  zarr_path,
  group,
  zarr_version = if (has_metadata_v2) 2L else 3L
)
}
\arguments{
\item{zarr_path}{A character vector of length 1. This provides the
path to a Zarr store.}

\item{group}{A character vector of length 1. This provides the name of the
group to create. If \code{""}, the root group will be created.}

\item{zarr_version}{The version of the Zarr specification to use. If a
metadata file already exists, the version will be inferred from the file.
Otherwise, the default is \code{3}.}
}
\description{
Initialize a Zarr group
}
\details{
Nested groups are created recursively. For example, if
\code{group = "foo/bar"}, then the group \code{foo} will be created first,
followed by the group \code{bar} inside of it.
}
\examples{
zarr_v2 <- withr::local_tempfile(fileext = ".zarr")
write_zarr_group(zarr_v2, "test/deep/nested/group", zarr_version = 2L)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_data.R
\name{.compress_and_write_chunk}
\alias{.compress_and_write_chunk}
\title{Compress and write a single chunk}
\usage{
.compress_and_write_chunk(input_chunk, chunk_path, metadata)
}
\arguments{
\item{input_chunk}{Array containing the chunk data to be compressed.  Will be
converted to a raw vector before compression.}

\item{chunk_path}{Character string giving the path to the chunk that should
be written.}
}
\value{
Returns \code{TRUE} if writing is successful.  Mostly called for the
side-effect of writing the compressed chunk to disk.
}
\description{
Compress and write a single chunk
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_data.R
\name{read_zarr_array}
\alias{read_zarr_array}
\title{Read a Zarr array}
\usage{
read_zarr_array(zarr_array_path, index, s3_client = NULL)
}
\arguments{
\item{zarr_array_path}{Path to a Zarr array. A character vector of length 1.
This can either be a location on a local file system or the URI to an array
in S3 storage.}

\item{index}{A list of the same length as the number of dimensions in the
Zarr array.  Each entry in the list provides the indices in that dimension
that should be read from the array.  Setting a list entry to \code{NULL} will
read everything in the associated dimension.  If this argument is missing
the entirety of the the Zarr array will be read.}

\item{s3_client}{Object created by \code{\link[paws.storage:s3]{paws.storage::s3()}}. Only required for a
file on S3. Leave as \code{NULL} for a file on local storage.}
}
\value{
An array with the same number of dimensions as the input array. The
extent of each dimension will correspond to the length of the values
provided to the \code{index} argument.
}
\description{
Read a Zarr array
}
\examples{

## Using a local file provided with the package
## This array has 3 dimensions
z1 <- system.file("extdata", "zarr_examples", "row-first", "int32.zarr", package = "Rarr")

## read the entire array
read_zarr_array(zarr_array_path = z1)

## extract values for first 10 rows, all columns, first slice
read_zarr_array(zarr_array_path = z1, index = list(1:10, NULL, 1))

\donttest{
## using a Zarr file hosted on Amazon S3
## This array has a single dimension with length 2729077
z2 <- "https://noaa-nwm-retro-v2-zarr-pds.s3.amazonaws.com/feature_id/"

## read the entire array
read_zarr_array(zarr_array_path = z2)

## read alternating elements
read_zarr_array(zarr_array_path = z2, index = list(seq(1, 576, 2)))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_metadata.R
\name{.read_consolidated_metadata}
\alias{.read_consolidated_metadata}
\title{Read consolidated metadata file}
\usage{
.read_consolidated_metadata(
  zarr_path,
  nodes = c("group", "array"),
  s3_client = NULL
)
}
\arguments{
\item{zarr_path}{A character vector of length 1.  This provides the
path to a Zarr array or group of arrays. This can either be on a local file
system or on S3 storage.}

\item{s3_client}{A list representing an S3 client.  This should be produced
by \code{\link[paws.storage:s3]{paws.storage::s3()}}.}
}
\description{
Read consolidated metadata file
}
\details{
This is stored in the \code{.zmetadata} file at the root of a Zarr store.
Note that it is not documented in the official Zarr specification, because
it is not (yet?) part of the standard.

It is implemented in zarr-python and discussed under the
"consolidated metadata" phrase.

In particular, it lists the location of all the metadata files for arrays in
the current group, so it is not necessary to crawl to discover them.
}
\references{
\url{https://zarr.readthedocs.io/en/latest/user-guide/consolidated_metadata.html}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/store_utils.R
\name{.normalize_array_path}
\alias{.normalize_array_path}
\title{Normalize a Zarr array path}
\usage{
.normalize_array_path(path)
}
\arguments{
\item{path}{Character vector of length 1 giving the path to be normalised.}
}
\value{
A character vector of length 1 containing the normalised path.
}
\description{
Taken from https://zarr.readthedocs.io/en/stable/spec/v2.html#logical-storage-paths
}
\keyword{internal}
